{***************************************************************************}
{ TMS Async32                                                               }
{ for Delphi 4.0,5.0,6.0 & C++Builder 4.0,5.0,6.0                           }
{                                                                           }
{ Copyright 2002 by TMS Software                                            }
{ Email : info@tmssoftware.com                                              }
{ Web : http://www.tmssoftware.com                                          }
{                                                                           }
{ The source code is given as is. The author is not responsible             }
{ for any possible damage done due to the use of this code.                 }
{ The component can be freely used in any application. The complete         }
{ source code remains property of the author and may not be distributed,    }
{ published, given or sold in any form as such. No parts of the source      }
{ code can be included in any other component or application without        }
{ written authorization of the author.                                      }
{***************************************************************************}

unit VaANSIEmulation;

interface

{$I VALIB.INC}

uses
  Classes, VaTerminal, VaComm, Graphics;

const
  ANSITxtBlack   = #27'[30m';
  ANSITxtRed     = #27'[31m';
  ANSITxtGreen   = #27'[32m';
  ANSITxtYellow  = #27'[33m';
  ANSITxtBlue    = #27'[34m';
  ANSITxtFuchsia = #27'[35m';
  ANSITxtAqua    = #27'[36m';
  ANSITxtWhite   = #27'[37m';

  ANSIBkgBlack   = #27'[40m';
  ANSIBkgRed     = #27'[41m';
  ANSIBkgGreen   = #27'[42m';
  ANSIBkgYellow  = #27'[43m';
  ANSIBkgBlue    = #27'[44m';
  ANSIBkgFuchsia = #27'[45m';
  ANSIBkgAqua    = #27'[46m';
  ANSIBkgWhite   = #27'[47m';


type

  TVaANSIEmulation = class(IVaTerminalEmulation)
  private
    FLastChar: Char;
    FLastCommand: string;
    FLastColor: TColor;
    FLastBkg: TColor;
    FCommandMode: Boolean;
  public
    constructor Create(AOwner: TComponent);override;
    destructor Destroy;override;
    procedure DataArrived(Data: PVaData; Count: Integer); override;
    procedure KeyDown(var Key: Word; Shift: TShiftState); override;
    procedure KeyPress(var Key: Char);override;
    procedure KeyUp(var Key: Word; Shift: TShiftState); override;
    function IsANSICommand(const Cmd: string): Boolean; virtual;
  end;


implementation

{ TVaANSIEmulation }

constructor TVaANSIEmulation.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  FLastChar := #32;
  FLastColor := clNone;
  FLastBkg := clNone;
  FCommandMode := False;
end;

procedure TVaANSIEmulation.DataArrived(Data: PVaData; Count: Integer);
var
  s: String;
  i,beg: Integer;
  FOldColor,FOldBkg: TColor;
begin
  inherited;
  beg := 0;
  if (FLastChar = #13) and (Data[0] = #10) then
    beg := 1;

  if FLastColor = clNone then
    FLastColor := FTerminal.Font.Color;
  if FLastBkg = clNone then
    FLastBkg := FTerminal.Color;

  FOldColor := FLastColor;
  FOldBkg := FLastBkg;

  s := '';
  for i := beg to Count - 1 do
  begin
    if (Data[i] = #27) then
    begin
      FCommandMode := True;
      FLastCommand := '';
    end;

    if not FCommandMode then
      s := s + Data[i]
    else
    begin
      FLastCommand := FLastCommand + Data[i];
      if IsANSICommand(FLastCommand) then
      begin
        if FTerminal <> nil then
          FTerminal.AddControlText(s,FOldColor,FOldBkg);
        s := '';
        FCommandMode := False;
      end;
    end;
  end;

  FLastChar := Data[Count - 1];


  if FTerminal <> nil then
    FTerminal.AddControlText(s,FLastColor,FLastBkg);
end;

destructor TVaANSIEmulation.Destroy;
begin
  inherited;
end;

function TVaANSIEmulation.IsANSICommand(const Cmd: string): Boolean;
var
  FC,LC,MC: Char;
  sub,value: string;
  v1,v2,err1,err2: Integer;
  TmpColor: TColor;

begin
  Result := False;

  if Length(Cmd) > 1 then
  begin
    FC := Cmd[1];
    LC := Cmd[Length(Cmd)];
    MC := #0;

    if Length(Cmd) > 2 then
      MC := Cmd[2];


    if (FC = #27) and (MC = '[') and (LC = 'H') then
    begin
      sub := Copy(cmd, 3, Length(Cmd) - 3);
      if sub <> '' then
      begin
        if Pos(';',sub) > 0 then
        begin
          value := Copy(sub,1,Pos(';',sub) - 1);
          Delete(sub,1,Pos(';',sub));

          val(value,v1,err1);

          val(sub,v2,err2);

          if (err1 = 0) and (err2 = 0) then
          begin
            if Assigned(FTerminal) then
            begin
              FTerminal.CaretX := v1;
              FTerminal.CaretY := v2;
            end;
            Result := True;
          end;
        end;
      end
      else
      begin
        Result := True;
      end;
    end;


    if (FC = #27) and (MC = '[') and (LC = 'A') then
    begin
    end;

    if (FC = #27) and (MC = '[') and (LC = 'B') then
    begin
    end;

    if (FC = #27) and (MC = '[') and (LC = 'C') then
    begin
    end;


    if (FC = #27) and (MC = '[') and (LC = 'm') then
    begin
      sub := Copy(cmd, 3, Length(Cmd) - 3);
      sub := sub + ';';
      while Pos(';',sub) > 0 do
      begin
        value := Copy(sub,1,Pos(';',sub) - 1);
        Delete(sub,1,Pos(';',sub));

        val(value,v1,err1);

        case v1 of
        0: if Assigned(FTerminal) then
           begin
               FLastColor := FTerminal.Font.Color;
               FLastBkg := FTerminal.Color;
           end;
        7:begin
            TmpColor := FLastColor;
            FLastColor := FLastBkg;
            FLastBkg := TmpColor;
          end;
        8:begin
            FLastColor := FLastBkg;
          end;
        30: FLastColor := clBlack;
        31: FLastColor := clRed;
        32: FLastColor := clGreen;
        33: FLastColor := clYellow;
        34: FLastColor := clBlue;
        35: FLastColor := clFuchsia;
        36: FLastColor := clAqua;
        37: FLastColor := clWhite;

        40: FLastBkg := clBlack;
        41: FLastBkg := clRed;
        42: FLastBkg := clGreen;
        43: FLastBkg := clYellow;
        44: FLastBkg := clBlue;
        45: FLastBkg := clFuchsia;
        46: FLastBkg := clAqua;
        47: FLastBkg := clWhite;
        end;

      end;
      Result := True;
    end;
  end;



end;

procedure TVaANSIEmulation.KeyDown(var Key: Word; Shift: TShiftState);
begin
  //Does nothing. In ANSI emulation we send only characters
end;

procedure TVaANSIEmulation.KeyPress(var Key: Char);
begin
  if FTerminal = nil then
    Exit;
  if Key = #13 then
    FTerminal.WriteText(#13#10)
  else
    FTerminal.WriteChar(Key);
end;

procedure TVaANSIEmulation.KeyUp(var Key: Word; Shift: TShiftState);
begin
  //Does nothing. In ANSI emulation we send only characters
end;


end.
