{***************************************************************************}
{ TMS Async32                                                               }
{ for Delphi 4.0,5.0,6.0 & C++Builder 4.0,5.0,6.0                           }
{                                                                           }
{ Copyright 1996 - 2002 by TMS Software                                     }
{ Email : info@tmssoftware.com                                              }
{ Web : http://www.tmssoftware.com                                          }
{                                                                           }
{ The source code is given as is. The author is not responsible             }
{ for any possible damage done due to the use of this code.                 }
{ The component can be freely used in any application. The complete         }
{ source code remains property of the author and may not be distributed,    }
{ published, given or sold in any form as such. No parts of the source      }
{ code can be included in any other component or application without        }
{ written authorization of the author.                                      }
{***************************************************************************}

unit VaBuffer;

{$I VALIB.INC}

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  VaTypes, VaComm;

const
  DefSize = 2048;

type
  TVaBuffer = class(TVaCommComponent)
  private
    FData: PChar;
    FSize: Integer;
    FStartPtr: Integer;
    FEndPtr: Integer;
    FFull: Boolean;
    FOnChange: TNotifyEvent;
    FOnOverflow: TNotifyEvent;
    procedure SetSize(Value: Integer);
  protected
    procedure Changed; virtual;
    procedure DataChanged(Data: PVaData; Count: Integer); override;
    procedure IncPtr(var Ptr: Integer; const Value: Integer);
    property StartPtr: Integer read FStartPtr;
    property EndPtr: Integer read FEndPtr;
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    procedure Clear;
    function Peek(var Buf; Count: Integer): Boolean;
    function Remove(Count: Integer): Boolean;
    function Read(var Buf; Count: Integer): Boolean;
    function Write(const Buf; Count: Integer): Boolean;
    function BufFree: Integer;
    function BufUsed: Integer;
    property Full: Boolean read FFull;
  published
    property Size: Integer read FSize write SetSize default DefSize;
    property OnChange: TNotifyEvent read FOnChange write FOnChange;
    property OnOverflow: TNotifyEvent read FOnOverflow write FOnOverflow;
    property Active;
  end;



implementation


{ TVaBuffer }

constructor TVaBuffer.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  FSize := 0;
  SetSize(DefSize);
end;

destructor TVaBuffer.Destroy;
begin
  SetSize(0);
  inherited Destroy;
end;

procedure TVaBuffer.Changed;
begin
  if Assigned(FOnChange) then
    FOnChange(Self);
end;

procedure TVaBuffer.SetSize(Value: Integer);
begin
  if FSize <> Value then
  begin
    Clear;
    ReallocMem(FData, Value);
    FSize := Value;
  end;
end;

procedure TVaBuffer.Clear;
begin
  FStartPtr := 0;
  FEndPtr := 0;
  FFull := False;
  Changed;
end;

procedure TVaBuffer.IncPtr(var Ptr: Integer; const Value: Integer);
begin
  if Ptr + Value < FSize then
    Inc(Ptr, Value)
  else
    Inc(Ptr, Value - FSize);
end;

function TVaBuffer.BufFree: Integer;
begin
  Result := FSize - BufUsed;
end;

function TVaBuffer.BufUsed: Integer;
begin
  if Full then
    Result := FSize
  else if FStartPtr <= FEndPtr then
    Result := FEndPtr - FStartPtr
  else
    Result := FEndPtr + FSize - FStartPtr;
end;

function TVaBuffer.Peek(var Buf; Count: Integer): Boolean;
var
  SizeToEnd: Integer;
begin
  Result := False;
  if (Count > 0) and (Count <= BufUsed) then
  begin
    if FStartPtr < FEndPtr then
      Move(FData[FStartPtr], Buf, Count)
    else
    begin
      SizeToEnd := FSize - FStartPtr;
      if Count <= SizeToEnd then
        Move(FData[FStartPtr], Buf, Count)
      else
      begin
        Move(FData[FStartPtr], Buf, SizeToEnd);
        Move(FData[0], PChar(@Buf)[SizeToEnd], Count - SizeToEnd);
      end;
    end;
    Result := True;
  end;
end;

function TVaBuffer.Remove(Count: Integer): Boolean;
begin
  Result := Count <= BufUsed;
  if Result then
  begin
    IncPtr(FStartPtr, Count);
    if Count > 0 then FFull := False;
  end;
end;

function TVaBuffer.Read(var Buf; Count: Integer): Boolean;
begin
  Result := Peek(Buf, Count);
  if Result then
  begin
    Remove(Count);
    Changed;
  end;
end;

function TVaBuffer.Write(const Buf; Count: Integer): Boolean;
var
  SizeToEnd: Integer;
begin
  Result := False;
  if (Count > 0) and (Count <= BufFree) then
  begin
    if FStartPtr <= FEndPtr then
    begin
      SizeToEnd := FSize - FEndPtr;
      if Count <= SizeToEnd then
        Move(Buf, FData[FEndPtr], Count)
      else
      begin
        Move(Buf, FData[FEndPtr], SizeToEnd);
        Move(PChar(@Buf)[SizeToEnd], FData[0], Count - SizeToEnd);
      end;
    end else
      Move(Buf, FData[FEndPtr], Count);
    IncPtr(FEndPtr, Count);
    FFull := (FStartPtr = FEndPtr);
    Result := True;
    Changed;
  end else
  begin
    if Assigned(FOnOverflow) then
      FOnOverflow(Self);
  end;
end;

procedure TVaBuffer.DataChanged(Data: PVaData; Count: Integer);
begin
  if Count > 0 then Self.Write(Data^, Count);
end;


end.
