{***************************************************************************}
{ TMS Async32                                                               }
{ for Delphi & C++Builder                                                   }
{ v1.3                                                                      }
{ Copyright 1996 - 2005 by TMS Software                                     }
{ Email : info@tmssoftware.com                                              }
{ Web : http://www.tmssoftware.com                                          }
{                                                                           }
{ The source code is given as is. The author is not responsible             }
{ for any possible damage done due to the use of this code.                 }
{ The component can be freely used in any application. The complete         }
{ source code remains property of the author and may not be distributed,    }
{ published, given or sold in any form as such. No parts of the source      }
{ code can be included in any other component or application without        }
{ written authorization of the author.                                      }
{***************************************************************************}

unit VaComm;

{$I VALIB.INC}

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  VaConst, VaTypes, VaClasses, VaObjects, VaPrst, VaUtils, VaDisplay, Registry;

const
  MAJ_VER = 1; // Major version nr.
  MIN_VER = 3; // Minor version nr.
  REL_VER = 5; // Release nr.
  BLD_VER = 0; // Build nr.

  // version history
  // 1.3.5 : Added ReadCharWait, WaitChar functions

type
  PVaData = ^TVaData;
  TVaData = array[0..32767] of Char;

  TVaCommRxCharEvent = procedure(Sender: TObject; Count: Integer) of object;
  TVaCommRxBufEvent = procedure(Sender: TObject; Data: PVaData; Count: Integer) of object;
  TVaCommErrorEvent = procedure(Sender: TObject; Errors: Integer) of object;
  TVaCommLineEvent = procedure(Sender: TObject; EvtMask: dword) of object;

  TVaCommEvent = (ceBreak, ceCts, ceDsr, ceError, ceRing, ceRlsd,
    ceRxChar, ceTxEmpty, ceRxFlag, cePErr, ceRx80Full, ceEvent1,
    ceEvent2);
  TVaCommEvents = set of TVaCommEvent;

  TVaCommOption = (coErrorChar, coNullStrip, coParityCheck);
  TVaCommOptions = set of TVaCommOption;

  TVaCommStatus = record
    Errors: dword;
    Status: TComStat;
  end;

  TVaBaudrate = (brUser, br110, br300, br600, br1200, br2400, br4800, br9600, br14400,
    br19200, br38400, br56000, br57600, br115200, br128000, br256000);
  TVaParity = (paNone, paOdd, paEven, paMark, paSpace);
  TVaDatabits = (db4, db5, db6, db7, db8);
  TVaStopbits = (sb1, sb15, sb2);

  TVaDispatcher = class;
  TVaCommComponent = class;
  TVaCustomComm = class;

  TVaCommEventThread = class(TThread)
  private
    FEvtMask: DWord;
    FEvent: TVaWaitEvent;
    FOwner: TVaCustomComm;
    FSync: boolean;
  protected
    procedure DoEvent;
    procedure Execute; override;
    procedure StopThread;
  public
    constructor Create(AOwner: TVaCustomComm);
    destructor Destroy; override;
    property Sync: boolean read FSync write FSync;
  end;

  TVaCommBuffer = class(TObject)
  private
    FData: PChar;
    FSize: Integer;
  public
    constructor Create(BufSize: Integer);
    destructor Destroy; override;
    function DataPtr: PChar;
    procedure Clear;
    procedure Write(var Buf; Count: Integer);
    procedure Read(var Buf; Count: Integer);
    procedure Peek(var Buf; Count: Integer);
    procedure Remove(Count: Integer);
    property Size: Integer read FSize;
  end;

  TVaCommWriteThread = class(TThread)
  private
    FBuffer: TVaCommBuffer;
    FWriteEvent: TVaWaitEvent;
    FCloseEvent: TVaWaitEvent;
    FFlushEvent: TVaWaitEvent;
    FFlushDoneEvent: TVaWaitEvent;
    FOverlappedEvent: TVaWaitEvent;
    FOwner: TVaCustomComm;
    FSync: boolean;
  protected
    procedure DoTxEmpty;
    procedure DoWriteError;
    procedure Execute; override;
    procedure StopThread;
  public
    constructor Create(AOwner: TVaCustomComm);
    destructor Destroy; override;
    procedure PurgeWrite;
    function WriteData(var Buf; Count: Integer): Boolean;
    property Buffer: TVaCommBuffer read FBuffer;
    property Sync: boolean read FSync write FSync;
  end;

  TWaitThread = class(TThread)
  private
    { Private declarations }
    FComm: TVaCustomComm;
    FDone: Boolean;
  protected
    procedure Execute; override;
  public
    constructor Create(AComm: TVaCustomComm);
    property Comm: TVaCustomComm read FComm write FComm;
    property Done: Boolean read FDone write FDone;
  end;


  TVaCustomComm = class(TVaComponent)
  private
    FHandle: THandle;
    FWriteSection: TVaCriticalSection;
    FGlobalSection: TVaCriticalSection;
    FProperties: TCommProp;
    FEvent: TVaWaitEvent;
    FEventThread: TVaCommEventThread;
    FWriteThread: TVaCommWriteThread;
    FEventMask: Integer;
    FAutoOpen: Boolean;
    FPortNum: TVaPortInt;
    FDeviceName: string;
    FMonitorEvents: TVaCommEvents;
    FBuffers: TVaBuffers;
    FEventPriority: TThreadPriority;
    FWritePriority: TThreadPriority;
    FDirectWrite: Boolean;
    FControlsEnabled: Boolean;
    FClients: TList;
    FBaudrate: TVaBaudrate;
    FParity: TVaParity;
//  FParityCheck: Boolean;
    FDatabits: TVaDatabits;
    FStopbits: TVaStopbits;
    FOptions: TVaCommOptions;
    FFlowControl: TVaFlowControl;
    FEventChars: TVaEventChars;
    FUpdateDCB: Boolean;
    FUserBaudrate: Integer;
    FOnOpen: TNotifyEvent;
    FOnClose: TNotifyEvent;
    FOnCts: TNotifyEvent;
    FOnDsr: TNotifyEvent;
    FOnRlsd: TNotifyEvent;
    FOnRing: TNotifyEvent;
    FOnBreak: TNotifyEvent;
    FOnError: TVaCommErrorEvent;
    FOnRxChar: TVaCommRxCharEvent;
    FOnRxFlag: TNotifyEvent;
    FOnRxBuf: TVaCommRxBufEvent;
    FOnTxEmpty: TNotifyEvent;
    FOnPErr: TNotifyEvent;
    FOnRx80Full: TNotifyEvent;
    FOnEvent1: TNotifyEvent;
    FOnEvent2: TNotifyEvent;
    FDisplay: TVaDisplay;
    FSyncThreads: boolean;
    function GetComStat(Index: Integer): Boolean;
    function GetModemStatus(Index: Integer): Boolean;
    function GetProperties: TCommProp;
    procedure SetBaudrate(Value: TVaBaudrate);
    procedure SetParity(Value: TVaParity);
//  procedure SetParityCheck(Value: Boolean);
    procedure SetDatabits(Value: TVaDatabits);
    procedure SetStopbits(Value: TVaStopbits);
    procedure SetOptions(Value: TVaCommOptions);
    procedure SetEventChars(Value: TVaEventChars);
    procedure SetFlowControl(Value: TVaFlowControl);
    procedure SetPortNum(Value: TVaPortInt);
    procedure SetMonitorEvents(Value: TVaCommEvents);
    procedure SetBuffers(Value: TVaBuffers);
    procedure SetEventPriority(Value: TThreadPriority);
    procedure SetWritePriority(Value: TThreadPriority);
    procedure SetDirectWrite(Value: Boolean);
    procedure SetUpdateDCB(Value: Boolean);
    procedure SetDeviceName(const Value: string);
    procedure SetUserBaudrate(Value: Integer);
    procedure UpdateBuffers;
    procedure UpdateTimeouts;
    procedure UpdateEventMask;
    procedure PurgeBuffers(Mode: Integer);
    procedure ToggleCommState(AState: dword);
    procedure BuffersChanged(Sender: TObject);
    procedure EventCharsChanged(Sender: TObject);
    procedure FlowControlChanged(Sender: TObject);
    procedure UpdateDisplay;
    procedure SetDisplay(const Value: TVaDisplay);
    function GetVersion: string;
    function GetVersionNr: integer;    
    procedure SetVersion(const Value: string);
  protected
    procedure Loaded; override;
    procedure CreateHandle; virtual;
    procedure DestroyHandle;
    procedure CreateThreads; virtual;
    procedure Notification(AComponent: TComponent; Operation: TOperation); override;
    procedure DestroyThreads;
    procedure CheckOpen;
    procedure DoOpen;
    procedure DoClose;
    procedure WriteLock;
    procedure WriteUnlock;
    procedure GlobalLock;
    procedure GlobalUnlock;
    procedure SetDCB;
    procedure ReadDCB(const DCB: TDCB);
    procedure Error(Msg: string);
    procedure Watch(B: Boolean; Msg: string);
    function ClearLineErrors: Boolean;
    function GetCommStatus: TVaCommStatus;
    function GetDeviceName: string;
    procedure HandleDataEvent; virtual;
    procedure HandleLineEvent(EvtMask: dword); virtual;
    procedure DestroyClients;
    procedure WriteEvents(Events: dword); virtual;
    procedure WriteBuffer(Data: PVaData; Count: Integer); virtual;
    property AutoOpen: Boolean read FAutoOpen write FAutoOpen default false;
    property Baudrate: TVaBaudrate read FBaudrate write SetBaudrate default br1200;
    property Parity: TVaParity read FParity write SetParity default paNone;
//  property ParityCheck: Boolean read FParityCheck write SetParityCheck default false;
    property Databits: TVaDatabits read FDatabits write SetDatabits default db8;
    property Stopbits: TVaStopbits read FStopbits write SetStopbits default sb1;
    property EventChars: TVaEventChars read FEventChars write SetEventChars;
    property FlowControl: TVaFlowControl read FFlowControl write SetFlowControl;
    property Options: TVaCommOptions read FOptions write SetOptions default [];
    property PortNum: TVaPortInt read FPortNum write SetPortNum default 0;
    property DeviceName: string read FDeviceName write SetDeviceName;
    property MonitorEvents: TVaCommEvents read FMonitorEvents write SetMonitorEvents default [ceError, ceRxChar, ceTxEmpty];
    property Buffers: TVaBuffers read FBuffers write SetBuffers;
    property EventPriority: TThreadPriority read FEventPriority write SetEventPriority default tpNormal;
    property WritePriority: TThreadPriority read FWritePriority write SetWritePriority default tpHighest;
    property DirectWrite: Boolean read FDirectWrite write SetDirectWrite default false;
    property UpdateDCB: Boolean read FUpdateDCB write SetUpdateDCB default True;
    property UserBaudrate: Integer read FUserBaudrate write SetUserBaudrate default 0;
    property OnOpen: TNotifyEvent read FOnOpen write FOnOpen;
    property OnClose: TNotifyEvent read FOnClose write FOnClose;
    property OnCts: TNotifyEvent read FOnCts write FOnCts;
    property OnDsr: TNotifyEvent read FOnDsr write FOnDsr;
    property OnRlsd: TNotifyEvent read FOnRlsd write FOnRlsd;
    property OnRing: TNotifyEvent read FOnRing write FOnRing;
    property OnBreak: TNotifyEvent read FOnBreak write FOnBreak;
    property OnError: TVaCommErrorEvent read FOnError write FOnError;
    property OnRxChar: TVaCommRxCharEvent read FOnRxChar write FOnRxChar;
    property OnRxBuf: TVaCommRxBufEvent read FOnRxBuf write FOnRxBuf;
    property OnRxFlag: TNotifyEvent read FOnRxFlag write FOnRxFlag;
    property OnTxEmpty: TNotifyEvent read FOnTxEmpty write FOnTxEmpty;
    property OnPErr: TNotifyEvent read FOnPErr write FOnPErr;
    property OnRx80Full: TNotifyEvent read FOnRx80Full write FOnRx80Full;
    property OnEvent1: TNotifyEvent read FOnEvent1 write FOnEvent1;
    property OnEvent2: TNotifyEvent read FOnEvent2 write FOnEvent2;
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    procedure Open;
    procedure Close;
    function Active: Boolean;
    procedure InsertClient(Value: TVaDispatcher);
    procedure RemoveClient(Value: TVaDispatcher);
    function WriteBuf(var Buf; Count: Integer): Integer;
    function WriteText(const s: string): Boolean;
    function WriteChar(Ch: Char): Boolean;
    function ReadBuf(var Buf; Count: Integer): Integer;
    function ReadText: string;
    function ReadChar(var Ch: Char): Boolean;
    function ReadCharWait(var Ch: Char; TimeOut: Integer): Boolean;    
    function WaitChar(TimeOut: Integer): boolean;
    function ReadBufFree: Integer;
    function ReadBufUsed: Integer;
    function WriteBufFree: Integer;
    function WriteBufUsed: Integer;
    procedure PurgeRead;
    procedure PurgeWrite;
    procedure PurgeReadWrite;
    procedure SetDTR(Value: Boolean);
    procedure SetRTS(Value: Boolean);
    procedure SetXon(Value: Boolean);
    procedure SetBreak(Value: Boolean);
    procedure ResetDev;
    procedure EnableControls;
    procedure DisableControls;
    procedure ResetPortParameters;
    procedure GetComPortNames(Value: TStringList);
    property Handle: THandle read FHandle;
    property CTS: Boolean index 1 read GetModemStatus;
    property DSR: Boolean index 2 read GetModemStatus;
    property RING: Boolean index 3 read GetModemStatus;
    property RLSD: Boolean index 4 read GetModemStatus;
    property CTSHold: Boolean index 1 read GetComStat;
    property DSRHold: Boolean index 2 read GetComStat;
    property RLSDHold: Boolean index 3 read GetComStat;
    property XOffHold: Boolean index 4 read GetComStat;
    property XOffSent: Boolean index 5 read GetComStat;
    property Properties: TCommProp read GetProperties;
    property Display : TVaDisplay read FDisplay write SetDisplay;
    property SyncThreads: boolean read FSyncThreads write FSyncThreads;
    property Version: string read GetVersion write SetVersion;
  end;

  TVaComm = class(TVaCustomComm)
    property AutoOpen;
    property Baudrate;
    property Parity;
//  property ParityCheck;
    property Databits;
    property Stopbits;
    property EventChars;
    property FlowControl;
    property Options;
    property PortNum;
    property DeviceName;
    property MonitorEvents;
    property Buffers;
    property EventPriority;
    property WritePriority;
    property DirectWrite;
    property UpdateDCB;
    property UserBaudrate;
    property OnOpen;
    property OnClose;
    property OnCts;
    property OnDsr;
    property OnRlsd;
    property OnRing;
    property OnBreak;
    property OnError;
    property OnRxChar;
    property OnRxBuf;
    property OnRxFlag;
    property OnTxEmpty;
    property OnPErr;
    property OnRx80Full;
    property OnEvent1;
    property OnEvent2;
    property Display;
    property Version;
  end;

  IVaCommUser = interface(IUnknown)
    procedure LineChanged(Events: dword);
    procedure DataChanged(Data: PVaData; Count: Integer);
  end;

  TVaDispatcher = class(TObject)
  private
    FActive: Boolean;
    FComm: TVaCustomComm;
    FOwner: IVaCommUser;
    procedure LineEvent(Events: dword);
    procedure DataEvent(Data: PVaData; Count: Integer);
  public
    constructor Create(AOwner: IVaCommUser);
    destructor Destroy; override;
    procedure PurgeRead;
    procedure PurgeWrite;
    procedure PurgeReadWrite;
    function WriteBuf(var Buf; Count: Integer): Integer;
    function WriteText(const s: string): Boolean;
    function WriteChar(Ch: Char): Boolean;
    procedure SetDTRState(Value: Boolean);
    property Active: Boolean read FActive write FActive;
    property Comm: TVaCustomComm read FComm write FComm;
  end;

  TVaCommComponent = class(TVaComponent, IVaCommUser)
  private
    FComm: TVaComm;
    FDispatcher: TVaDispatcher;
    function GetActive: Boolean;
    procedure SetActive(Value: Boolean);
    procedure SetComm(Value: TVaComm);
  protected
    procedure Notification(AComponent: TComponent; Operation: TOperation); override;
    procedure DataChanged(Data: PVaData; Count: Integer); virtual;
    procedure LineChanged(Events: dword); virtual;
    property Active: Boolean read GetActive write SetActive;
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    property Dispatcher: TVaDispatcher read FDispatcher;
  published
    property Comm: TVaComm read FComm write SetComm;
  end;

                                
implementation


const
  sCantPerform = 'Can''t perform this operation while port is open';
  sCantWhileClosed = 'Can''t perform this operation while port is closed';
  sInvalidDevice = 'Specified device not found (%s).';
  sErrorOpen = 'Can''t open specified device %s';
  sErrorRead = 'Read error (%d), %s';
  sErrorWrite = 'Write error (%d), %s';
  sErrorSetTimeouts = 'Error setting timeouts (%d), %s';
  sErrorSetCommMask = ' Error setting EventMask (%d), %s';
  sErrorSetupComm = 'Error initializing buffers (%d), %s';
  sErrorPurgeComm = 'Error clearing buffers (%d), %s';
  sErrorClearCommError = 'Error retrieving statistics (%d), %s';
  sErrorGetCommState = 'Error reading dcb (%d), %s';
  sErrorSetCommState = 'Error writing dcb (%d), %s';
  sErrorEscapeComm = 'Error calling comm function (%d), %s';
  sErrorGetModemStatus = 'Error reading modem state (%d), %s';
  sErrorProperties = 'Error reading port properties (%d), %s';
  sErrorDTRToggle = 'DTR line updates not allowed when flowcontrol is set to Handshake.';
  sErrorRTSToggle = 'RTS line updates not allowed when flowcontrol is set to Handshake or Toggle.';
  sOutputBufferFull = 'Output buffer is full. Please check comm status.';
  sErrDefaultConfig = 'Error reading default port parameters.';

  EventList: array[TVaCommEvent] of dword =
    (EV_BREAK, EV_CTS, EV_DSR, EV_ERR, EV_RING, EV_RLSD, EV_RXCHAR,
     EV_TXEMPTY, EV_RXFLAG, EV_PERR, EV_RX80FULL, EV_EVENT1, EV_EVENT2);

  DCB_BaudRates: array[TVaBaudRate] of dword =
    ( 0, CBR_110, CBR_300, CBR_600, CBR_1200, CBR_2400, CBR_4800, CBR_9600,
      CBR_14400, CBR_19200, CBR_38400, CBR_56000, CBR_57600, CBR_115200,
      CBR_128000, CBR_256000);

  DCB_DataBits: array[TVaDatabits] of Integer =
    ( 4, 5, 6, 7, 8);

  DCB_Parity: array[TVaParity] of Integer =
    (NOPARITY, ODDPARITY, EVENPARITY, MARKPARITY, SPACEPARITY);

//DCB_PARITYCHECK: array[Boolean] of Integer = (0, fParity);

  DCB_StopBits: array[TVaStopbits] of Integer =
    (ONESTOPBIT, ONE5STOPBITS, TWOSTOPBITS);

  DCB_Options: array[TVaCommOption] of Integer =
    (fErrorChar, fNull, fParity);

  DCB_XONXOFFOUT: array[Boolean] of Integer = (0, fOutX);
  DCB_XONXOFFIN: array[Boolean] of Integer  = (0, fInX);
  DCB_OUTCTSFLOW: array[Boolean] of Integer = (0, fOutxCtsFlow);
  DCB_OUTDSRFLOW: array[Boolean] of Integer = (0, fOutxDsrFlow);
  DCB_CONTROLDTR: array[TVaControlDtr] of Integer =
    (fDtrControlDisable, fDtrControlEnable, fDtrControlHandshake);
  DCB_CONTROLRTS: array[TVaControlRts] of Integer =
    (fRtsControlDisable, fRtsControlEnable, fRtsControlHandshake, fRtsControlToggle);

  ESC_DTR: array[Boolean] of dword = (Windows.CLRDTR, Windows.SETDTR);
  ESC_RTS: array[Boolean] of dword = (Windows.CLRRTS, Windows.SETRTS);
  ESC_BREAK: array[Boolean] of dword = (Windows.CLRBREAK, Windows.SETBREAK);
  ESC_XOFFXON: array[Boolean] of dword = (Windows.SETXOFF, Windows.SETXON);

  MAX_XONLIM = 1024; //NT 4 only


{ TVaCommEventThread }

constructor TVaCommEventThread.Create(AOwner: TVaCustomComm);
begin
  FOwner := AOwner;
  FEvent := TVaWaitEvent.Create;
  FreeOnTerminate := False;
  FSync := true;
  inherited Create(True); //create suspended
end;

destructor TVaCommEventThread.Destroy;
begin
  FEvent.Free;
  inherited Destroy;
end;

procedure TVaCommEventThread.Execute;
var
  Overlapped: TOverlapped;

  procedure HandleEvent;
  begin
    if not Terminated then
    begin
      if Sync then
        Synchronize(DoEvent)
      else
        DoEvent;  
      FEvent.ResetEvent;
    end;
  end;

begin
  FillChar(Overlapped, Sizeof(Overlapped), 0);
  Overlapped.hEvent := FEvent.Handle;
  repeat
    if WaitCommEvent(FOwner.Handle, FEvtMask, @Overlapped) then
      HandleEvent
     else
      //if GetLastError = ERROR_IO_PENDING then
      //begin
        if (FEvent.WaitFor(INFINITE) = wrSignaled) then
          HandleEvent;
      //end;
  until Terminated;
end;

procedure TVaCommEventThread.StopThread;
begin
  Terminate;
  FEvent.SetEvent;
end;

procedure TVaCommEventThread.DoEvent;
begin
  try
    FOwner.UpdateDisplay;
    if (FEvtMask and EV_RXCHAR) > 0 then
    begin
      FOwner.HandleDataEvent;
      FEvtMask := FEvtMask and not EV_RXCHAR;
    end;
    if FEvtMask <> 0 then
      FOwner.HandleLineEvent(FEvtMask);
  except
    Application.HandleException(Self);
  end;
end;

{ TVaCommBuffer }

constructor TVaCommBuffer.Create(BufSize: Integer);
begin
  FSize := 0;
  GetMem(FData, BufSize);
end;

destructor TVaCommBuffer.Destroy;
begin
  FreeMem(FData);
  inherited Destroy;
end;

function TVaCommBuffer.DataPtr: PChar;
begin
  Result := FData;
end;

procedure TVaCommBuffer.Clear;
begin
  FSize := 0;
end;

procedure TVaCommBuffer.Write(var Buf; Count: Integer);
begin
  Move(Buf, FData[FSize], Count);
  Inc(FSize, Count);
end;

procedure TVaCommBuffer.Peek(var Buf; Count: Integer);
begin
  Move(FData[0], Buf, Count);
end;

procedure TVaCommBuffer.Read(var Buf; Count: Integer);
begin
  Move(FData[0], Buf, Count);
  Remove(Count);
end;

procedure TVaCommBuffer.Remove(Count: Integer);
begin
  Move(FData[Count], FData[0], FSize - Count);
  Dec(FSize, Count);
end;

{ TVaCommWriteThread }

constructor TVaCommWriteThread.Create(AOwner: TVaCustomComm);
begin
  FOwner := AOwner;
  FWriteEvent := TVaWaitEvent.Create;
  FCloseEvent := TVaWaitEvent.Create;
  FFlushEvent := TVaWaitEvent.Create;
  FFlushDoneEvent := TVaWaitEvent.Create;
  FOverlappedEvent := TVaWaitEvent.Create;
  FBuffer := TVaCommBuffer.Create(FOwner.Buffers.WriteSize);
  FSync := true;  
  inherited Create(True);
  FreeOnTerminate := False;
end;

destructor TVaCommWriteThread.Destroy;
begin
  FBuffer.Free;
  FWriteEvent.Free;
  FCloseEvent.Free;
  FFlushEvent.Free;
  FFlushDoneEvent.Free;
  FOverlappedEvent.Free;
  inherited Destroy;
end;

procedure TVaCommWriteThread.DoTxEmpty;
begin
  FOwner.HandleLineEvent(EV_TXEMPTY);
end;

procedure TVaCommWriteThread.DoWriteError;
begin
  FOwner.HandleLineEvent(EV_ERR);
end;

procedure TVaCommWriteThread.Execute;
var
  BytesToWrite, BytesWritten: DWORD;
  WaitHandles: array[0..2] of THandle;
  WriteHandles: array[0..2] of THandle;
  BufPtr: PChar;
  Overlapped: TOverlapped;

  procedure WriteError;
  begin
    if Sync then
      Synchronize(DoWriteError)
    else
      DoWriteError;
  end;

  procedure WriteCompleted;
  begin
    FOwner.WriteLock;
    try
      Buffer.Remove(BytesWritten);
    finally
      FOwner.WriteUnlock;
    end;
    if Buffer.Size > 0 then
      FWriteEvent.SetEvent
    else
      begin
        FWriteEvent.ResetEvent;
        if Sync then
          Synchronize(DoTxEmpty)
        else
          DoTxEmpty;  
      end;
  end;

begin
  WaitHandles[0] := FCloseEvent.Handle;
  WaitHandles[1] := FFlushEvent.Handle;
  WaitHandles[2] := FWriteEvent.Handle;

  WriteHandles[0] := FCloseEvent.Handle;
  WriteHandles[1] := FFlushEvent.Handle;
  WriteHandles[2] := FOverlappedEvent.Handle;

  repeat
    case WaitForMultipleObjects(3, @WaitHandles, False, INFINITE) of
      WAIT_OBJECT_0:
        Terminate;
      WAIT_OBJECT_0 + 1:
        begin
          FFlushEvent.ResetEvent;
          FFlushDoneEvent.SetEvent;
        end;
      WAIT_OBJECT_0 + 2:
        begin
          FOwner.WriteLock;
          BytesToWrite := Buffer.Size;
          BufPtr := Buffer.DataPtr;
          FOwner.WriteUnlock;

          FillChar(Overlapped, Sizeof(Overlapped), 0);
          Overlapped.hEvent := FOverlappedEvent.Handle;

          if WriteFile(FOwner.Handle, BufPtr[0], BytesToWrite, BytesWritten, @Overlapped) then
            WriteCompleted
          else
            if GetLastError = ERROR_IO_PENDING then
            begin
              case WaitForMultipleObjects(3, @WriteHandles, False, INFINITE) of
                WAIT_OBJECT_0:
                  Terminate;
                WAIT_OBJECT_0 + 1:
                  begin
                    FFlushEvent.ResetEvent;
                    FFlushDoneEvent.SetEvent;
                  end;
                WAIT_OBJECT_0 + 2:
                  if GetOverlappedResult(FOwner.Handle, Overlapped, BytesWritten, True) then
                    WriteCompleted;
              end;
            end else
              WriteError;
        end;
    end;
  until Terminated;
end;

procedure TVaCommWriteThread.StopThread;
begin
  FCloseEvent.SetEvent;
end;

procedure TVaCommWriteThread.PurgeWrite;
begin
  FFlushEvent.SetEvent;
  if FFlushDoneEvent.WaitFor(2000) = wrSignaled then
  begin
    FOwner.WriteLock;
    FBuffer.Clear;
    FOwner.WriteUnlock;
  end;
end;

function TVaCommWriteThread.WriteData(var Buf; Count: Integer): Boolean;
var
  NewSize: Integer;
  TimeEvent: TVaTimeEvent;
begin
  Result := false;
  NewSize := FBuffer.Size + Count;
  if NewSize > FOwner.Buffers.WriteSize then
  begin
    //protect the buffer from ever overflowing!
    //this will block all running threads...
    //but with proper settings this will never occure.
    InitTimer(TimeEvent, FOwner.Buffers.WriteTimeout);
    repeat
      NewSize := FBuffer.Size + Count;
      if TimerExpired(TimeEvent) then Exit;
    until NewSize < FOwner.Buffers.WriteSize;
  end;

  FOwner.WriteLock;
  try
    FBuffer.Write(Buf, Count);
    FWriteEvent.SetEvent;
    Result := True;
  finally
    FOwner.WriteUnlock;
  end;
end;

{ TVaCustomComm }

constructor TVaCustomComm.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  FSyncThreads := true;
  FHandle := INVALID_HANDLE_VALUE;
  FBaudrate := br1200;
  FParity := paNone;
  FDatabits := db8;
  FStopbits := sb1;
  FAutoOpen := false;
  FPortNum := 0;
  FDeviceName := 'COM%d';
  FMonitorEvents := [ceError, ceRxChar, ceTxEmpty];
  FOptions := [];
  FEventPriority := tpNormal;
  FWritePriority := tpHighest;
  FUpdateDCB := True;
  FDirectWrite := false;
  FControlsEnabled := True;
  FClients := TList.Create;
  FBuffers := TVaBuffers.Create;
  FBuffers.OnChange := BuffersChanged;
  FEventChars := TVaEventChars.Create;
  FEventChars.OnChange := EventCharsChanged;
  FFlowControl := TVaFlowControl.Create;
  FFlowControl.OnChange := FlowControlChanged;
  FEvent := TVaWaitEvent.Create;
  FWriteSection := TVaCriticalSection.Create;
  FGlobalSection := TVaCriticalSection.Create;
end;

destructor TVaCustomComm.Destroy;
begin
  Close;
  DestroyClients;
  FClients.Free;
  FBuffers.Free;
  FEventChars.Free;
  FFlowControl.Free;
  FEvent.Free;
  FWriteSection.Free;
  FGlobalSection.Free;
  inherited Destroy;
end;

procedure TVaCustomComm.Loaded;
begin
  inherited Loaded;
  if not (csDesigning in ComponentState) then
    if AutoOpen then Open;
end;

procedure TVaCustomComm.WriteLock;
begin
  FWriteSection.Enter;
end;

procedure TVaCustomComm.WriteUnlock;
begin
  FWriteSection.Leave;
end;

procedure TVaCustomComm.GlobalLock;
begin
  FGlobalSection.Enter;
end;

procedure TVaCustomComm.GlobalUnlock;
begin
  FGlobalSection.Leave;
end;

function TVaCustomComm.GetDeviceName: string;
var
  verinfo: TOSVersionInfo;
begin
  VerInfo.dwOSVersionInfoSize := SizeOf(TOSVersionInfo);
  GetVersionEx(verinfo);

  if (verinfo.dwPlatformId = VER_PLATFORM_WIN32_NT) then
    Result := '\\.\' + Format(DeviceName, [PortNum])
  else
    Result := Format(DeviceName, [PortNum]);  
end;

procedure TVaCustomComm.Error(Msg: string);
var
  ErrorCode: Integer;
begin
  ErrorCode := GetLastError;
  raise EVaCommError.CreateFmt(Msg, [ErrorCode,
    SysErrorMessage(ErrorCode)]);
end;

procedure TVaCustomComm.Watch(B: Boolean; Msg: string);
begin
  if not B then Error(Msg);
end;

procedure TVaCustomComm.CreateHandle;
begin
  FHandle := CreateFile(PChar(GetDeviceName),
    GENERIC_READ or GENERIC_WRITE, 0, nil,
    OPEN_EXISTING, FILE_FLAG_OVERLAPPED, 0);

  if FHandle = INVALID_HANDLE_VALUE then
  begin
    if GetLastError = ERROR_FILE_NOT_FOUND then
      Error(Format(sInvalidDevice, [GetDeviceName]))
    else
      Error(Format(sErrorOpen, [GetDeviceName])+ ' (Error %d, %s)');
  end;
end;

procedure TVaCustomComm.DestroyHandle;
begin
  CloseHandle(FHandle);
  FHandle := INVALID_HANDLE_VALUE;
end;

procedure TVaCustomComm.CreateThreads;
begin
  if FEventMask <> 0 then
  begin
    FEventThread := TVaCommEventThread.Create(Self);
    FEventThread.Priority := EventPriority;
    FEventThread.Sync := SyncThreads;
    FEventThread.Resume;
  end;

  if not FDirectWrite then
  begin
    FWriteThread := TVaCommWriteThread.Create(Self);
    FWriteThread.Priority := WritePriority;
    FWriteThread.Sync := SyncThreads;
    FWriteThread.Resume;
  end;
end;

procedure TVaCustomComm.DestroyThreads;
begin
  if FEventThread <> nil then
  begin
    FEventThread.StopThread;
    FEventThread.WaitFor;  //Do not free until thread is finished executing
    FEventThread.Free;
    FEventThread := nil;
  end;

  if FWriteThread <> nil then
  begin
    FWriteThread.StopThread;
    FWriteThread.WaitFor;  //Do not free until thread is finished executing
    FWriteThread.Free;
    FWriteThread := nil;
  end;
end;

procedure TVaCustomComm.UpdateEventMask;
var
  EvtIndex: TVaCommEvent;
begin
  FEventMask := 0;
  for EvtIndex := Low(TVaCommEvent) to High(TVaCommEvent) do
    if EvtIndex in MonitorEvents then
      FEventMask := dword(FEventMask) or EventList[EvtIndex];

  if not FDirectWrite then
    FEventMask := FEventMask and not EV_TXEMPTY;

  if Active then
    Watch(SetCommMask(Handle, FEventMask), sErrorSetCommMask);
end;

procedure TVaCustomComm.PurgeBuffers(Mode: Integer);
begin
  if Active then
  begin
    GlobalLock;
    try
      if (Mode <> cPurgeRead) and (FWriteThread <> nil) then
        FWriteThread.PurgeWrite;
      Watch(PurgeComm(Handle, Mode), sErrorPurgeComm);
    finally
      GlobalUnlock;
    end;
  end;
end;

procedure TVaCustomComm.UpdateTimeouts;
var
  CommTimeouts: TCommTimeouts;
begin
  FillChar(CommTimeOuts, Sizeof(CommTimeOuts), 0);
  CommTimeouts.ReadIntervalTimeout := MAXDWORD;
  Watch(SetCommTimeOuts(Handle, CommTimeOuts), sErrorSetTimeouts);
end;

procedure TVaCustomComm.UpdateBuffers;
begin
  if Active then
    with Buffers do
      Watch(SetupComm(Handle, ReadSize, WriteSize), sErrorSetupComm);
end;

procedure TVaCustomComm.ToggleCommState(AState: dword);
begin
  if Active then
    Watch(EscapeCommFunction(Handle, AState), sErrorEscapeComm);
end;

function TVaCustomComm.GetProperties: TCommProp;
begin
  FillChar(FProperties, Sizeof(FProperties), 0);
  if Active then
    Watch(GetCommProperties(Handle, FProperties), sErrorProperties)
  else Error(sCantWhileClosed);
end;

procedure TVaCustomComm.Open;
begin
  CheckOpen;
  CreateHandle;
  try
    if Active then
    begin
      SetDCB;
      UpdateBuffers;
      UpdateTimeouts;
      UpdateEventMask;
      CreateThreads;
      DoOpen;
    end;
  except
    DestroyHandle;
    raise;
  end;
end;

procedure TVaCustomComm.Close;
begin
  if Active then
  begin
    GlobalLock;
    try
      DestroyThreads;
    finally
      GlobalUnlock;
    end;
    DestroyHandle;
    DoClose;
    Sleep(300); // avoid typical serial comm API issue that closing serial port takes a while to propagate
  end;
end;

function TVaCustomComm.ClearLineErrors: Boolean;
var
  Status: TVaCommStatus;
begin
  Status := GetCommStatus;
  Result := Status.Errors <> 0;
  try
    if (Assigned(FOnError)) and (Result) then
      FOnError(Self, Status.Errors);
  except //make sure we continue
    Application.HandleException(Self);
  end;
end;

function TVaCustomComm.WriteBuf(var Buf; Count: Integer): Integer;
var
  Overlapped: TOverlapped;
begin
  if Active then
  begin
    if not FDirectWrite then
    begin
      if FWriteThread.WriteData(Buf, Count) then
        Result := Count else Error(sOutputBufferFull);
      Exit;
    end;

    WriteLock;
    try
      FillChar(Overlapped, Sizeof(Overlapped), 0);
      Overlapped.hEvent := FEvent.Handle;
      if (not WriteFile(Handle, Buf, Count, LongWord(Result),
        @Overlapped)) and (GetLastError <> ERROR_IO_PENDING) then
          Error(sErrorWrite);
      if FEvent.WaitFor(Buffers.ReadTimeout) <> wrSignaled then
        Result := 0
      else GetOverlappedResult(Handle, Overlapped, LongWord(Result), False);
    finally
      WriteUnlock;
    end;
  end;
end;

function TVaCustomComm.ReadBuf(var Buf; Count: Integer): Integer;
var
  Overlapped: TOverlapped;
begin
  if Active then
  begin
    GlobalLock;
    try
      FillChar(Overlapped, Sizeof(Overlapped), 0);
      Overlapped.hEvent := FEvent.Handle;
      if (not ReadFile(Handle, Buf, Count, LongWord(Result), @Overlapped)) and
        (GetLastError <> ERROR_IO_PENDING) then
          Error(sErrorRead);
      if FEvent.WaitFor(Buffers.ReadTimeout) <> wrSignaled then
        Result := 0
      else GetOverlappedResult(Handle, Overlapped, LongWord(Result), False);
    finally
      GlobalUnlock;
    end;
  end;
end;

function TVaCustomComm.WriteText(const S: string): Boolean;
var
  Len: Integer;
begin
  Len := Length(S);
  Result := WriteBuf(Pointer(S)^, Len) = Len;
end;

function TVaCustomComm.WriteChar(Ch: Char): Boolean;
begin
  Result := WriteBuf(Ch, 1) = 1;
end;

function TVaCustomComm.ReadText: string;
var
  Count: Integer;
begin
  Count := ReadBufUsed;
  if Count > 0 then
  begin
    SetLength(Result, Count);
    ReadBuf(Pointer(Result)^, Count);
  end
  else
    Result := '';
end;

function TVaCustomComm.WaitChar(TimeOut: Integer): boolean;
var
  wt: TWaitThread;
begin
  Result := false;
  wt := TWaitThread.Create(Self);
  if WaitForSingleObject(wt.Handle, Timeout) = WAIT_OBJECT_0 then
    Result := true
  else
    wt.Done := true;
end;

function TVaCustomComm.ReadCharWait(var Ch: Char; TimeOut: Integer): Boolean;
begin
  WaitChar(TimeOut);
  Result := ReadBuf(Ch, 1) = 1;
end;

function TVaCustomComm.ReadChar(var Ch: Char): Boolean;
begin
  Result := ReadBuf(Ch, 1) = 1;
end;

function TVaCustomComm.GetCommStatus: TVaCommStatus;
begin
  GlobalLock;
  try
    FillChar(Result, Sizeof(Result), 0);
    if Active then
      Watch(ClearCommError(Handle, Result.Errors, @Result.Status), sErrorClearCommError);
  finally
    GlobalUnlock;
  end;
end;

function TVaCustomComm.ReadBufUsed: Integer;
begin
  if Active then
    Result := GetCommStatus.Status.cbInQue
  else
    Result := 0;
end;

function TVaCustomComm.ReadBufFree: Integer;
begin
  if Active then
    Result := dword(Buffers.ReadSize) - GetCommStatus.Status.cbInQue
  else Result := 0;
end;

function TVaCustomComm.WriteBufUsed: Integer;
begin
  Result := 0;
  if Active then
  begin
    if FDirectWrite then
      Result := GetCommStatus.Status.cbOutQue
    else
    begin
      WriteLock;
      try
        if FWriteThread <> nil then
          Result := FWriteThread.Buffer.Size;
      finally
        WriteUnlock;
      end;
    end;
  end;
end;

function TVaCustomComm.WriteBufFree: Integer;
begin
  Result := 0;
  if Active then
  begin
    if FDirectWrite then
      Result := dword(Buffers.WriteSize) - GetCommStatus.Status.cbOutQue
    else
    begin
      WriteLock;
      try
        if FWriteThread <> nil then
          Result := Buffers.WriteSize - FWriteThread.Buffer.Size;
      finally
        WriteUnlock;
      end;
    end;
  end;
end;

procedure TVaCustomComm.PurgeRead;
begin
  PurgeBuffers(cPurgeRead);
end;

procedure TVaCustomComm.PurgeWrite;
begin
  PurgeBuffers(cPurgeWrite);
end;

procedure TVaCustomComm.PurgeReadWrite;
begin
  PurgeBuffers(cPurgeReadWrite);
end;

procedure TVaCustomComm.HandleDataEvent;
var
  Data: PVaData;
  Count: Integer;
begin
  Count := ReadBufUsed;
  if (Count > 0) then
  begin
    if FClients.Count > 0 then
    begin
      GetMem(Data, Count);
      try
        ReadBuf(Data^, Count);
        WriteBuffer(Data, Count);
      finally
        FreeMem(Data, Count);
      end;
    end else
    if Assigned(FOnRxChar) then
      FOnRxChar(Self, Count);
  end;
end;

procedure TVaCustomComm.HandleLineEvent(EvtMask: dword);
begin
  if BitOn(EV_ERR, EvtMask) then
    ClearLineErrors;

  if FClients.Count > 0 then
    WriteEvents(EvtMask);

  if BitOn(EV_RXFLAG, EvtMask) then
    if Assigned(FOnRxFlag) then FOnRxFlag(Self);

  if BitOn(EV_CTS, EvtMask) then
    if Assigned(FOnCts) then FOnCts(Self);

  if BitOn(EV_DSR, EvtMask) then
    if Assigned(FOnDsr) then FOnDsr(Self);

  if BitOn(EV_RING, EvtMask) then
    if Assigned(FOnRing) then FOnRing(Self);

  if BitOn(EV_RLSD, EvtMask) then
    if Assigned(FOnRlsd) then FOnRlsd(Self);

  if BitOn(EV_TXEMPTY, EvtMask) then
    if Assigned(FOnTxEmpty) then FOnTxEmpty(Self);

  if BitOn(EV_BREAK, EvtMask) then
    if Assigned(FOnBreak) then FOnBreak(Self);

  if BitOn(EV_PERR, EvtMask) then
    if Assigned(FOnPErr) then FOnPErr(Self);

  if BitOn(EV_RX80FULL, EvtMask) then
    if Assigned(FOnRx80Full) then FOnRx80Full(Self);

  if BitOn(EV_EVENT1, EvtMask) then
    if Assigned(FOnEvent1) then FOnEvent1(Self);

  if BitOn(EV_EVENT2, EvtMask) then
    if Assigned(FOnEvent2) then FOnEvent2(Self);
end;

procedure TVaCustomComm.DoOpen;
begin
  try
    if Assigned(FOnOpen) then
      FOnOpen(Self);
  except
    Application.HandleException(Self);
  end;
end;

procedure TVaCustomComm.DoClose;
begin
  if not (csDestroying in ComponentState) then
    if Assigned(FOnClose) then FOnClose(Self);
end;

function TVaCustomComm.Active: Boolean;
begin
  Result := Handle <> INVALID_HANDLE_VALUE;
end;

procedure TVaCustomComm.CheckOpen;
begin
  if Active then Error(sCantPerform);
end;

procedure TVaCustomComm.SetDTR(Value: Boolean);
begin
  if (FlowControl.ControlDtr = dtrHandshake) then
    raise EVaCommError.Create(sErrorDTRToggle);
  ToggleCommState(ESC_DTR[Value]);
end;

procedure TVaCustomComm.SetRTS(Value: Boolean);
begin
  if (FlowControl.ControlRts = rtsHandshake) or
     (FlowControl.ControlRts = rtsToggle) then
       raise EVaCommError.Create(sErrorRTSToggle);
  ToggleCommState(ESC_RTS[Value]);
end;

procedure TVaCustomComm.SetBreak(Value: Boolean);
begin
  ToggleCommState(ESC_BREAK[Value]);
end;

procedure TVaCustomComm.SetXOn(Value: Boolean);
begin
  ToggleCommState(ESC_XOFFXON[Value]);
end;

procedure TVaCustomComm.ResetDev;
begin
  ToggleCommState(Windows.RESETDEV);
end;

procedure TVaCustomComm.SetDCB;
var
  DCB: TDCB;
  Index: TVaCommOption;
begin
  if (Active) and (UpdateDCB) then
  begin
    FillChar(DCB, Sizeof(DCB), 0);

    if Baudrate = brUser then
      DCB.BaudRate := FUserBaudrate
    else DCB.BaudRate := DCB_BAUDRATES[FBaudRate];

    DCB.Parity := DCB_PARITY[FParity];
    DCB.Stopbits := DCB_STOPBITS[FStopbits];
    DCB.Bytesize := DCB_DATABITS[FDatabits];

    DCB.XOnChar := EventChars.XonChar;
    DCB.XOffChar := EventChars.XOffChar;
    DCB.ErrorChar := EventChars.ErrorChar;
    DCB.EvtChar := EventChars.EventChar;
    DCB.EofChar := EventChars.EofChar;
    DCB.XOnLim := Buffers.XonLimit;
    DCB.XOffLim := Buffers.XOffLimit;

    DCB.Flags := fBinary; {or fAbortOnError;}

    with FFlowControl do
    begin
      DCB.Flags := DCB.Flags or
        DCB_OUTCTSFLOW[OutCtsFlow] or
        DCB_OUTDSRFLOW[OutDsrFlow] or
        DCB_XONXOFFOUT[XonXoffOut] or
        DCB_XONXOFFIN [XonXoffIn ] or
        DCB_CONTROLDTR[ControlDtr] or
        DCB_CONTROLRTS[ControlRts];
      if DsrSensitivity then
        DCB.Flags := DCB.Flags or fDsrSensitivity;
      if TxContinueOnXoff then
        DCB.Flags := DCB.Flags or fTxContinueOnXoff;
    end;

    for Index := Low(TVaCommOption) to High(TVaCommOption) do
      if Index in Options then
        DCB.Flags := DCB.Flags or DCB_Options[Index];
    Watch(SetCommState(Handle, DCB), sErrorSetCommState);
  end;
end;

procedure TVaCustomComm.ReadDCB(const DCB: TDCB);
var
  Br: TVaBaudrate;
  Pa: TVaParity;
  Sb: TVaStopbits;
  Db: TVaDatabits;
  Dtr: TVaControlDtr;
  Rts: TVaControlRts;
  Option: TVaCommOption;
  Valid: Boolean;
begin
  Valid := false;
  for Br := Low(TVaBaudRate) to High(TVaBaudRate) do
    if DCB.Baudrate = DCB_BAUDRATES[Br] then
    begin
      Valid := True;
      FBaudrate := Br;
      FUserBaudRate := 0;
      Break;
    end;

  if not Valid then
  begin
    FBaudrate := brUser;
    FUserBaudrate := DCB.Baudrate;
  end;

  Valid := false;
  for Pa := Low(TVaParity) to High(TVaParity) do
    if DCB.Parity = DCB_PARITY[Pa] then
    begin
      Valid := True;
      FParity := Pa;
      Break;
    end;
  if not Valid then FParity := paNone;

  Valid := false;
  for Db := Low(TVaDataBits) to High(TVaDataBits) do
    if DCB.ByteSize = DCB_DataBits[Db] then
    begin
      Valid := True;
      FDatabits := Db;
      Break;
    end;
  if not Valid then FDatabits := db8;

  Valid := false;
  for Sb := Low(TVaStopBits) to High(TVaStopBits) do
    if DCB.StopBits = DCB_StopBits[Sb] then
    begin
      Valid := True;
      FStopbits := Sb;
      Break;
    end;
  if not Valid then FStopbits := sb1;

  with FFlowControl do
  begin
    Valid := false;
    for Dtr := Low(TVaControlDtr) to High(TVaControlDtr) do
      if (DCB.Flags and DCB_ControlDtr[Dtr]) > 0 then
      begin
        Valid := True;
        ControlDtr := Dtr;
        Break;
      end;
    if not Valid then ControlDtr := dtrHandshake;

    Valid := false;
    for Rts := Low(TVaControlRts) to High(TVaControlRts) do
      if (DCB.Flags and DCB_ControlRts[Rts]) > 0 then
      begin
        Valid := True;
        ControlRts := Rts;
        Break;
      end;
    if not Valid then ControlRts := rtsHandshake;

    OutCtsFlow := DCB.Flags and fOutxCtsFlow > 0;
    OutDsrFlow  := DCB.Flags and fOutxDsrFlow > 0;
    XonXoffOut := DCB.Flags and fOutX > 0;
    XonXoffIn := DCB.Flags and fInX > 0;
    DsrSensitivity := DCB.Flags and fDsrSensitivity > 0;
    TxContinueOnXoff := DCB.Flags and fTxContinueOnXoff > 0;
  end;

  EventChars.XonChar := DCB.XOnChar;
  EventChars.XOffChar := DCB.XOffChar;
  EventChars.ErrorChar := DCB.ErrorChar;
  EventChars.EventChar := DCB.EvtChar;
  EventChars.EofChar := DCB.EofChar;

  Buffers.XonLimit := DCB.XOnLim;
  Buffers.XOffLimit := DCB.XOffLim;

  FOptions := [];
  for Option := Low(TVaCommOption) to High(TVaCommOption) do
    if DCB.Flags and DCB_Options[Option] <> 0 then
      FOptions := FOptions + [Option];
end;

procedure TVaCustomComm.ResetPortParameters;
var
  BufSize: Dword;
  CommConfig: TCommConfig;
begin
  FillChar(CommConfig, Sizeof(CommConfig), 0);
  BufSize := Sizeof(TCommConfig);
  if GetDefaultCommConfig(PChar(GetDeviceName), CommConfig, BufSize) then
    ReadDCB(CommConfig.dcb)
  else raise EVaCommError.Create(sErrDefaultConfig);
end;

function TVaCustomComm.GetComStat(Index: Integer): Boolean;
begin
  Result := TComStateFlag(Index - 1) in GetCommStatus.Status.Flags;
end;

function TVaCustomComm.GetModemStatus(Index: Integer): Boolean;
const
  Mask: array[1..4] of dword =
    (MS_CTS_ON, MS_DSR_ON, MS_RING_ON, MS_RLSD_ON);
var
  Status: dword;
begin
  Status := 0;
  if Active then
    Watch(GetCommModemStatus(Handle, Status), sErrorGetModemStatus);
  Result := Status and Mask[Index] <> 0;
end;

procedure TVaCustomComm.SetBuffers(Value: TVaBuffers);
begin
  CheckOpen;
  FBuffers.Assign(Value);
end;

procedure TVaCustomComm.SetEventChars(Value: TVaEventChars);
begin
  FEventChars.Assign(Value);
end;

procedure TVaCustomComm.SetFlowControl(Value: TVaFlowControl);
begin
  FFlowControl.Assign(Value);
end;

procedure TVaCustomComm.BuffersChanged(Sender: TObject);
begin
  CheckOpen;
  UpdateBuffers;
  SetDCB;
end;

procedure TVaCustomComm.EventCharsChanged(Sender: TObject);
begin
  SetDCB;
end;

procedure TVaCustomComm.FlowControlChanged(Sender: TObject);
begin
  SetDCB;
end;

procedure TVaCustomComm.SetDirectWrite(Value: Boolean);
begin
  CheckOpen;
  FDirectWrite := Value;
end;

procedure TVaCustomComm.SetUpdateDCB(Value: Boolean);
begin
  CheckOpen;
  FUpdateDCB := Value;
end;

procedure TVaCustomComm.SetDeviceName(const Value: string);
begin
  CheckOpen;
  FDeviceName := Value;
end;

procedure TVaCustomComm.SetUserBaudrate(Value: Integer);
begin
  if FUserBaudrate <> Value then
  begin
    FUserBaudrate := Value;
    FBaudrate := brUser;
    SetDCB;
  end;
end;

procedure TVaCustomComm.SetBaudrate(Value: TVaBaudrate);
begin
  if FBaudrate <> Value then
  begin
    FBaudrate := Value;
    if FBaudrate <> brUser then
      FUserBaudrate := 0;
    SetDCB;
  end;
end;

procedure TVaCustomComm.SetParity(Value: TVaParity);
begin
  if FParity <> Value then
  begin
    FParity := Value;
    SetDCB;
  end;
end;

procedure TVaCustomComm.SetDatabits(Value: TVaDatabits);
begin
  if FDatabits <> Value then
  begin
    FDatabits := Value;
    SetDCB;
  end;
end;

procedure TVaCustomComm.SetStopbits(Value: TVaStopbits);
begin
  if FStopbits <> Value then
  begin
    FStopbits := Value;
    SetDCB;
  end;
end;

procedure TVaCustomComm.SetOptions(Value: TVaCommOptions);
begin
  if FOptions <> Value then
  begin
    FOptions := Value;
    SetDCB;
  end;
end;

procedure TVaCustomComm.SetPortNum(Value: TVaPortInt);
begin
  CheckOpen;
  FPortNum := Value;
end;

procedure TVaCustomComm.SetMonitorEvents(Value: TVaCommEvents);
begin
  CheckOpen;
  FMonitorEvents := Value;
end;

procedure TVaCustomComm.SetEventPriority(Value: TThreadPriority);
begin
  if FEventPriority <> Value then
  begin
    FEventPriority := Value;
    if FEventThread <> nil then
      FEventThread.Priority := FEventPriority;
  end;
end;

procedure TVaCustomComm.SetWritePriority(Value: TThreadPriority);
begin
  if FWritePriority <> Value then
  begin
    FWritePriority := Value;
    if FWriteThread <> nil then
      FWriteThread.Priority := FWritePriority;
  end;
end;

procedure TVaCustomComm.InsertClient(Value: TVaDispatcher);
begin
  Value.Comm := Self;
  FClients.Add(Value);
end;

procedure TVaCustomComm.RemoveClient(Value: TVaDispatcher);
var
  I: Integer;
begin
  I := FClients.IndexOf(Value);
  if I <> -1 then
  begin
    Value.Comm := nil;
    FClients.Delete(I);
  end;
end;

procedure TVaCustomComm.DestroyClients;
begin
  while FClients.Count > 0 do
    RemoveClient(TVaDispatcher(FClients.Last));
end;

procedure TVaCustomComm.EnableControls;
begin
  FControlsEnabled := True;
end;

procedure TVaCustomComm.DisableControls;
begin
  FControlsEnabled := false;
end;

procedure TVaCustomComm.WriteEvents(Events: dword);
var
  I: Integer;
begin
  if FControlsEnabled then
  begin
    for I := 0 to FClients.Count - 1 do
      TVaDispatcher(FClients[I]).LineEvent(Events);
  end;
end;

procedure TVaCustomComm.WriteBuffer(Data: PVaData; Count: Integer);
var
  I: Integer;
begin
  if FControlsEnabled then
  begin
    for I := 0 to FClients.Count - 1 do
      TVaDispatcher(FClients[I]).DataEvent(Data, Count);
  end;
  if Assigned(FOnRxBuf) then FOnRxBuf(Self, Data, Count);
end;

procedure TVaCustomComm.UpdateDisplay;
begin
  if Assigned(FDisplay) then
    FDisplay.UpdateDisplay(Self);
end;

procedure TVaCustomComm.SetDisplay(const Value: TVaDisplay);
begin
  FDisplay := Value;
  if Value <> nil then
    Value.FreeNotification(Self);
end;

procedure TVaCustomComm.Notification(AComponent: TComponent; Operation: TOperation);
begin
  if (AComponent = FDisplay) and (Operation = opRemove) then
    FDisplay := nil;
  inherited;
end;

procedure TVaCustomComm.GetComPortNames(Value: TStringList);
var
  Reg : TRegistry;
  SerPtSL : TStringList;
  i : integer;
  CommStr : string;
const
  CommPNPKey : string = '\Enum\BIOS\*PNP0501';
  HardwareKey : string = '\hardware\devicemap\serialcomm';
begin
  Value.Clear;

  // stringlist to hold key or value names during search
  SerPtSL := TStringList.Create;

  Reg := TRegistry.Create;
  Reg.Access := KEY_READ;	
  with Reg do
  begin
    RootKey := HKEY_LOCAL_MACHINE;
    // check PNP entries
    if OpenKey(CommPNPKey, false) then
    begin
      // get all serial port keys - one key for each interupt used
      GetKeyNames(SerPtSL);
      //get the Comm names for all the keys - into CommSL
      for i := 0 to SerPtSL.Count - 1 do
      begin
        OpenKey(CommPNPKey + '\' + SerPtSL.Strings[i], false);
        if GetDataType('PortName') = rdString then
        begin
          Value.Add(ReadString('PortName'));
        end;
      end;
    end;
    SerPtSL.Clear; // to use for hardware value names

    // check the hardware entries}
    if OpenKey(HardwareKey, false) then
    begin
      // get the value names for the commports - NT is "Serialn" W95 is "COMn"
      GetValueNames(SerPtSL);
      // now get the data value for each commport
      for i := 0 to SerPtSL.Count - 1 do
        if GetDataType(SerPtSL.Strings[i]) = rdString then
        begin
          CommStr := ReadString(SerPtSL.Strings[i]);
          // if its not in CommNames already ...
          if Value.IndexOf(CommStr) < 0 then // ... add it
            Value.Add(CommStr);
        end;
    end;
    Free;
  end;
  SerPtSL.Free;
end;

function TVaCustomComm.GetVersion: string;
var
  vn: Integer;
begin
  vn := GetVersionNr;
  Result := IntToStr(Hi(Hiword(vn))) + '.' + IntToStr(Lo(Hiword(vn))) + '.' + IntToStr(Hi(Loword(vn))) + '.' + IntToStr(Lo(Loword(vn)));
end;

function TVaCustomComm.GetVersionNr: Integer;
begin
  Result := MakeLong(MakeWord(BLD_VER, REL_VER), MakeWord(MIN_VER, MAJ_VER));
end;

procedure TVaCustomComm.SetVersion(const Value: string);
begin

end;

{ TVaDispatcher }

constructor TVaDispatcher.Create(AOwner: IVaCommUser);
begin
  FOwner := AOwner;
  FActive := True;
end;

destructor TVaDispatcher.Destroy;
begin
  if Comm <> nil then
    Comm.FClients.Remove(Self);
  inherited Destroy;
end;

procedure TVaDispatcher.SetDTRState(Value: Boolean);
begin
  if Comm <> nil then Comm.SetDTR(Value);
end;

procedure TVaDispatcher.PurgeRead;
begin
  if Comm <> nil then Comm.PurgeRead;
end;

procedure TVaDispatcher.PurgeWrite;
begin
  if Comm <> nil then Comm.PurgeWrite;
end;

procedure TVaDispatcher.PurgeReadWrite;
begin
  if Comm <> nil then Comm.PurgeReadWrite;
end;

function TVaDispatcher.WriteBuf(var Buf; Count: Integer): Integer;
begin
  Result := 0;
  if Comm <> nil then Result := Comm.WriteBuf(Buf, Count);
end;

function TVaDispatcher.WriteText(const s: string): Boolean;
begin
  Result := false;
  if Comm <> nil then Result := Comm.WriteText(s);
end;

function TVaDispatcher.WriteChar(Ch: Char): Boolean;
begin
  Result := false;
  if Comm <> nil then Result := Comm.WriteChar(Ch);
end;

procedure TVaDispatcher.DataEvent(Data: PVaData; Count: Integer);
begin
  if (Active) then FOwner.DataChanged(Data, Count);
end;

procedure TVaDispatcher.LineEvent(Events: dword);
begin
  if (Active) then FOwner.LineChanged(Events);
end;

{ TVaCommComponent }

constructor TVaCommComponent.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  FDispatcher := TVaDispatcher.Create(Self);
end;

destructor TVaCommComponent.Destroy;
begin
  FDispatcher.Free;
  inherited Destroy;
end;

procedure TVaCommComponent.SetComm(Value: TVaComm);
begin
  if FComm <> nil then
    FComm.RemoveClient(FDispatcher);
  FComm := Value;
  if FComm <> nil then
    FComm.InsertClient(FDispatcher);
end;

procedure TVaCommComponent.Notification(AComponent: TComponent;
  Operation: TOperation);
begin
  inherited Notification(AComponent, Operation);
  if (Operation = opRemove) then
    if AComponent = Comm then Comm := nil;
end;

procedure TVaCommComponent.SetActive(Value: Boolean);
begin
  FDispatcher.Active := Value;
end;

function TVaCommComponent.GetActive: Boolean;
begin
  Result := FDispatcher.Active;
end;

procedure TVaCommComponent.DataChanged(Data: PVaData; Count: Integer);
begin
end;

procedure TVaCommComponent.LineChanged(Events: dword);
begin
end;

{ TWaitThread }

constructor TWaitThread.Create(AComm: TVaCustomComm);
begin
  FComm := AComm;
  inherited Create(false);
end;

procedure TWaitThread.Execute;
begin
  FreeOnTerminate := true;
  Done := false;
  repeat
  until (FComm.ReadBufUsed > 0) or (Done);
end;

initialization

{$IFDEF FREEWARE}
  if (FindWindow('TApplication', nil) = 0) or
{    (FindWindow('TAlignPalette', nil) = 0) or
    (FindWindow('TPropertyInspector', nil) = 0) or }
    (FindWindow('TAppBuilder', nil) = 0) then
  begin
    ShowMessage('TMS Async32 evaluation version');
  end
{$ENDIF}


end.
