{***************************************************************************}
{ TMS Async32                                                               }
{ for Delphi 4.0,5.0,6.0,7.0 & C++Builder 4.0,5.0,6.0                       }
{ v1.1                                                                      }
{ Copyright 1996 - 2004 by TMS Software                                     }
{ Email : info@tmssoftware.com                                              }
{ Web : http://www.tmssoftware.com                                          }
{                                                                           }
{ The source code is given as is. The author is not responsible             }
{ for any possible damage done due to the use of this code.                 }
{ The component can be freely used in any application. The complete         }
{ source code remains property of the author and may not be distributed,    }
{ published, given or sold in any form as such. No parts of the source      }
{ code can be included in any other component or application without        }
{ written authorization of the author.                                      }
{***************************************************************************}

unit VaControls;

{$I VALIB.INC}

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Forms, Controls,
  ExtCtrls;

const
  { Error messages }
  sValueOutOfRange = 'Value out of range.';

type
  EVrException = class(Exception);

  TVrVersion = string;

  TVrDrawStyle = (dsOwnerDraw, dsNormal);

  TVrTransparentMode = (tmPixel, tmColor);

  TVrTextAlignment = (vtaLeft, vtaCenter, vtaRight,
    vtaTopLeft, vtaTop, vtaTopRight,
    vtaBottomLeft, vtaBottom, vtaBottomRight);

  TVrAlignment = (vaLeftJustify, vaRightJustify, vaCenter);
  TVrOrientation = (voVertical, voHorizontal);
  TVrImageTextLayout = (ImageLeft, ImageRight, ImageTop, ImageBottom);

const
  cfTextAlign: array[TVrTextAlignment] of Integer =
    (DT_LEFT + DT_VCENTER, DT_CENTER + DT_VCENTER, DT_RIGHT + DT_VCENTER,
     DT_TOP + DT_LEFT, DT_TOP + DT_CENTER, DT_TOP + DT_RIGHT,
     DT_BOTTOM + DT_LEFT, DT_BOTTOM + DT_CENTER, DT_BOTTOM + DT_RIGHT);


type
  TGraphicControlCanvas = class(TGraphicControl)
  public
    property Canvas;
  end;

  TCustomControlCanvas = class(TCustomControl)
  public
    property Canvas;
  end;

  TVaGraphicControl = class(TGraphicControl)
  private
    FUpdateCount: Integer;
    FOnMouseEnter: TNotifyEvent;
    FOnMouseLeave: TNotifyEvent;
    FOnFontChanged: TNotifyEvent;
    FOnTextChanged: TNotifyEvent;
    procedure CMMouseEnter(var Message: TMessage); message CM_MOUSEENTER;
    procedure CMMouseLeave(var Message: TMessage); message CM_MOUSELEAVE;
    procedure CMFontChanged(var Message: TMessage); message CM_FONTCHANGED;
    procedure CMTextChanged(var Message: TMessage); message CM_TEXTCHANGED;
  protected
    function Designing: Boolean;
    function Loading: Boolean;
    procedure ClearClientCanvas;
    procedure UpdateControlCanvas; virtual;
    procedure UpdateControlBounds; virtual;
    procedure AdjustControlSize; virtual;
    procedure ShowDesignFrame(Dest: TCanvas);
    procedure MouseEnter; virtual;
    procedure MouseLeave; virtual;
    procedure FontChanged; virtual;
    procedure TextChanged; virtual;
    property OnMouseEnter: TNotifyEvent read FOnMouseEnter write FOnMouseEnter;
    property OnMouseLeave: TNotifyEvent read FOnMouseLeave write FOnMouseLeave;
    property OnFontChanged: TNotifyEvent read FOnFontChanged write FOnFontChanged;
    property OnTextChanged: TNotifyEvent read FOnTextChanged write FOnTextChanged;
  public
    constructor Create(AOwner: TComponent); override;
    procedure BeginUpdate;
    procedure EndUpdate;
  published
  end;

implementation

function CompareRect(R1, R2: TRect): Boolean;
begin
  Result := (R1.Left = R2.Left) and
            (R1.Top = R2.Top) and
            (R1.Right = R2.Right) and
            (R1.Bottom = R2.Bottom);
end;


{ TVaGraphicControl }

constructor TVaGraphicControl.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
end;

function TVaGraphicControl.Designing: Boolean;
begin
  Result := (csDesigning in ComponentState);
end;

function TVaGraphicControl.Loading: Boolean;
begin
  Result := (csLoading in ComponentState);
end;

procedure TVaGraphicControl.AdjustControlSize;
var
  R: TRect;
begin
  R := BoundsRect;
  if not Loading then
  begin
    SetBounds(Left, Top, Width, Height);
    if CompareRect(BoundsRect, R) then
      UpdateControlCanvas; //force update
  end;
end;

{ Used in combination with AutoSize property }
procedure TVaGraphicControl.UpdateControlBounds;
var
  R: TRect;
begin
  if (AutoSize) and (Align in [alNone, alTop, alBottom, alLeft, alRight]) then
  begin
    R := BoundsRect;
    if not Loading then
      SetBounds(Left, Top, Width, Height);
    if CompareRect(BoundsRect, R) then
      UpdateControlCanvas; //force update
  end else UpdateControlCanvas;
end;

procedure TVaGraphicControl.ClearClientCanvas;
begin
  with inherited Canvas do
  begin
    Brush.Style := bsSolid;
    Brush.Color := Self.Color;
    FillRect(ClientRect);
  end;
end;

procedure TVaGraphicControl.ShowDesignFrame(Dest: TCanvas);
begin
  if Designing then
    with Dest do
    begin
      Pen.Style := psDot;
      Pen.Color := clBlack;
      Brush.Style := bsClear;
      Rectangle(0, 0, Width, Height);
    end;
end;

procedure TVaGraphicControl.UpdateControlCanvas;
begin
  if not Loading then
    if FUpdateCount = 0 then Repaint;
end;

procedure TVaGraphicControl.BeginUpdate;
begin
  Inc(FUpdateCount);
end;

procedure TVaGraphicControl.EndUpdate;
begin
  Dec(FUpdateCount);
  UpdateControlCanvas;
end;

procedure TVaGraphicControl.FontChanged;
begin
  if Assigned(FOnFontChanged) then
    FOnFontChanged(Self);
end;

procedure TVaGraphicControl.TextChanged;
begin
  if Assigned(FOnTextChanged) then
    FOnTextChanged(Self);
end;

procedure TVaGraphicControl.MouseEnter;
begin
  if Assigned(FOnMouseEnter) then
    FOnMouseEnter(Self);
end;

procedure TVaGraphicControl.MouseLeave;
begin
  if Assigned(FOnMouseLeave) then
    FOnMouseLeave(Self);
end;

procedure TVaGraphicControl.CMMouseEnter(var Message: TMessage);
begin
  inherited;
  MouseEnter;
end;

procedure TVaGraphicControl.CMMouseLeave(var Message: TMessage);
begin
  inherited;
  MouseLeave;
end;

procedure TVaGraphicControl.CMFontChanged(var Message: TMessage);
begin
  inherited;
  FontChanged;
end;

procedure TVaGraphicControl.CMTextChanged(var Message: TMessage);
begin
  inherited;
  TextChanged;
end;

initialization

end.
