{***************************************************************************}
{ TMS Async32                                                               }
{ for Delphi 4.0,5.0,6.0,7.0,2005 & C++Builder 4.0,5.0,6.0                  }
{ v1.3                                                                      }
{ Copyright 1996 - 2004 by TMS Software                                     }
{ Email : info@tmssoftware.com                                              }
{ Web : http://www.tmssoftware.com                                          }
{                                                                           }
{ The source code is given as is. The author is not responsible             }
{ for any possible damage done due to the use of this code.                 }
{ The component can be freely used in any application. The complete         }
{ source code remains property of the author and may not be distributed,    }
{ published, given or sold in any form as such. No parts of the source      }
{ code can be included in any other component or application without        }
{ written authorization of the author.                                      }
{***************************************************************************}

unit VaDisplay;

{$I VALIB.INC}

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  VaControls;

type
  TVrDisplayType = (ltRounded, ltRectangle);
  TVrDisplayMode = (ltHide, ltScale);
  
  TVaDisplay = class(TVaGraphicControl)
  public
    procedure UpdateDisplay(AOwner: TComponent); virtual; abstract;
  end;

  TVaLEDDisplay = class(TVaDisplay)
  private
    FShowCTS: Boolean;
    FShowDSR: Boolean;
    FShowXOffHold: Boolean;
    FShowRLSD: Boolean;
    FShowRLSDHold: Boolean;
    FShowXOffSent: Boolean;
    FShowRING: Boolean;
    FShowCTSHold: Boolean;
    FShowDSRHold: Boolean;
    FOnColor: TColor;
    FOffColor: TColor;
    FDisplayType: TVrDisplayType;
    FDisplayMode: TVrDisplayMode;
    FDisplaycaption: Boolean;
    FLEDImage: TBitmap;
    procedure SetOffColor(const Value: TColor);
    procedure SetOnColor(const Value: TColor);
    procedure SetShowCTS(const Value: Boolean);
    procedure SetShowCTSHold(const Value: Boolean);
    procedure SetShowDSR(const Value: Boolean);
    procedure SetShowDSRHold(const Value: Boolean);
    procedure SetShowRING(const Value: Boolean);
    procedure SetShowRLSD(const Value: Boolean);
    procedure SetShowRLSDHold(const Value: Boolean);
    procedure SetShowXOffHold(const Value: Boolean);
    procedure SetShowXOffSent(const Value: Boolean);
    procedure SetDisplayType(const Value: TVrDisplayType);
    procedure SetDisplayMode(const Value: TVrDisplayMode);
    procedure SetDisplaycaption(const Value: Boolean);
  protected
    LastCTS: Boolean;
    LastDSR: Boolean;
    LastRING: Boolean;
    LastRLSD: Boolean;
    LastCTSHold: Boolean;
    LastDSRHold: Boolean;
    LastRLSDHold: Boolean;
    LastXOffHold: Boolean;
    LastXOffSent: Boolean;
    visibleLeds: integer;
    procedure DrawALed(LedLeft, Ledwidth: integer; color: TColor; Caption: string);
    procedure DoALed(var LedLeft: integer; Ledwidth: integer;
      Active: Boolean; Caption: string);
    procedure DoAllLed(var LedLeft: integer; Ledwidth: integer);
    function GetLedWidth: integer;
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    procedure UpdateDisplay(AOwner: TComponent); override;
    procedure Paint; override;
    procedure SetBounds(ALeft, ATop, AWidth, AHeight: Integer); override;
  published
    property Color;
    property ShowCTS: Boolean read FShowCTS write SetShowCTS default True;
    property ShowDSR: Boolean read FShowDSR write SetShowDSR default True;
    property ShowRING: Boolean read FShowRING write SetShowRING default True;
    property ShowRLSD: Boolean read FShowRLSD write SetShowRLSD default True;
    property ShowCTSHold: Boolean read FShowCTSHold write SetShowCTSHold default True;
    property ShowDSRHold: Boolean read FShowDSRHold write SetShowDSRHold default True;
    property ShowRLSDHold: Boolean read FShowRLSDHold write SetShowRLSDHold default True;
    property ShowXOffHold: Boolean read FShowXOffHold write SetShowXOffHold default True;
    property ShowXOffSent: Boolean read FShowXOffSent write SetShowXOffSent default True;
    property OnColor: TColor read FOnColor write SetOnColor default clBlue;
    property OffColor: TColor read FOffColor write SetOffColor default clNavy;
    property DisplayType: TVrDisplayType read FDisplayType write SetDisplayType
      default ltRounded;
    property DisplayMode: TVrDisplayMode read FDisplayMode write SetDisplayMode
      default ltScale;
    property Displaycaption: Boolean read FDisplaycaption write SetDisplaycaption
      default True;
    property Font;


  end;

implementation

uses VaComm;
{ TVaLEDDisplay }

constructor TVaLEDDisplay.Create(AOwner: TComponent);
begin
  inherited;
  Font.Color := clWhite;
  FDisplayType := ltRounded;
  FDisplayMode := ltScale;
  FDisplaycaption := True;
  FShowCTS := True;
  FShowDSR := True;
  FShowXOffHold := True;
  FShowRLSD := True;
  FShowRLSDHold := True;
  FShowXOffSent := True;
  FShowRING := True;
  FShowCTSHold := True;
  FShowDSRHold := True;
  FOnColor := clBlue;
  FOffColor := clNavy;
  LastCTS := False;
  LastDSR := False;
  LastRING := False;
  LastRLSD := False;
  LastCTSHold := False;
  LastDSRHold := False;
  LastRLSDHold := False;
  LastXOffHold := False;
  LastXOffSent := False;
  FLEDImage := TBitmap.Create;
  SetBounds(Left, Top, 316, 18);
end;

procedure TVaLEDDisplay.DoALed(var LedLeft: Integer; Ledwidth: integer;
  Active: Boolean; Caption: string);
var
  color: TColor;
begin
  if Active then
    Color := FOnColor
  else
    Color := FOffColor;

  DrawALed(LedLeft, Ledwidth, color, Caption);
  Inc(LedLeft, Ledwidth);
end;

procedure TVaLEDDisplay.DrawALed(LedLeft, Ledwidth: Integer; Color: TColor;
  Caption: string);
var
  r: Trect;
  DrawStyle: integer;
begin
  FLEDImage.Canvas.Pen.Color := ClBlack;
  FLEDImage.Canvas.Brush.Color := Color;

  R := Rect(LedLeft + 1, 0, LedLeft + Ledwidth - 1, Height);
  if FDisplayType = ltRectangle then
    FLEDImage.Canvas.Rectangle(r.Left,r.Top,r.Right,r.Bottom)
  else
    FLEDImage.Canvas.Ellipse(r.Left,r.Top,r.Right,r.Bottom);

  if FDisplayCaption then
  begin
    with FLEDImage.Canvas do
    begin
      Brush.Style := bsClear;
      Font.Assign(Self.Font);
      if FDisplayType = ltRectangle then DrawStyle :=
          DT_SINGLELine + DT_NOPREFIX + DT_CENTER + DT_TOP
      else
        DrawStyle := DT_SINGLELine + DT_NOPREFIX + DT_CENTER + DT_VCENTER;
      DrawText(Handle, PChar(Caption), - 1, r, DrawStyle);
    end;
  end;
end;

procedure TVaLEDDisplay.DoAllLed(var LedLeft: integer; Ledwidth: integer);
begin
  LedLeft := 0;

  FLEDImage.Canvas.Brush.Color := Color;
  FLEDImage.Canvas.Pen.Color := Color;

  FLEDImage.Canvas.Rectangle(0,0,Width,Height);

  if FShowCTS then DoALed(LedLeft, Ledwidth, LastCTS, 'CTS');
  if FShowDSR then DoALed(LedLeft, Ledwidth, LastDSR, 'DSR');
  if FShowRING then DoALed(LedLeft, Ledwidth, LastRING, 'RING');
  if FShowRLSD then DoALed(LedLeft, Ledwidth, LastRLSD, 'RLSD');
  if FShowCTSHold then DoALed(LedLeft, Ledwidth, LastCTSHold, 'CTSH');
  if FShowDSRHold then DoALed(LedLeft, Ledwidth, LastDSRHold, 'DSRH');
  if FShowRLSDHold then DoALed(LedLeft, Ledwidth, LastRLSDHold, 'LSDH');
  if FShowXOffHold then DoALed(LedLeft, Ledwidth, LastXOffHold, 'XH');
  if FShowXOffSent then DoALed(LedLeft, Ledwidth, LastXOffSent, 'XS');
end;


function TVaLEDDisplay.GetLedWidth: Integer;
begin
  visibleLeds := 0;
  if FShowCTS then Inc(visibleLeds);
  if FShowDSR then Inc(visibleLeds);
  if FShowXOffHold then Inc(visibleLeds);
  if FShowRLSD then Inc(visibleLeds);
  if FShowRLSDHold then Inc(visibleLeds);
  if FShowXOffSent then Inc(visibleLeds);
  if FShowRING then Inc(visibleLeds);
  if FShowCTSHold then Inc(visibleLeds);
  if FShowDSRHold then Inc(visibleLeds);
  if visibleLeds = 0 then Result := 0
  else
    Result := Trunc(Width / visibleLeds);
end;

procedure TVaLEDDisplay.Paint;
var
  LedWidth, LedLeft: integer;
begin
  inherited;
  LedWidth := GetLedWidth;
  if not (FDisplayMode = ltscale) then
    LedWidth := Trunc(Width / 9);
  DoAllLed(LedLeft, LedWidth);

  Canvas.Draw(0,0,FLEDImage);
end;

procedure TVaLEDDisplay.SetDisplayType(const Value: TVrDisplayType);
begin
  if Value <> FDisplayType then
  begin
    FDisplayType := Value;
    UpdateControlCanvas;
  end;
end;

procedure TVaLEDDisplay.SetOffColor(const Value: TColor);
begin
  if Value <> FOffColor then
  begin
    FOffColor := Value;
    UpdateControlCanvas;
  end;
end;

procedure TVaLEDDisplay.SetOnColor(const Value: TColor);
begin
  if Value <> FOnColor then
  begin
    FOnColor := Value;
    UpdateControlCanvas;
  end;
end;

procedure TVaLEDDisplay.SetShowCTS(const Value: Boolean);
begin
  if Value <> FShowCTS then
  begin
    FShowCTS := Value;
    UpdateControlCanvas;
  end;
end;

procedure TVaLEDDisplay.SetShowCTSHold(const Value: Boolean);
begin
  if Value <> FShowCTSHold then
  begin
    FShowCTSHold := Value;
    UpdateControlCanvas;
  end;
end;

procedure TVaLEDDisplay.SetShowDSR(const Value: Boolean);
begin
  if Value <> FShowDSR then
  begin
    FShowDSR := Value;
    UpdateControlCanvas;
  end;
end;

procedure TVaLEDDisplay.SetShowDSRHold(const Value: Boolean);
begin
  if Value <> FShowDSRHold then
  begin
    FShowDSRHold := Value;
    UpdateControlCanvas;
  end;
end;

procedure TVaLEDDisplay.SetDisplayMode(const Value: TVrDisplayMode);
begin
  if Value <> FDisplayMode then
  begin
    FDisplayMode := Value;
    UpdateControlCanvas;
  end;
end;

procedure TVaLEDDisplay.SetShowRING(const Value: Boolean);
begin
  if Value <> FShowRING then
  begin
    FShowRING := Value;
    UpdateControlCanvas;
  end;
end;

procedure TVaLEDDisplay.SetShowRLSD(const Value: Boolean);
begin
  if Value <> FShowRLSD then
  begin
    FShowRLSD := Value;
    UpdateControlCanvas;
  end;
end;

procedure TVaLEDDisplay.SetShowRLSDHold(const Value: Boolean);
begin
  if Value <> FShowRLSDHold then
  begin
    FShowRLSDHold := Value;
    UpdateControlCanvas;
  end;
end;

procedure TVaLEDDisplay.SetShowXOffHold(const Value: Boolean);
begin
  if Value <> FShowXOffHold then
  begin
    FShowXOffHold := Value;
    UpdateControlCanvas;
  end;
end;

procedure TVaLEDDisplay.SetShowXOffSent(const Value: Boolean);
begin
  if Value <> FShowXOffSent then
  begin
    FShowXOffSent := Value;
    UpdateControlCanvas;
  end;
end;

procedure TVaLEDDisplay.UpdateDisplay(AOwner: TComponent);
var
  theComm: TVaCustomComm;
begin
  if not (AOwner is TVaCustomComm) then
  begin
    Beep;
    Exit;
  end;
  
  theComm := AOwner as TVaCustomComm;
  LastCTS := theComm.CTS;
  LastDSR := theComm.DSR;
  LastRING := theComm.RING;
  LastRLSD := theComm.RLSD;
  LastCTSHold := theComm.CTSHold;
  LastDSRHold := theComm.DSRHold;
  LastRLSDHold := theComm.RLSDHold;
  LastXOffHold := theComm.XOffHold;
  LastXOffSent := theComm.XOffSent;

  UpdateControlCanvas;
end;

procedure TVaLEDDisplay.SetDisplayCaption(const Value: Boolean);
begin
  if FDisplaycaption <> Value then
  begin
    FDisplaycaption := Value;
    UpdateControlCanvas;
  end;
end;

destructor TVaLEDDisplay.Destroy;
begin
  FLEDImage.Free;
  inherited;
end;

procedure TVaLEDDisplay.SetBounds(ALeft, ATop, AWidth, AHeight: Integer);
begin
  inherited;
  FLEDImage.Width := AWidth;
  FLEDImage.Height := AHeight;
end;

end.