{***************************************************************************}
{ TMS Async32                                                               }
{ for Delphi 4.0,5.0,6.0 & C++Builder 4.0,5.0,6.0                           }
{                                                                           }
{ Copyright 1996 - 2002 by TMS Software                                     }
{ Email : info@tmssoftware.com                                              }
{ Web : http://www.tmssoftware.com                                          }
{                                                                           }
{ The source code is given as is. The author is not responsible             }
{ for any possible damage done due to the use of this code.                 }
{ The component can be freely used in any application. The complete         }
{ source code remains property of the author and may not be distributed,    }
{ published, given or sold in any form as such. No parts of the source      }
{ code can be included in any other component or application without        }
{ written authorization of the author.                                      }
{***************************************************************************}

unit VaModem;

{$I VALIB.INC}

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  VaTypes, VaClasses, VaComm, VaUtils, VaTriggers;

type
  EVaModemException = class(Exception);
  TVaRingDetect = (rdLine, rdMsg, rdNone);
  TVaRingEvent = procedure(Sender: TObject; Rings: Integer; var AcceptCall: Boolean) of object;
  TVaConnectEvent = procedure(Sender: TObject; const ConnectString: string) of object;
  TVaModemAction = (maNone, maInit, maReset, maDial, maAnswer, maHangup);

  TVaCustomModem = class(TVaCommComponent)
  private
    FOK: string;
    FRing: string;
    FBusy: string;
    FVoice: string;
    FNoCarrier: string;
    FNoDialTone: string;
    FError: string;
    FConnect: string;
    FConfig: TStrings;
    FDialTimeout: Integer;
    FAnswerTimeout: Integer;
    FCommandTimeout: Integer;
    FCharDelay: Integer;
    FConnectString: string;
    FRingDetect: TVaRingDetect;
    FRingCount: Integer;
    FRingTimer: TVaTimer;
    FActionTimer: TVaTimer;
    FModemAction: TVaModemAction;
    FDtrDropDelay: Integer;
    FOnRingDetect: TVaRingEvent;
    FOnRingWaitTimeout: TNotifyEvent;
    FOnOK: TNotifyEvent;
    FOnBusy: TNotifyEvent;
    FOnVoice: TNotifyEvent;
    FOnNoCarrier: TNotifyEvent;
    FOnNoDialTone: TNotifyEvent;
    FOnError: TNotifyEvent;
    FOnConnect: TVaConnectEvent;
    FOnAnswerTimeout: TNotifyEvent;
    FOnDialTimeout: TNotifyEvent;
    FOnCommandTimeout: TNotifyEvent;
    function GetRingWaitTimeout: Integer;
    procedure SetRingWaitTimeout(Value: Integer);
    procedure SetConfig(Value: TStrings);
    procedure ConfigChanged(Sender: TObject);
  protected
    Resp: string;
    TermChar: Char;
    procedure ReceiveChar(Ch: Char);
    procedure HandleRingEvent; virtual;
    procedure RingWaitTimeoutEvent(Sender: TObject);
    procedure ActionWaitTimeoutEvent(Sender: TObject);
    procedure CheckWaiting;
    procedure CheckResponse;
    procedure InitConfig;
    function ConfigValue(Tag: string): string;
    procedure LineChanged(Events: dword); override;
    procedure DataChanged(Data: PVaData; Count: Integer); override;
    property Config: TStrings read FConfig write SetConfig;
    property DialTimeout: Integer read FDialTimeout write FDialTimeout default 30000;
    property AnswerTimeout: Integer read FAnswerTimeout write FAnswerTimeout default 30000;
    property RingWaitTimeout: Integer read GetRingWaitTimeout write SetRingWaitTimeout default 5000;
    property CommandTimeout: Integer read FCommandTimeout write FCommandTimeout default 5000;
    property CharDelay: Integer read FCharDelay write FCharDelay default 0;
    property RingDetect: TVaRingDetect read FRingDetect write FRingDetect default rdLine;
    property DtrDropDelay: Integer read FDtrDropDelay write FDtrDropDelay default 1000;
    property OnRingDetect: TVaRingEvent read FOnRingDetect write FOnRingDetect;
    property OnRingWaitTimeout: TNotifyEvent read FOnRingWaitTimeout write FOnRingWaitTimeout;
    property OnOK: TNotifyEvent read FOnOK write FOnOK;
    property OnBusy: TNotifyEvent read FOnBusy write FOnBusy;
    property OnVoice: TNotifyEvent read FOnVoice write FOnVoice;
    property OnNoCarrier: TNotifyEvent read FOnNoCarrier write FOnNoCarrier;
    property OnNoDialTone: TNotifyEvent read FOnNoDialTone write FOnNoDialTone;
    property OnError: TNotifyEvent read FOnError write FOnError;
    property OnConnect: TVaConnectEvent read FOnConnect write FOnConnect;
    property OnAnswerTimeout: TNotifyEvent read FOnAnswerTimeout write FOnAnswerTimeout;
    property OnDialTimeout: TNotifyEvent read FOnDialTimeout write FOnDialTimeout;
    property OnCommandTimeout: TNotifyEvent read FOnCommandTimeout write FOnCommandTimeout;
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    procedure Init;
    procedure Reset;
    procedure Dial(PhoneNumber: string);
    procedure Answer;
    procedure Cancel;
    procedure Hangup(DropDTR: Boolean);
    procedure PutCommand(Cmd: string);
    procedure WaitForResponse(Delay: Integer);
    property ConnectString: string read FConnectString;
    property ModemAction: TVaModemAction read FModemAction;
  end;

  TVaModem = class(TVaCustomModem)
  published
    property Active;
    property Config;
    property DialTimeout;
    property AnswerTimeout;
    property RingWaitTimeout;
    property CommandTimeout;
    property CharDelay;
    property RingDetect;
    property DtrDropDelay;
    property OnRingDetect;
    property OnRingWaitTimeout;
    property OnOK;
    property OnBusy;
    property OnVoice;
    property OnNoCarrier;
    property OnNoDialTone;
    property OnError;
    property OnConnect;
    property OnAnswerTimeout;
    property OnDialTimeout;
    property OnCommandTimeout;
  end;


implementation


const
  sErrorWrite = 'TVaModem: error writing to comm.';
  sWaitingOp = 'Waiting for previous operation to complete.';

  RESETCMD    = 'RESETCMD';
  DIALCMD     = 'DIALCMD';
  DIALTERM    = 'DIALTERM';
  HANGUPCMD   = 'HANGUPCMD';
  INITCMD     = 'INITCMD';
  ANSWERCMD   = 'ANSWERCMD';
  ROK         = 'ROK';
  RCONNECT    = 'RCONNECT';
  RBUSY       = 'RBUSY';
  RVOICE      = 'RVOICE';
  RNOCARRIER  = 'RNOCARRIER';
  RNODIALTONE = 'RNODIALTONE';
  RERROR      = 'RERROR';
  RRING       = 'RRING';
  RTERM       = 'RTERM';


constructor TVaCustomModem.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  FConfig := TStringList.Create;
  TStringList(FConfig).OnChange := ConfigChanged;
  InitConfig;
  FDialTimeout := 30000;
  FAnswerTimeout := 30000;
  FCommandTimeout := 5000;
  FCharDelay := 0;
  FRingDetect := rdLine;
  FModemAction := maNone;
  FDtrDropDelay := 1000;
  FRingTimer := TVaTimer.Create(nil);
  FRingTimer.Enabled := false;
  FRingTimer.Interval := 5000; //5 seconds
  FRingTimer.OnTimer := RingWaitTimeoutEvent;
  FActionTimer := TVaTimer.Create(nil);
  FActionTimer.Enabled := false;
  FActionTimer.Interval := 5000; //5 seconds
  FActionTimer.OnTimer := ActionWaitTimeoutEvent;
end;

destructor TVaCustomModem.Destroy;
begin
  FConfig.Free;
  FRingTimer.Free;
  FActionTimer.Free;
  inherited Destroy;
end;

procedure TVaCustomModem.PutCommand(Cmd: string);
var
  I: Integer;
begin
  Cmd := strCtrl(Cmd);
  for I := 1 to Length(Cmd) do
  begin
    if not Dispatcher.WriteChar(Cmd[I]) then
      raise EVaModemException.Create(sErrorWrite);
    if CharDelay > 0 then
      SysDelay(CharDelay, False);
  end;
end;

procedure TVaCustomModem.ActionWaitTimeoutEvent(Sender: TObject);
begin
  FActionTimer.Enabled := false;
  case FModemAction of
    maDial:
      if Assigned(FOnDialTimeout) then
        FOnDialTimeout(Self);
    maAnswer:
      begin
        FRingCount := 0;
        if Assigned(FOnAnswerTimeout) then
          FOnAnswerTimeout(Self);
      end;
    else
      if Assigned(FOnCommandTimeout) then
        FOnCommandTimeout(Self);
  end;
end;

procedure TVaCustomModem.RingWaitTimeoutEvent(Sender: TObject);
begin
  FRingTimer.Enabled := false;
  FRingCount := 0;
  if Assigned(FOnRingWaitTimeout) then
    FOnRingWaitTimeout(Self);
end;

procedure TVaCustomModem.HandleRingEvent;
var
  AcceptCall: Boolean;
begin
  FRingTimer.Enabled := false;
  Inc(FRingCount);
  AcceptCall := false;
  if Assigned(FOnRingDetect) then
    FOnRingDetect(Self, FRingCount, AcceptCall);
  if AcceptCall then Answer
  else FRingTimer.Enabled := True;
end;

function TVaCustomModem.GetRingWaitTimeout: Integer;
begin
  Result := FRingTimer.Interval;
end;

procedure TVaCustomModem.SetRingWaitTimeout(Value: Integer);
begin
  FRingTimer.Interval := Value;
end;

procedure TVaCustomModem.CheckWaiting;
begin
  if FActionTimer.Enabled then
    raise EVaModemException.Create(sWaitingOp);
end;

procedure TVaCustomModem.WaitForResponse(Delay: Integer);
begin
  FActionTimer.Interval := Delay;
  FActionTimer.Enabled := True;
end;

procedure TVaCustomModem.Cancel;
begin
  FActionTimer.Enabled := false;
  FModemAction := maNone;
end;

procedure TVaCustomModem.CheckResponse;
var
  Index: Integer;
begin
  Index := -1;
  Resp := Trim(Resp);
  if Resp = FOk then Index := 1
  else if Resp = FRing then Index := 2
  else if Resp = FBusy then Index := 3
  else if Resp = FVoice then Index := 4
  else if Resp = FNoCarrier then Index := 5
  else if Resp = FNoDialTone then Index := 6
  else if Resp = FError then Index := 7
  else
    begin
      if Copy(Resp, 1, Length(FConnect)) = FConnect then
      begin
        Index := 8;
        FConnectString := Resp;
      end;
    end;

  if Index <> -1 then
    FActionTimer.Enabled := false;

  case Index of
    1: if Assigned(FOnOK) then FOnOK(Self);
    2: begin
         if RingDetect = rdMsg then
           HandleRingEvent;
       end;
    3: if Assigned(FOnBusy) then FOnBusy(Self);
    4: if Assigned(FOnVoice) then FOnVoice(Self);
    5: if Assigned(FOnNoCarrier) then FOnNoCarrier(Self);
    6: if Assigned(FOnNoDialTone) then FOnNoDialTone(Self);
    7: if Assigned(FOnError) then FOnError(Self);
    8: if Assigned(FOnConnect) then FOnConnect(Self, FConnectString);
   end; //case
end;

procedure TVaCustomModem.ReceiveChar(Ch: Char);
begin
  if Ch = TermChar then
  begin
    if Resp <> '' then
      CheckResponse;
    Resp := '';
  end else
  Resp := Resp + CN(Ch);
end;

procedure TVaCustomModem.Init;
begin
  CheckWaiting;
  FModemAction := maInit;
  Dispatcher.PurgeReadWrite;
  PutCommand(ConfigValue(INITCMD));
  WaitForResponse(CommandTimeout);
end;

procedure TVaCustomModem.Reset;
begin
  CheckWaiting;
  FModemAction := maReset;
  Dispatcher.PurgeReadWrite;
  PutCommand(ConfigValue(RESETCMD));
  WaitForResponse(CommandTimeout);
end;

procedure TVaCustomModem.Dial(PhoneNumber: string);
var
  Cmd: string;
begin
  CheckWaiting;
  FModemAction := maDial;
  Cmd := ConfigValue(DIALCMD) + PhoneNumber + ConfigValue(DIALTERM);
  Dispatcher.PurgeReadWrite;
  PutCommand(Cmd);
  WaitForResponse(DialTimeout);
end;

procedure TVaCustomModem.Answer;
begin
  CheckWaiting;
  FModemAction := maAnswer;
  Dispatcher.PurgeReadWrite;
  PutCommand(ConfigValue(ANSWERCMD));
  WaitForResponse(AnswerTimeout);
end;

procedure TVaCustomModem.Hangup(DropDTR: Boolean);
begin
  CheckWaiting;
  FModemAction := maHangup;
  Dispatcher.PurgeReadWrite;
  if DropDTR then
  begin
    Dispatcher.SetDTRState(False);
    SysDelay(DtrDropDelay, True);
    Dispatcher.SetDTRState(True);
  end else
  begin
    PutCommand(ConfigValue(HANGUPCMD));
    WaitForResponse(CommandTimeout);
  end;
end;

procedure TVaCustomModem.SetConfig(Value: TStrings);
begin
  FConfig.Assign(Value);
end;

procedure TVaCustomModem.ConfigChanged(Sender: TObject);
var
  Tmp: string;
begin
  FOk := ConfigValue(ROK);
  FRing := ConfigValue(RRING);
  FBusy := ConfigValue(RBUSY);
  FVoice := ConfigValue(RVOICE);
  FNoCarrier := ConfigValue(RNOCARRIER);
  FNoDialTone := ConfigValue(RNODIALTONE);
  FError := ConfigValue(RERROR);
  FConnect := ConfigValue(RCONNECT);
  Tmp := strCtrl(ConfigValue(RTERM));
  if Tmp <> '' then
    TermChar := Tmp[1] else TermChar := #0;
end;

function TVaCustomModem.ConfigValue(Tag: string): string;
begin
  Result := Trim(FConfig.Values[Tag]);
end;

procedure TVaCustomModem.InitConfig;
begin
  FConfig.BeginUpdate;
  try
    FConfig.Clear;
    FConfig.Add(RESETCMD+'=ATZ^M');
    FConfig.Add(DIALCMD+'=ATDT');
    FConfig.Add(DIALTERM+'=^M');
    FConfig.Add(HANGUPCMD+'=+++~~~ATH0^M');
    FConfig.Add(INITCMD+'=ATZE1M1^M');
    FConfig.Add(ANSWERCMD+'=ATA^M');
    FConfig.Add(ROK+'=OK');
    FConfig.Add(RCONNECT+'=CONNECT');
    FConfig.Add(RBUSY+'=BUSY');
    FConfig.Add(RVOICE+'=VOICE');
    FConfig.Add(RNOCARRIER+'=NO CARRIER');
    FConfig.Add(RNODIALTONE+'=NO DIALTONE');
    FConfig.Add(RERROR+'=ERROR');
    FConfig.Add(RRING+'=RING');
    FConfig.Add(RTERM+'=^M');
  finally
    FConfig.EndUpdate;
  end;
end;

procedure TVaCustomModem.LineChanged(Events: dword);
begin
  if (EV_RING and Events > 0) then
    if RingDetect = rdLine then HandleRingEvent;
end;

procedure TVaCustomModem.DataChanged(Data: PVaData; Count: Integer);
var
  I: Integer;
begin
  for I := 0 to Count - 1 do
    ReceiveChar(Data^[I]);
end;



end.
