{***************************************************************************}
{ TMS Async32                                                               }
{ for Delphi 4.0,5.0,6.0,7.0,2005 & C++Builder 4.0,5.0,6.0                  }
{                                                                           }
{ Copyright 1996 - 2004 by TMS Software                                     }
{ Email : info@tmssoftware.com                                              }
{ Web : http://www.tmssoftware.com                                          }
{                                                                           }
{ The source code is given as is. The author is not responsible             }
{ for any possible damage done due to the use of this code.                 }
{ The component can be freely used in any application. The complete         }
{ source code remains property of the author and may not be distributed,    }
{ published, given or sold in any form as such. No parts of the source      }
{ code can be included in any other component or application without        }
{ written authorization of the author.                                      }
{***************************************************************************}

unit VaObjects;

{$I VALIB.INC}

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs;


type
  TVaHandleObject = class(TObject)
  private
    FHandle: THandle;
    FError: Integer;
  public
    destructor Destroy; override;
    property Handle: THandle read FHandle;
    property Error: Integer read FError;
  end;

  TVaEventWaitResult = (wrSignaled, wrTimeout, wrAbandoned, wrError);

  TVaEvent = class(TVaHandleObject)
  public
    constructor Create(EventAttributes: PSecurityAttributes; ManualReset,
      InitialState: Boolean; const Name: string);
    function WaitFor(Timeout: dWord): TVaEventWaitResult;
    procedure SetEvent;
    procedure ResetEvent;
  end;

  TVaWaitEvent = class(TVaEvent)
  public
    constructor Create;
  end;

  TVaCriticalSection = class(TObject)
  private
    FSection: TRTLCriticalSection;
  public
    constructor Create;
    destructor Destroy; override;
    procedure Acquire;
    procedure Release;
    procedure Enter;
    procedure Leave;
  end;

implementation


{ TVaHandleObject }

destructor TVaHandleObject.Destroy;
begin
  CloseHandle(FHandle);
  inherited Destroy;
end;

{ TVaEvent }

constructor TVaEvent.Create(EventAttributes: PSecurityAttributes; ManualReset,
  InitialState: Boolean; const Name: string);
begin
  FHandle := CreateEvent(EventAttributes, ManualReset, InitialState, PChar(Name));
end;

function TVaEvent.WaitFor(Timeout: dWord): TVaEventWaitResult;
begin
  case WaitForSingleObject(Handle, Timeout) of
    WAIT_ABANDONED:
      Result := wrAbandoned;
    WAIT_OBJECT_0:
      Result := wrSignaled;
    WAIT_TIMEOUT:
      Result := wrTimeout;
    WAIT_FAILED:
      begin
        Result := wrError;
        FError := GetLastError;
      end;
  else
    Result := wrError;
  end;
end;

procedure TVaEvent.SetEvent;
begin
  Windows.SetEvent(Handle);
end;

procedure TVaEvent.ResetEvent;
begin
  Windows.ResetEvent(Handle);
end;

{ TVaWaitEvent }

constructor TVaWaitEvent.Create;
begin
  FHandle := CreateEvent(nil, True, False, nil);
end;

{ TVaCriticalSection }

constructor TVaCriticalSection.Create;
begin
  inherited Create;
  InitializeCriticalSection(FSection);
end;

destructor TVaCriticalSection.Destroy;
begin
  DeleteCriticalSection(FSection);
  inherited Destroy;
end;

procedure TVaCriticalSection.Acquire;
begin
  EnterCriticalSection(FSection);
end;

procedure TVaCriticalSection.Release;
begin
  LeaveCriticalSection(FSection);
end;

procedure TVaCriticalSection.Enter;
begin
  Acquire;
end;

procedure TVaCriticalSection.Leave;
begin
  Release;
end;


end.
