{***************************************************************************}
{ TMS Async32                                                               }
{ for Delphi 4.0,5.0,6.0 & C++Builder 4.0,5.0,6.0                           }
{ v1.1                                                                      }
{ Copyright 1996 - 2002 by TMS Software                                     }
{ Email : info@tmssoftware.com                                              }
{ Web : http://www.tmssoftware.com                                          }
{                                                                           }
{ The source code is given as is. The author is not responsible             }
{ for any possible damage done due to the use of this code.                 }
{ The component can be freely used in any application. The complete         }
{ source code remains property of the author and may not be distributed,    }
{ published, given or sold in any form as such. No parts of the source      }
{ code can be included in any other component or application without        }
{ written authorization of the author.                                      }
{***************************************************************************}

unit VaProtocol;

{$I VALIB.INC}

interface

uses
  Windows, Messages, SysUtils, Classes, VaClasses, Forms,
  VaComm, VaObjects;

const
  MAX_RXSIZE = 4096;
  MAX_PACKET = 1024;

  ErrorNone           = 0;
  ErrorTimeout        = -1;
//ErrorWriteChar      = -2;
  ErrorLocalCan       = -3;
  ErrorRemoteCan      = -4;
  ErrorOutOfSync      = -5;
//ErrorChecksum       = -6;
  ErrorStreamRead     = -7;
  ErrorStreamWrite    = -8;
  ErrorMaxErrors      = -9;
  ErrorRemoteStartup  = -10;
  ErrorFileCreate     = -11;
  ErrorFileOpen       = -12;
  ErrorBufferOverrun  = -14;
  ErrorInvalidPacket  = -15;
  ErrorInvalidChar    = -16;

  ErrorStreamSeek     = -17;
//ErrorRemoteAbort    = -18;
  ErrorUnknownHeader  = -19;
  ErrorRepeatHeader   = -20;
  ErrorRemoteFileErr  = -21;
  ErrorProgramExit    = -22;
  ErrorPacketToLarge  = -23;
  ErrorRemoteShutdown = -24;
  ErrorZCommand       = -25;

  RX0 = $00;
  SOH = $01;
  STX = $02;
  EOT = $04;
  ACK = $06;
  NAK = $15;
  CAN = $18;
  XYN = NAK;
  XYC = Ord('C');
  XYG = Ord('G');

type
  EProtAbort = Exception;

  TVaProtocol = class;
  TVaTransferMode = (tmUpload, tmDownload);

  TVaTransferEndEvent = procedure(Sender: TObject; ExitCode: Integer) of object;
  TVaTransferErrorEvent = procedure(Sender: TObject; ErrorCode, ErrorCount: Integer) of object;
  TVaPacketEvent = procedure(Sender: TObject; Packet, ByteCount: Integer) of object;
  TVaFileInfoEvent = procedure(Sender: TObject; const Name: string; Size: Integer) of object;

  TVaProtocolThread = class(TThread)
  private
    FBuffer: TVaCommBuffer;
    FStream: TFileStream;
    FErrorCode: Integer;
    FErrorCount: Integer;
    FAbortTransfer: Boolean;
    FBufferOverrun: Boolean;
    FCriticalSection: TVaCriticalSection;
    FTimeEvent: TVaWaitEvent;
    FOwner: TVaProtocol;
  protected
    procedure Lock;
    procedure Unlock;
    function ReadChar: Integer; virtual;
    function PeekChar: Integer;
    function CharReceived: Boolean;
    procedure PutBuf(var Buf; Count: Integer);
    procedure WriteBuf(var Buf; Count: Integer);
    procedure WriteChar(aByte: Byte);
    procedure Upload; virtual; abstract;
    procedure Download; virtual; abstract;
    procedure Initialize; virtual;
    procedure OpenStream(FileName: string);
    procedure CreateStream(FileName: string);
    procedure CloseStream;
    function EndOfStream: Boolean;
    procedure SetExitCode(Value: Integer);
    procedure RemoteCan; virtual;
    procedure DoError;
    procedure ErrorEvent(ErrorCode: Integer; RCan: Boolean);
    property Stream: TFileStream read FStream write FStream;
    property Owner: TVaProtocol read FOwner;
    property AbortTransfer: Boolean read FAbortTransfer;
  public
    constructor Create(AOwner: TVaProtocol);
    destructor Destroy; override;
    procedure Cancel;
    procedure Execute; override;
    procedure Terminate;
    property ErrorCode: Integer read FErrorCode;
    property ErrorCount: Integer read FErrorCount;
  end;

  TVaProtocol = class(TVaCommComponent)
  private
    FThread: TVaProtocolThread;
    FTimeout: Integer;
    FMode: TVaTransferMode;
    FExitCode: Integer;
    FMaxErrors: Integer;
    FBufferSize: Integer;
    FOnTransferStart: TNotifyEvent;
    FOnTransferEnd: TVaTransferEndEvent;
    FOnError: TVaTransferErrorEvent;
    function GetErrorCode: Integer;
    function GetErrorCount: Integer;
    function GetIsActive: Boolean;
    procedure CheckActive;
    procedure Terminate(Sender: TObject);
  protected
    function CreateThread: TVaProtocolThread; virtual; abstract;
    procedure DataChanged(Data: PVaData; Count: Integer); override;
    procedure TransferStart; virtual;
    procedure TransferEnd; virtual;
    procedure ErrorEvent;
    procedure SetExitCode(Value: Integer);
    procedure Initialize; virtual;
    property ExitCode: Integer read FExitCode;
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    procedure Execute; virtual;
    procedure Cancel; virtual;
    property ErrorCode: Integer read GetErrorCode;
    property ErrorCount: Integer read GetErrorCount;
    property IsActive: Boolean read GetIsActive;
  published
    property Mode: TVaTransferMode read FMode write FMode;
    property Timeout: Integer read FTimeout write FTimeout default 60000;
    property MaxErrors: Integer read FMaxErrors write FMaxErrors default 10;
    property BufferSize: Integer read FBufferSize write FBufferSize default 4096;
    property OnTransferStart: TNotifyEvent read FOnTransferStart write FOnTransferStart;
    property OnTransferEnd: TVaTransferEndEvent read FOnTransferEnd write FOnTransferEnd;
    property OnError: TVaTransferErrorEvent read FOnError write FOnError;
  end;

  TVaPacket = array[0..MAX_PACKET - 1] of Byte;

  TVaXYModemThread = class(TVaProtocolThread)
  private
    FPacket: TVaPacket;
    FPacketNum: Integer;
    FByteCount: Integer;
    FFileName: string;
    FFileSize: Integer;
    procedure FileInfo(Name: string; Size: Integer);
    procedure PacketEvent(Packet, Bytes: Integer);
  protected
    procedure RemoteCan; override;
    procedure RemoteTxStartup(var NCGByte: Byte);
    procedure RemoteRxStartup(NCGByte: Byte);
    procedure ResetPacket;
    procedure ReadPacket(PacketNum: Integer; var PacketSize: Integer; NCGByte: Byte);
    procedure ReadFile(Name: string; Size: Integer; NCGByte: Byte);
    procedure WriteEOT;
    procedure WritePacket(PacketNum, PacketSize: Integer; NCGByte: Byte);
    procedure WriteFile(Name: string);
    property PacketNum: Integer read FPacketNum;
    property ByteCount: Integer read FByteCount;
    property FileName: string read FFileName;
    property FileSize: Integer read FFileSize;
  end;

  TVaXYModem = class(TVaProtocol)
  private
    FOnFileInfo: TVaFileInfoEvent;
    FOnPacketEvent: TVaPacketEvent;
  protected
    procedure FileInfo;
    procedure PacketEvent;
  published
    property OnFileInfo: TVaFileInfoEvent read FOnFileInfo write FOnFileInfo;
    property OnPacketEvent: TVaPacketEvent read FOnPacketEvent write FOnPacketEvent;
  end;

  TVaXModemThread = class(TVaXYModemThread)
  protected
    procedure Upload; override;
    procedure Download; override;
  end;

  TVaXProtocol = (XModem, XModem1K, XModemG);
  TVaXModem = class(TVaXYModem)
  private
    FFileName: string;
    FProtocol: TVaXProtocol;
  protected
    function CreateThread: TVaProtocolThread; override;
  public
    constructor Create(AOwner: TComponent); override;
  published
    property FileName: string read FFileName write FFileName;
    property Protocol: TVaXProtocol read FProtocol write FProtocol;
  end;

  TVaYModemThread = class(TVaXYModemThread)
  private
    function GetFiles: TStrings;
  protected
    procedure Upload; override;
    procedure Download; override;
    property Files: TStrings read GetFiles;
  end;

  TVaYProtocol = (YModem, YModemG);
  TVaYModem = class(TVaXYModem)
  private
    FFiles: TStrings;
    FTargetFolder: string;
    FProtocol: TVaYProtocol;
    procedure SetFiles(Value: TStrings);
  protected
    function CreateThread: TVaProtocolThread; override;
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
  published
    property Files: TStrings read FFiles write SetFiles;
    property Protocol: TVaYProtocol read FProtocol write FProtocol;
    property TargetFolder: string read FTargetFolder write FTargetFolder;
  end;


implementation

uses
  VaUtils, VaCRC32;

const
  NCGX: array[TVaXProtocol] of Byte = (XYN, XYC, XYG);
  NCGY: array[TVaYProtocol] of Byte = (XYC, XYG);


{ TVaProtocolThread }

constructor TVaProtocolThread.Create(AOwner: TVaProtocol);
begin
  FOwner := AOwner;
  FreeOnTerminate := True;
  FCriticalSection := TVaCriticalSection.Create;
  FBuffer := TVaCommBuffer.Create(FOwner.BufferSize);
  FBufferOverrun := false;
  FTimeEvent := TVaWaitEvent.Create;
  inherited Create(True);
  Priority := tpHigher;
end;

destructor TVaProtocolThread.Destroy;
begin
  FBuffer.Free;
  FTimeEvent.Free;
  FCriticalSection.Free;
  CloseStream;
  inherited Destroy;
end;

procedure TVaProtocolThread.Lock;
begin
  FCriticalSection.Enter;
end;

procedure TVaProtocolThread.Unlock;
begin
  FCriticalSection.Leave;
end;

procedure TVaProtocolThread.PutBuf(var Buf; Count: Integer);
var
  NewSize: Integer;
begin
  if not FBufferOverrun then
  begin
    Lock;
    try
      NewSize := FBuffer.Size + Count;
      if NewSize > MAX_RXSIZE then
        FBufferOverrun := True
      else FBuffer.Write(Buf, Count);
      FTimeEvent.SetEvent;
    finally
      Unlock;
    end;
  end;
end;

function TVaProtocolThread.ReadChar: Integer;
var
  B: Byte;
begin
  if (FBuffer.Size = 0) and (not FAbortTransfer) then
  begin
    repeat
      FTimeEvent.ResetEvent;
      if FTimeEvent.WaitFor(FOwner.Timeout) <> wrSignaled then
        ErrorEvent(ErrorTimeout, False);
    until (FBuffer.Size > 0) or (FAbortTransfer) or (FBufferOverrun);
  end;

  if (FBufferOverrun) then
  begin
    RemoteCan;
    SetExitCode(ErrorBufferOverrun);
  end;

  if FAbortTransfer then
    Result := RX0
  else
  begin
    Lock;
    try
      FBuffer.Read(B, 1);
      Result := B;
    finally
      Unlock;
    end;
  end;
end;

function TVaProtocolThread.PeekChar: Integer;
var
  B: Byte;
begin
  if FBuffer.Size = 0 then
  begin
    Result := -1;
    Exit;
  end;
  FBuffer.Peek(B, 1);
  Result := B;
end;

function TVaProtocolThread.CharReceived: Boolean;
begin
  Result := FBuffer.Size > 0;
end;

procedure TVaProtocolThread.WriteBuf(var Buf; Count: Integer);
begin
  if not Terminated then
    with FOwner do Dispatcher.WriteBuf(Buf, Count);
end;

procedure TVaProtocolThread.WriteChar(aByte: Byte);
begin
  WriteBuf(aByte, 1);
end;

procedure TVaProtocolThread.Initialize;
begin
  FBuffer.Clear;
  FErrorCount := 0;
  FErrorCode := ErrorNone;
  FAbortTransfer := false;
end;

procedure TVaProtocolThread.OpenStream(FileName: string);
begin
  try
    FStream := TFileStream.Create(FileName, fmOpenReadWrite);
  except
    RemoteCan;
    SetExitCode(ErrorFileOpen);
  end;
end;

procedure TVaProtocolThread.CreateStream(FileName: string);
begin
  try
    FStream := TFileStream.Create(FileName, fmCreate);
  except
    RemoteCan;
    SetExitCode(ErrorFileCreate);
  end;
end;

procedure TVaProtocolThread.CloseStream;
begin
  if FStream <> nil then
  begin
    FStream.Free;
    FStream := nil;
  end;
end;

function TVaProtocolThread.EndOfStream: Boolean;
begin
  Result := FStream.Position = FStream.Size;
end;

procedure TVaProtocolThread.SetExitCode(Value: Integer);
begin
  FOwner.SetExitCode(Value);
  Abort;
//raise EProtAbort.Create('Async32: Protocol Abort.');
end;

procedure TVaProtocolThread.RemoteCan;
begin
end;

procedure TVaProtocolThread.DoError;
begin
  if not Terminated then
    FOwner.ErrorEvent;
end;

procedure TVaProtocolThread.ErrorEvent(ErrorCode: Integer;
  RCan: Boolean);
begin
  Inc(FErrorCount);
  FErrorCode := ErrorCode;
  if FErrorCount = FOwner.MaxErrors then
  begin
    if RCan then RemoteCan;
    SetExitCode(ErrorMaxErrors);
  end;
  Synchronize(DoError);
end;

procedure TVaProtocolThread.Cancel;
begin
  FAbortTransfer := True;
  FTimeEvent.SetEvent;
end;

procedure TVaProtocolThread.Execute;
begin
  repeat
    try
      Initialize;
      if FOwner.Mode = tmUpload then
        Upload else Download;
    except
      Terminate;
    end;
  until Terminated;
end;

procedure TVaProtocolThread.Terminate;
begin
  inherited;
  FAbortTransfer := True;
  FTimeEvent.SetEvent;
end;

{ TVaProtocol }

constructor TVaProtocol.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  Active := false;
  FMode := tmUpload;
  FTimeout := 60000;
  FMaxErrors := 10;
  FBufferSize := 4096;
end;

destructor TVaProtocol.Destroy;
begin
  if FThread <> nil then
  begin
    FThread.Terminate;
{$IFNDEF VER140}
    FThread.WaitFor;
{$ENDIF}
  end;
  inherited Destroy;
end;

function TVaProtocol.GetErrorCode: Integer;
begin
  Result := ErrorNone;
  if FThread <> nil then
    Result := FThread.ErrorCode;
end;

function TVaProtocol.GetErrorCount: Integer;
begin
  Result := 0;
  if FThread <> nil then
    Result := FThread.ErrorCount;
end;

procedure TVaProtocol.DataChanged(Data: PVaData; Count: Integer);
begin
  if Count > 0 then
    if FThread <> nil then
      FThread.PutBuf(Data^, Count);
end;

procedure TVaProtocol.SetExitCode(Value: Integer);
begin
  FExitCode := Value;
end;

procedure TVaProtocol.CheckActive;
begin
  if FThread <> nil then
    raise EProtAbort.Create('Transfer already in progress...');
end;

function TVaProtocol.GetIsActive: Boolean;
begin
  Result := FThread <> nil;
end;

procedure TVaProtocol.TransferStart;
begin
  try
    if Assigned(FOnTransferStart) then
      FOnTransferStart(Self);
  except
    Application.HandleException(Self);
  end;
end;

procedure TVaProtocol.TransferEnd;
begin
  try
    if Assigned(FOnTransferEnd) then
      FOnTransferEnd(Self, ExitCode);
  except
    Application.HandleException(Self);
  end;
end;

procedure TVaProtocol.ErrorEvent;
begin
  try
    if Assigned(FOnError) then
      FOnError(Self, ErrorCode, ErrorCount);
  except
    Application.HandleException(Self);
  end;
end;

procedure TVaProtocol.Initialize;
begin
  FExitCode := ErrorNone;
  Dispatcher.PurgeReadWrite;
  Active := True;
end;

procedure TVaProtocol.Cancel;
begin
  if FThread <> nil then
    FThread.Cancel;
end;

procedure TVaProtocol.Execute;
begin
  CheckActive;
  Initialize;
  FThread := CreateThread;
  FThread.OnTerminate := Terminate;
  TransferStart;
  FThread.Resume;
end;

procedure TVaProtocol.Terminate(Sender: TObject);
begin
  Active := false;
  if not (csDestroying in ComponentState) then
  begin
    FThread := nil;
    TransferEnd;
  end;
end;

{ TVaXYModemThread }

procedure TVaXYModemThread.FileInfo(Name: string; Size: Integer);
begin
  FFileName := Name;
  FFileSize := Size;
  Synchronize(TVaXYModem(FOwner).FileInfo);
end;

procedure TVaXYModemThread.PacketEvent(Packet, Bytes: Integer);
begin
  FPacketNum := Packet;
  FByteCount := Bytes;
  Synchronize(TVaXYModem(FOwner).PacketEvent);
end;

procedure TVaXYModemThread.ResetPacket;
begin
  FillChar(FPacket, Sizeof(FPacket), 0);
end;

procedure TVaXYModemThread.ReadPacket(PacketNum: Integer;
  var PacketSize: Integer; NCGByte: Byte);
var
  PacketOK: Boolean;
  Checksum, RxChecksum: Word;
  I, Code, RxPacketNum: Integer;
begin
  ResetPacket;
  PacketSize := 0;
  repeat
    Code := ReadChar;
    case Code of
      RX0:; //abort transfer code
      SOH: PacketSize := 128;
      STX: PacketSize := 1024;
      CAN: SetExitCode(ErrorRemoteCan);
      EOT:
        begin
          WriteChar(ACK);
          SetExitCode(ErrorNone);
        end;
      else
        SetExitCode(ErrorOutOfSync);
    end;

    RxPacketNum := ReadChar and $00ff; //packetnum
    ReadChar; //complement

    Checksum := 0;
    for I := 0 to PacketSize - 1 do
    begin
      FPacket[I] := ReadChar;
      if NCGByte = NAK then
        Checksum := (Checksum + FPacket[I]) and $00ff
      else Checksum := CalcCRC16(FPacket[I], Checksum);
    end;

    RxChecksum := ReadChar and $00ff;
    if NCGByte <> NAK then
    begin
      Code := ReadChar and $00ff;
      RxChecksum := (RxChecksum shl 8) or Code;
    end;

    if FAbortTransfer then
    begin
      RemoteCan;
      SetExitCode(ErrorLocalCan);
    end;

    if PacketNum and $00ff <> RxPacketNum then
    begin
      RemoteCan;
      SetExitCode(ErrorOutOfSync);
    end;

    PacketOK := Checksum = RxChecksum;
    if not PacketOK then
    begin
      ErrorEvent(ErrorInvalidPacket, True);
      WriteChar(NAK);
    end;

  until PacketOK;
end;

procedure TVaXYModemThread.ReadFile(Name: string; Size: Integer; NCGByte: Byte);
var
  Ready: Boolean;
  BytesLeft: Integer;
  PacketNum, PacketSize: Integer;
begin
  PacketNum := 0;

  if FileExists(Name) then
    Name := CreateUniqueFileName(Name);
  CreateStream(Name);

  try
    FileInfo(Name, Size);

    WriteChar(NCGByte);
    Ready := false;
    repeat
      Inc(PacketNum);
      ReadPacket(PacketNum, PacketSize, NCGByte);

      BytesLeft := Size - Stream.Position;
      if (Size = 0) or (BytesLeft > PacketSize) then
        BytesLeft := PacketSize;

      try
        Stream.Write(FPacket, BytesLeft);
      except
        RemoteCan;
        SetExitCode(ErrorStreamWrite);
      end;

      PacketEvent(PacketNum, Stream.Position);

      if NCGByte <> XYG then WriteChar(ACK);
    until Ready;
  finally
    CloseStream;
  end;
end;

procedure TVaXYModemThread.RemoteCan;
begin
  WriteChar(CAN);
  WriteChar(CAN);
end;

procedure TVaXYModemThread.RemoteTxStartup(var NCGByte: Byte);
var
  I: Integer;
begin
  for I := 1 to 10 do
  begin
    NCGByte := ReadChar;
    case NCGByte of
      RX0:
        begin
          RemoteCan;
          SetExitCode(ErrorLocalCan);
        end;
      XYN, XYC, XYG: Exit;
      else
        ErrorEvent(ErrorInvalidChar, false);
    end;
  end;
  SetExitCode(ErrorOutOfSync);
end;

procedure TVaXYModemThread.RemoteRxStartup(NCGByte: Byte);
var
  I: Integer;
  TimeEvent: TVaTimeEvent;
begin
  for I := 1 to 10 do
  begin
    WriteChar(NCGByte);
    if not CharReceived then
    begin
      InitTimer(TimeEvent, 10000); //10 seconds wait
      repeat
        if TimerExpired(TimeEvent) then
        begin
          ErrorEvent(ErrorTimeout, False);
          Break;
        end;
      until (CharReceived) or (AbortTransfer);
    end;
    if CharReceived then
      Exit;
  end;
  SetExitCode(ErrorRemoteStartup);
end;


procedure TVaXYModemThread.WriteEOT;
var
  I: Integer;
  Code: Integer;
begin
  for I := 1 to 10 do
  begin
    WriteChar(EOT);
    Code := ReadChar;
    case Code of
      RX0:
        begin
          RemoteCan;
          SetExitCode(ErrorLocalCan);
        end;
      ACK: Exit;
      else
        ErrorEvent(ErrorInvalidChar, false);
    end;
  end;
  SetExitCode(ErrorOutOfSync);
end;

procedure TVaXYModemThread.WritePacket(PacketNum, PacketSize: Integer;
  NCGByte: Byte);
var
  I, Code: Integer;
  Checksum: Word;
  PacketOK: Boolean;
begin
  PacketOK := false;
  repeat
    if FAbortTransfer then
    begin
      RemoteCan;
      SetExitCode(ErrorLocalCan);
    end;

    case PacketSize of
      128: WriteChar(SOH);
     1024: WriteChar(STX);
    end;

    Code := PacketNum and $00ff;
    WriteChar(Code);
    WriteChar(255 - Code);

    Checksum := 0;
    for I := 0 to PacketSize - 1 do
    begin
      if NCGByte = NAK then
        Checksum := Checksum + FPacket[I]
      else Checksum := CalcCRC16(FPacket[I], Checksum);
    end;
    WriteBuf(FPacket, PacketSize);

    if NCGByte = NAK then
      WriteChar(Checksum and $00ff)
    else
    begin
      WriteChar((Checksum shr 8) and $00ff);
      WriteChar(Checksum and $00ff);
    end;

    if NCGByte = XYG then
    begin
      if PeekChar = CAN then
        SetExitCode(ErrorRemoteCan);
      Exit;
    end;

    Code := ReadChar;
    case Code of
      RX0:; //abort transfer code
      ACK: PacketOK := True;
      NAK: ErrorEvent(ErrorInvalidPacket, True);
      CAN: SetExitCode(ErrorRemoteCan);
      else
        SetExitCode(ErrorOutOfsync);
    end;
  until PacketOK;
end;

procedure TVaXYModemThread.WriteFile(Name: string);
var
  Ready: Boolean;
  PacketNum, PacketSize: Integer;
  NCGByte: Byte;
begin
  PacketNum := 0;
  OpenStream(Name);
  try
    FileInfo(Name, Stream.Size);

    RemoteTxStartup(NCGByte);

    Ready := false;
    repeat
      Inc(PacketNum);

      if NCGByte = NAK then PacketSize := 128
      else PacketSize := 1024;

      ResetPacket;
      try
        Stream.Read(FPacket, PacketSize);
      except
        RemoteCan;
        SetExitCode(ErrorStreamRead);
      end;

      WritePacket(PacketNum, PacketSize, NCGByte);

      PacketEvent(PacketNum, FStream.Position);

      if EndOfStream then
      begin
        WriteEOT;
        SetExitCode(ErrorNone);
      end;

    until Ready;

  finally
    CloseStream;
  end;
end;

{ TVaXYModem }

procedure TVaXYModem.FileInfo;
begin
  try
    if Assigned(FOnFileInfo) then
      with TVaXYModemThread(FThread) do
        FOnFileInfo(Self, FileName, FileSize);
  except
    Application.HandleException(Self);
  end;
end;

procedure TVaXYModem.PacketEvent;
begin
  try
    if Assigned(FOnPacketEvent) then
      with TVaXYModemThread(FThread) do
        FOnPacketEvent(Self, PacketNum, ByteCount);
  except
    Application.HandleException(Self);
  end;
end;

{ TVaXModemThread }

procedure TVaXModemThread.Upload;
begin
  WriteFile(TVaXModem(Owner).FileName);
end;

procedure TVaXModemThread.Download;
begin
  ReadFile(TVaXModem(Owner).FileName, 0, NCGX[TVaXModem(FOwner).Protocol]);
end;

{ TVaXModem }

constructor TVaXModem.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  FFileName := '';
  FProtocol := XModem;
end;

function TVaXModem.CreateThread: TVaProtocolThread;
begin
  Result := TVaXModemThread.Create(Self);
end;

{ TVaYModemThread }

function TVaYModemThread.GetFiles: TStrings;
begin
  Result := TVaYModem(Owner).Files;
end;

procedure TVaYModemThread.Upload;
var
  Temp: string;
  ShortName: string;
  FileSize: Integer;
  I, J, Index: Integer;
  NCGByte: Byte;
  BatchDone: Boolean;
begin
  for Index := 0 to Files.Count - 1 do
  begin
    RemoteTxStartup(NCGByte);

    ResetPacket;

    ShortName := ExtractFileName(Files[Index]);
    Filesize := GetFileSize(Files[Index]);

    I := 0;
    J := 1;
    while J <= Length(ShortName) do
    begin
      FPacket[I] := Ord(ShortName[J]);
      Inc(I);
      Inc(J);
    end;
    J := 1;
    Inc(I);
    Temp := IntToStr(FileSize);
    while J <= Length(Temp) do
    begin
      FPacket[I] := Ord(Temp[J]);
      Inc(I);
      Inc(J);
    end;
    WritePacket(0, 128, NCGByte); //write filename, filesize
    try
      WriteFile(Files[Index]);
    except
      if FOwner.ExitCode <> ErrorNone then //eot?
        raise;
    end;

    BatchDone := Index = Files.Count - 1;
    if BatchDone then
    begin
      RemoteTxStartup(NCGByte);
      ResetPacket;
      WritePacket(0, 128, NCGByte);
      SetExitCode(ErrorNone);
    end;
  end;
end;

procedure TVaYModemThread.Download;
var
  BatchDone: Boolean;
  FileName, Temp, Folder: string;
  I, FileSize, PacketSize: Integer;
  NCGByte: Byte;
begin
  Files.Clear;
  with TVaYModem(Owner) do
  begin
    Folder := AddPathSlash(TargetFolder);
    NCGByte := NCGY[Protocol];
  end;

  BatchDone := false;
  repeat
    RemoteRxStartup(NCGByte);

    ReadPacket(0, PacketSize, XYC); //read filename, filesize
    if FPacket[0] = 0 then
    begin
      if NCGByte <> XYG then WriteChar(ACK);
      SetExitCode(ErrorNone); //batch complete
    end;

    I := 0;
    FileName := '';
    while (FPacket[I] <> 0) do
    begin
      FileName := FileName + Chr(FPacket[I]);
      Inc(I);
    end;
    FileName := ExtractFileName(FileName);

    Inc(I);
    Temp := '';
    while (FPacket[I] <> 0) and (Chr(FPacket[I]) in ['0'..'9'])  do
    begin
      Temp := Temp + Chr(FPacket[I]);
      Inc(I);
    end;
    FileSize := StrToIntDef(Temp, MAXINT);

    if NCGByte <> XYG then WriteChar(ACK);

    try
      ReadFile(Folder + FileName, FileSize, NCGByte);
    except
      if FOwner.ExitCode <> ErrorNone then
        raise;
      Files.Add(Folder + FileName);
    end;
  until BatchDone;
end;

{ TVaYModem }

constructor TVaYModem.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  FProtocol := YModem;
  FFiles := TStringList.Create;
end;

destructor TVaYModem.Destroy;
begin
  FFiles.Free;
  inherited Destroy;
end;

procedure TVaYModem.SetFiles(Value: TStrings);
begin
  FFiles.Assign(Value);
end;

function TVaYModem.CreateThread: TVaProtocolThread;
begin
  Result := TVaYModemThread.Create(Self);
end;


end.
