{***************************************************************************}
{ TMS Async32                                                               }
{ for Delphi 4.0,5.0,6.0,7.0,2005 & C++Builder 4.0,5.0,6.0                  }
{                                                                           }
{ Copyright 1996 - 2004 by TMS Software                                     }
{ Email : info@tmssoftware.com                                              }
{ Web : http://www.tmssoftware.com                                          }
{                                                                           }
{ The source code is given as is. The author is not responsible             }
{ for any possible damage done due to the use of this code.                 }
{ The component can be freely used in any application. The complete         }
{ source code remains property of the author and may not be distributed,    }
{ published, given or sold in any form as such. No parts of the source      }
{ code can be included in any other component or application without        }
{ written authorization of the author.                                      }
{***************************************************************************}

unit VaPrst;

{$I VALIB.INC}

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  VaConst, VaTypes, VaClasses, VaObjects, VaUtils;

type
  TVaControlDtr = (dtrDisabled, dtrEnabled, dtrHandshake);
  TVaControlRts = (rtsDisabled, rtsEnabled, rtsHandshake, rtsToggle);

  TVaFlowControl = class(TVaPersistent)
  private
    FOutCtsFlow: Boolean;
    FOutDsrFlow: Boolean;
    FControlDtr: TVaControlDtr;
    FControlRts: TVaControlRts;
    FXonXoffOut: Boolean;
    FXonXoffIn:  Boolean;
    FDsrSensitivity: Boolean;
    FTxContinueOnXoff: Boolean;
    procedure SetOutCtsFlow(Value: Boolean);
    procedure SetOutDsrFlow(Value: Boolean);
    procedure SetControlDtr(Value: TVaControlDtr);
    procedure SetControlRts(Value: TVaControlRts);
    procedure SetXonXoffOut(Value: Boolean);
    procedure SetXonXoffIn(Value: Boolean);
    procedure SetDsrSensitivity(Value: Boolean);
    procedure SetTxContinueOnXoff(Value: Boolean);
  public
    constructor Create;
    procedure Assign(Source: TPersistent); override;
  published
    property OutCtsFlow: Boolean read FOutCtsFlow write SetOutCtsFlow;
    property OutDsrFlow: Boolean read FOutDsrFlow write SetOutDsrFlow;
    property ControlDtr: TVaControlDtr read FControlDtr write SetControlDtr;
    property ControlRts: TVaControlRts read FControlRts write SetControlRts;
    property XonXoffOut: Boolean read FXonXoffOut write SetXonXoffOut;
    property XonXoffIn:  Boolean read FXonXoffIn write SetXonXoffIn;
    property DsrSensitivity: Boolean read FDsrSensitivity write SetDsrSensitivity;
    property TxContinueOnXoff: Boolean read FTxContinueOnXoff write SetTxContinueOnXoff;
  end;

  TVaEventChars = class(TVaPersistent)
  private
    FXonChar: Char;
    FXoffChar: Char;
    FErrorChar: Char;
    FEventChar: Char;
    FEofChar: Char;
    procedure SetEventChar(Index: Integer; Value: Char);
  public
    constructor Create;
    procedure Assign(Source: TPersistent); override;
  published
    property XonChar: Char index 1 read FXOnChar write SetEventChar default #17;
    property XoffChar: Char index 2 read FXOffChar write SetEventChar default #19;
    property ErrorChar: Char index 3 read FErrorChar write SetEventChar default #0;
    property EventChar: Char index 4 read FEventChar write SetEventChar default #0;
    property EofChar: Char index 5 read FEventChar write SetEventChar default #0;
  end;

  TVaBuffers = class(TVaPersistent)
  private
    FReadSize: TVaIntW;
    FWriteSize: TVaIntW;
    FReadTimeout: TVaIntW;
    FWriteTimeout: TVaIntW;
    FXonLimit: TVaIntW;
    FXoffLimit: TVaIntW;
    procedure SetReadSize(Value: TVaIntW);
    procedure SetWriteSize(Value: TVaIntW);
    procedure SetXonLimit(Value: TVaIntW);
    procedure SetXoffLimit(Value: TVaIntW);
  public
    constructor Create;
    procedure Assign(Source: TPersistent); override;
  published
    property ReadSize: TVaIntW read FReadSize write SetReadSize default 4096;
    property WriteSize: TVaIntW read FWriteSize write SetWriteSize default 2048;
    property ReadTimeout: TVaIntW read FReadTimeout write FReadTimeout default 1000;
    property WriteTimeout: TVaIntW read FWriteTimeout write FWriteTimeout default 1000;
    property XonLimit: TVaIntW read FXonLimit write SetXonLimit default 1024;
    property XoffLimit: TVaIntW read FXoffLimit write SetXoffLimit default 1024;
  end;


implementation


{ TVaFlowControl }

constructor TVaFlowControl.Create;
begin
  inherited Create;
  FOutCtsFlow := false;
  FOutDsrFlow := false;
  FControlDtr := dtrDisabled;
  FControlRts := RtsDisabled;
  FXonXoffOut := false;
  FXonXoffIn := false;
  FDsrSensitivity := false;
  FTxContinueOnXoff := false;
end;

procedure TVaFlowControl.Assign(Source: TPersistent);
begin
  if (Source <> nil) and (Source is TVaFlowControl) then
  begin
    BeginUpdate;
    try
      OutCtsFlow := TVaFlowControl(Source).OutCtsFlow;
      OutDsrFlow := TVaFlowControl(Source).OutDsrFlow;
      ControlDtr := TVaFlowControl(Source).ControlDtr;
      ControlRts := TVaFlowControl(Source).ControlRts;
      XonXoffOut := TVaFlowControl(Source).XonXoffOut;
      XonXoffIn := TVaFlowControl(Source).XonXoffIn;
      DsrSensitivity := TVaFlowControl(Source).DsrSensitivity;
      TxContinueOnXoff := TVaFlowControl(Source).TxContinueOnXoff;
    finally
      EndUpdate;
    end;
  end else inherited Assign(Source);
end;

procedure TVaFlowControl.SetOutCtsFlow(Value: Boolean);
begin
  if FOutCtsFlow <> Value then
  begin
    FOutCtsFlow := Value;
    Changed;
  end;
end;

procedure TVaFlowControl.SetOutDsrFlow(Value: Boolean);
begin
  if FOutDsrFlow <> Value then
  begin
    FOutDsrFlow := Value;
    Changed;
  end;
end;

procedure TVaFlowControl.SetControlDtr(Value: TVaControlDtr);
begin
  if FControlDtr <> Value then
  begin
    FControlDtr := Value;
    Changed;
  end;
end;

procedure TVaFlowControl.SetControlRts(Value: TVaControlRts);
begin
  if FControlRts <> Value then
  begin
    FControlRts := Value;
    Changed;
  end;
end;

procedure TVaFlowControl.SetXonXoffOut(Value: Boolean);
begin
  if FXonXoffOut <> Value then
  begin
    FXonXoffOut := Value;
    Changed;
  end;
end;

procedure TVaFlowControl.SetXonXoffIn(Value: Boolean);
begin
  if FXonXoffIn <> Value then
  begin
    FXonXoffIn := Value;
    Changed;
  end;
end;

procedure TVaFlowControl.SetDsrSensitivity(Value: Boolean);
begin
  if FDsrSensitivity <> Value then
  begin
    FDsrSensitivity := Value;
    Changed;
  end;
end;

procedure TVaFlowControl.SetTxContinueOnXoff(Value: Boolean);
begin
  if FTxContinueOnXoff <> Value then
  begin
    FTxContinueOnXoff := Value;
    Changed;
  end;
end;


{ TVaEventChars }

constructor TVaEventChars.Create;
begin
  inherited Create;
  FXonChar := #17;
  FXoffChar := #19;
  FErrorChar := #0;
  FEventChar := #0;
  FEofChar := #0;
end;

procedure TVaEventChars.SetEventChar(Index: Integer; Value: Char);
begin
  case Index of
    1: FXOnChar := Value;
    2: FXOffChar := Value;
    3: FErrorChar := Value;
    4: FEventChar := Value;
    5: FEofChar := Value;
  end;
  Changed;
end;

procedure TVaEventChars.Assign(Source: TPersistent);
begin
  if (Source <> nil) and (Source is TVaEventChars) then
  begin
    BeginUpdate;
    try
      XonChar := TVaEventChars(Source).XonChar;
      XoffChar := TVaEventChars(Source).XoffChar;
      ErrorChar := TVaEventChars(Source).ErrorChar;
      EventChar := TVaEventChars(Source).EventChar;
      EofChar := TVaEventChars(Source).EofChar;
    finally
      EndUpdate;
    end;
  end else inherited Assign(Source);
end;

{ TVaBuffers }

constructor TVaBuffers.Create;
begin
  inherited Create;
  FReadSize := 4096;
  FReadTimeout := 1000;
  FWriteSize := 2048;
  FWriteTimeout := 1000;
  FXonLimit := 1024;
  FXoffLimit := 1024;
end;

procedure TVaBuffers.Assign(Source: TPersistent);
begin
  if (Source <> nil) and (Source is TVaBuffers) then
  begin
    BeginUpdate;
    try
      FReadSize := TVaBuffers(Source).ReadSize;
      FReadTimeout := TVaBuffers(Source).ReadTimeout;
      FWriteSize := TVaBuffers(Source).WriteSize;
      FWriteTimeout := TVaBuffers(Source).WriteTimeout;
      FXonLimit := TVaBuffers(Source).XonLimit;
      FXoffLimit := TVaBuffers(Source).XoffLimit;
    finally
      EndUpdate;
    end;
  end else inherited Assign(Source);
end;

procedure TVaBuffers.SetReadSize(Value: TVaIntW);
begin
  if FReadSize <> Value then
  begin
    FReadSize := Value;
    Changed;
  end;
end;

procedure TVaBuffers.SetWriteSize(Value: TVaIntW);
begin
  if FWriteSize <> Value then
  begin
    FWriteSize := Value;
    Changed;
  end;
end;

procedure TVaBuffers.SetXonLimit(Value: TVaIntW);
begin
  if FXonLimit <> Value then
  begin
    FXonLimit := Value;
    Changed;
  end;
end;

procedure TVaBuffers.SetXoffLimit(Value: TVaIntW);
begin
  if FXoffLimit <> Value then
  begin
    FXoffLimit := Value;
    Changed;
  end;
end;


end.
