{***************************************************************************}
{ TMS Async32                                                               }
{ for Delphi 4.0,5.0,6.0 & C++Builder 4.0,5.0,6.0                           }
{                                                                           }
{ Copyright 1996 - 2002 by TMS Software                                     }
{ Email : info@tmssoftware.com                                              }
{ Web : http://www.tmssoftware.com                                          }
{                                                                           }
{ The source code is given as is. The author is not responsible             }
{ for any possible damage done due to the use of this code.                 }
{ The component can be freely used in any application. The complete         }
{ source code remains property of the author and may not be distributed,    }
{ published, given or sold in any form as such. No parts of the source      }
{ code can be included in any other component or application without        }
{ written authorization of the author.                                      }
{***************************************************************************}

unit VaServer;

{$I VALIB.INC}

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  VaClasses, VaTypes, VaComm, VaUtils;

type
  TVaServer = class;

  TVaServerClient = class(TVaCustomComm)
  private
    FServer: TVaServer;
    procedure SetServer(Value: TVaServer);
  protected
    procedure HandleDataEvent; override;
    procedure HandleLineEvent(EvtMask: dword); override;
    procedure Notification(AComponent: TComponent; Operation: TOperation); override;
  public
    destructor Destroy; override;
  published
    property Server: TVaServer read FServer write SetServer;
    property AutoOpen;
    property Baudrate;
    property Parity;
//  property ParityCheck;
    property Databits;
    property Stopbits;
    property EventChars;
    property FlowControl;
    property Options;
    property PortNum;
    property DeviceName;
    property MonitorEvents;
    property Buffers;
    property EventPriority;
    property WritePriority;
    property DirectWrite;
    property UpdateDCB;
    property UserBaudrate;
    property OnOpen;
    property OnClose;
  end;

  TVaClientEvent = procedure(Sender: TObject;
    Client: TVaServerClient) of object;
  TVaClientRxCharEvent = procedure(Sender: TObject;
    Client: TVaServerClient; Count: Integer) of object;
  TVaClientErrorEvent = procedure(Sender: TObject;
    Client: TVaServerClient; Errors: Integer) of object;

  TVaServer = class(TVaComponent)
  private
    FClients: TList;
    FOnClientCts: TVaClientEvent;
    FOnClientDsr: TVaClientEvent;
    FOnClientRlsd: TVaClientEvent;
    FOnClientRing: TVaClientEvent;
    FOnClientBreak: TVaClientEvent;
    FOnClientError: TVaClientErrorEvent;
    FOnClientRxChar: TVaClientRxCharEvent;
    FOnClientRxFlag: TVaClientEvent;
    FOnClientTxEmpty: TVaClientEvent;
    FOnClientPErr: TVaClientEvent;
    FOnClientRx80Full: TVaClientEvent;
    FOnClientEvent1: TVaClientEvent;
    FOnClientEvent2: TVaClientEvent;
    function GetClientCount: Integer;
    function GetServerClient(Index: Integer): TVaServerClient;
  protected
    procedure AddClient(Client: TVaServerClient);
    procedure RemoveClient(Client: TVaServerClient);
    procedure SysError(Msg: string; ErrorCode: Integer);
    procedure HandleDataEvent(Client: TVaServerClient);
    procedure HandleLineEvent(Client: TVaServerClient; EvtMask: dword);
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    procedure BroadCast(var Buf; Count: Integer);
    procedure CloseAll; virtual;
    property ServerClient[Index: Integer]: TVaServerClient read GetServerClient;
    property ClientCount: Integer read GetClientCount;
  published
    property OnClientCts: TVaClientEvent read FOnClientCts write FOnClientCts;
    property OnClientDsr: TVaClientEvent read FOnClientDsr write FOnClientDsr;
    property OnClientRlsd: TVaClientEvent read FOnClientRlsd write FOnClientRlsd;
    property OnClientRing: TVaClientEvent read FOnClientRing write FOnClientRing;
    property OnClientBreak: TVaClientEvent read FOnClientBreak write FOnClientBreak;
    property OnClientError: TVaClientErrorEvent read FOnClientError write FOnClientError;
    property OnClientRxChar: TVaClientRxCharEvent read FOnClientRxChar write FOnClientRxChar;
    property OnClientRxFlag: TVaClientEvent read FOnClientRxFlag write FOnClientRxFlag;
    property OnClientTxEmpty: TVaClientEvent read FOnClientTxEmpty write FOnClientTxEmpty;
    property OnClientPErr: TVaClientEvent read FOnClientPErr write FOnClientPErr;
    property OnClientRx80Full: TVaClientEvent read FOnClientRx80Full write FOnClientRx80Full;
    property OnClientEvent1: TVaClientEvent read FOnClientEvent1 write FOnClientEvent1;
    property OnClientEvent2: TVaClientEvent read FOnClientEvent2 write FOnClientEvent2;
  end;



implementation

const
  sErrorClearCommError = 'Error retrieving statistics (%d), %s';


{ TVaServerClient }

destructor TVaServerClient.Destroy;
begin
  Server := nil;
  inherited Destroy;
end;

procedure TVaServerClient.Notification(AComponent: TComponent;
  Operation: TOperation);
begin
  inherited Notification(AComponent, Operation);
  if (Operation = opRemove) then
    if AComponent = Server then Server := nil;
end;

procedure TVaServerClient.SetServer(Value: TVaServer);
begin
  if FServer <> nil then
    FServer.RemoveClient(Self);
  FServer := Value;
  if FServer <> nil then
    FServer.AddClient(Self);
end;

procedure TVaServerClient.HandleDataEvent;
begin
  if FServer <> nil then
    FServer.HandleDataEvent(Self);
end;

procedure TVaServerClient.HandleLineEvent(EvtMask: dword);
begin
  if FServer <> nil then
    FServer.HandleLineEvent(Self, EvtMask);
end;

{ TVaServer }

constructor TVaServer.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  FClients := TList.Create;
end;

destructor TVaServer.Destroy;
begin
  FClients.Free;
  inherited Destroy;
end;

procedure TVaServer.SysError(Msg: string; ErrorCode: Integer);
begin
  raise EVaCommError.CreateFmt(Msg, [ErrorCode, SysErrorMessage(ErrorCode)]);
end;

procedure TVaServer.AddClient(Client: TVaServerClient);
begin
  FClients.Add(Client);
end;

procedure TVaServer.RemoveClient(Client: TVaServerClient);
begin
  FClients.Remove(Client);
end;

function TVaServer.GetClientCount: Integer;
begin
  Result := FClients.Count;
end;

function TVaServer.GetServerClient(Index: Integer): TVaServerClient;
begin
  Result := FClients[Index];
end;

procedure TVaServer.BroadCast(var Buf; Count: Integer);
var
  I: Integer;
begin
  for I := 0 to FClients.Count - 1 do
    ServerClient[I].WriteBuf(Buf, Count);
end;

procedure TVaServer.CloseAll;
var
  I: Integer;
begin
  for I := 0 to FClients.Count - 1 do
    ServerClient[I].Close;
end;

procedure TVaServer.HandleDataEvent(Client: TVaServerClient);
var
  BytesRead: Integer;
begin
  BytesRead := Client.GetCommStatus.Status.cbInQue;
  if (BytesRead > 0) then
    if Assigned(FOnClientRxChar) then
      FOnClientRxChar(Self, Client, BytesRead);
end;

procedure TVaServer.HandleLineEvent(Client: TVaServerClient; EvtMask: dword);
var
  Status: TVaCommStatus;
begin
  if BitOn(EV_ERR, EvtMask) then
  begin
    Status := Client.GetCommStatus;
    if Assigned(FOnClientError) then
      FOnClientError(Self, Client, Status.Errors);
  end;

  if BitOn(EV_RXFLAG, EvtMask) then
    if Assigned(FOnClientRxFlag) then FOnClientRxFlag(Self, Client);

  if BitOn(EV_CTS, EvtMask) then
    if Assigned(FOnClientCts) then FOnClientCts(Self, Client);

  if BitOn(EV_DSR, EvtMask) then
    if Assigned(FOnClientDsr) then FOnClientDsr(Self, Client);

  if BitOn(EV_RING, EvtMask) then
    if Assigned(FOnClientRing) then FOnClientRing(Self, Client);

  if BitOn(EV_RLSD, EvtMask) then
    if Assigned(FOnClientRlsd) then FOnClientRlsd(Self, Client);

  if BitOn(EV_TXEMPTY, EvtMask) then
    if Assigned(FOnClientTxEmpty) then FOnClientTxEmpty(Self, Client);

  if BitOn(EV_BREAK, EvtMask) then
    if Assigned(FOnClientBreak) then FOnClientBreak(Self, Client);

  if BitOn(EV_PERR, EvtMask) then
    if Assigned(FOnClientPErr) then FOnClientPErr(Self, Client);

  if BitOn(EV_RX80FULL, EvtMask) then
    if Assigned(FOnClientRx80Full) then FOnClientRx80Full(Self, Client);

  if BitOn(EV_EVENT1, EvtMask) then
    if Assigned(FOnClientEvent1) then FOnClientEvent1(Self, Client);

  if BitOn(EV_EVENT2, EvtMask) then
    if Assigned(FOnClientEvent2) then FOnClientEvent2(Self, Client);
end;



end.
