{***************************************************************************}
{ TMS Async32                                                               }
{ for Delphi 4.0,5.0,6.0,7.0 & C++Builder 4.0,5.0,6.0                       }
{                                                                           }
{ Copyright 2002 - 2004 by TMS Software                                     }
{ Email : info@tmssoftware.com                                              }
{ Web : http://www.tmssoftware.com                                          }
{                                                                           }
{ The source code is given as is. The author is not responsible             }
{ for any possible damage done due to the use of this code.                 }
{ The component can be freely used in any application. The complete         }
{ source code remains property of the author and may not be distributed,    }
{ published, given or sold in any form as such. No parts of the source      }
{ code can be included in any other component or application without        }
{ written authorization of the author.                                      }
{***************************************************************************}

unit VaTerminal;

{$I VALIB.INC}
{$DEFINE AdditionalChecks}

interface
uses
  Dialogs, Windows, Messages, StdCtrls, SysUtils, Classes, Graphics, Controls,
  Forms, VaComm, Extctrls;

type
  TVaTerminalCursorType = (crsBlock, crsUnderline, crsNone);

  //Character item definition: char, foreground color, background color;
  TVaTerminalChar = record
    Ch: Char;
    FCol: TColor;
    BCol: TColor;
  end;

  //"Screen" buffer - actually larger than the screen in order to support scrollback
  TVaTerminalLine = array of TVaTerminalChar;
  TVaTerminalLineArray = array of TVaTerminalLine;
  TVaTerminal = class;

  IVaTerminalEmulation = class(TComponent)
  protected
    FTerminal: TVaTerminal;
    procedure DataArrived(Data: PVaData; Count: Integer); virtual; abstract;
    procedure KeyDown(var Key: Word; Shift: TShiftState); virtual; abstract;
    procedure KeyPress(var Key: Char); virtual; abstract;
    procedure KeyUp(var Key: Word; Shift: TShiftState); virtual; abstract;
    procedure SetTerminal(const Value: TVaTerminal);
    procedure Notification(AComponent: TComponent; Operation: TOperation); override;
  end;

  TVaTTYEmulation = class(IVaTerminalEmulation)
  private
    FLastChar: Char;
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    procedure DataArrived(Data: PVaData; Count: Integer); override;
    procedure KeyDown(var Key: Word; Shift: TShiftState); override;
    procedure KeyPress(var Key: Char); override;
    procedure KeyUp(var Key: Word; Shift: TShiftState); override;
  end;

  TVaTerminalBuffer = class(TObject)
  private
    FOwner: TVaTerminal;
    FBuf: TVaTerminalLineArray; // fbuf[x,y] - x=row number, y=column number
    FHeight: Integer;
    FWidth: Integer;
    //Position of the last line of the screen in the buffer
    FWinPos: Integer;
    FModified: Boolean;
    function GetLine(I: Integer): TVaTerminalLine;
    procedure SetLine(I: Integer; const Value: TVaTerminalLine);
    function GetCh(X, Y: Integer): TVaTerminalChar;
    procedure SetCh(X, Y: Integer; const Value: TVaTerminalChar);
    function GetBackground(X, Y: Integer): TColor;
    function GetCharacter(X, Y: Integer): Char;
    function GetColor(X, Y: Integer): TColor;
    procedure SetBackground(X, Y: Integer; const Value: TColor);
    procedure SetCharacter(X, Y: Integer; const Value: Char);
    procedure SetColor(X, Y: Integer; const Value: TColor);
  public
    constructor Create(AOwner: TVaTerminal; AWidth, AHeight: Integer);
    destructor Destroy; override;
    //Readjusts buffer size
    procedure AdjustWidth(Value: Integer);
    procedure AdjustHeight(Value: Integer);
    procedure Assign(const Value: TVaterminalBuffer);
    //Clears the whole buffer
    procedure ClearBuffer;
    //Clears only what is visible
    procedure ClearScreen;
    //Inserts a new line and moves the first line pointer one position
    procedure InsertLineAndScroll;
    //Gets/Sets the item on line Y and row X. [0,0]=BottomLeft
    property Ch[X, Y: Integer]: TVaTerminalChar read GetCh write SetCh;
    //Gets/Sets the value of the character on line Y and row X. [0,0]=BottomLeft
    property Character[X, Y: Integer]: Char read GetCharacter write SetCharacter;
    //Gets/Sets the color of the character on line Y and row X. [0,0]=BottomLeft
    property Color[X, Y: Integer]: TColor read GetColor write SetColor;
    //Gets/Sets the backgorund color of the character on line Y and row X. [0,0]=BottomLeft
    property Background[X, Y: Integer]: TColor read GetBackground write SetBackground;
    //Gets/Sets line I, starting from the last one "upwards"
    property Line[I: Integer]: TVaTerminalLine read GetLine write SetLine;
    //True if it was modified. It is updated on paint and writes
    property Modified: Boolean read FModified write FModified;
  end;

  TVaTerminal = class(TCustomPanel, IVaCommUser)
  private
    FBorderWidth: Integer;
    FPainted: Boolean; //True when scrollbar position changed and window scrolled - used in paint;
    FBlinkTime, FSavedBlinkTime: Integer;
    FFromScroll: Boolean;
    FDrawFinOK: Boolean;
    FBuffer: TVaTerminalBuffer;
    FFileCapt: file;
    FCaretHeight: Integer;
    FDrawBuffer: TBitmap; //For UpdateDisplay
    FOldBuffer: TVaTerminalBuffer;
    FColumns: Integer;
    FDispatcher: TVaDispatcher;
    FRows: Integer;
    FCaretShown: Boolean;
    FCaretCreated: Boolean;
    FComm: TVaComm;
    FCursorType: TVaTerminalCursorType;
    FLocalEcho: Boolean;
    FFont: TFont;
    FInternalLeading: Integer;
    FCapture: Boolean;
    FCaptureFile: string;
    FCharWidth: Integer;
    FLineHeight: Integer;
    FOwner: TComponent;
    FScrollBack: Boolean;
    FScrollBackRows: Integer;
    FScrollBar: TScrollBar;
    FCaretX: Integer;
    FCaretY: Integer;
    FWantAllKeys: Boolean;
    FColor: TColor;
    FOldFontColor: TColor;
    FInFontChanged: Boolean;
    FScrollPos: Integer;
    FTextTop: Integer;
    FTextLeft: Integer;
    FTextWidth: Integer;
    FTextHeight: Integer;
    FEmulation: IVaTerminalEmulation;
    FCaptureAppend: Boolean;
    FLocalEchoColor: TColor;
    procedure ReadjustWidth;
    procedure ReadjustHeight;
    procedure LineChanged(Events: dword);
    procedure DataChanged(Data: PVaData; Count: Integer);
    procedure DrawScroll(const nrLines: Integer);
    procedure FontChanged(Sender: TObject);
    procedure SetColumns(Value: Integer);
    procedure SetRows(Value: Integer);
    procedure SetComm(const Value: TVaComm);
    procedure SetCursorType(const Value: TVaTerminalCursorType);
    procedure SetBlinkTime(const Value: Integer);
    procedure SetWantAllKeys(const Value: Boolean);
    procedure SetLocalEcho(const Value: Boolean);
    procedure SetCapture(const Value: Boolean);
    procedure SetCaptureFile(const Value: string);
    procedure SetScrollBack(const Value: Boolean);
    procedure SetScrollBackRows(Value: Integer);
    procedure SetupFont;
    function GetCharacter(X, Y: Integer): Char;
    procedure SetCharacter(X, Y: Integer; const Value: Char);
    function GetCharacterBackground(X, Y: Integer): TColor;
    function GetCharacterColor(X, Y: Integer): TColor;
    procedure SetCharacterBackground(X, Y: Integer; const Value: TColor);
    procedure SetCharacterColor(X, Y: Integer; const Value: TColor);
    procedure SetBufferBackground(X, Y: Integer; const Value: TColor);
    procedure SetBufferChar(X, Y: Integer; const Value: Char);
    procedure SetBufferColor(X, Y: Integer; const Value: TColor);
    procedure SetFont(const Value: TFont);
    procedure SetColor(const Value: TColor);
    procedure SetColorForAll(const Value: TColor);
    procedure SetBackgroundForAll(const Value: TColor);
    procedure ScrollBarScroll(Sender: TObject; ScrollCode: TScrollCode; var ScrollPos: Integer);
    procedure SetEmulation(Value: IVaTerminalEmulation);
    procedure PositionCaret;
    procedure SetCaretX(const Value: Integer);
    procedure SetCaretY(const Value: Integer);
    procedure SaveAndSetCaretBlinkTime;
    procedure RestoreCaretBlinkTime;
  protected
    //Paints the control
    procedure Paint; override;
    //Sends keyboard input to emulation. It check first whether Key=0 and if so, exits
    procedure KeyDown(var Key: Word; Shift: TShiftState); override;
    //For character handling
    procedure KeyPress(var Key: Char); override;
    //Sends keyboard input to emulation. It check first whether Key=0 and if so, exits
    procedure KeyUp(var Key: Word; Shift: TShiftState); override;
    //To check if Comm is removed
    procedure Notification(AComponent: TComponent; Operation: TOperation); override;
    //We handle background erasing ourselves
    procedure WMEraseBkgnd(var M: TMessage); message WM_ERASEBKGND;
    //To enable all keys processing
    procedure WMGetDlgCode(var Message: TWMGetDlgCode); message WM_GETDLGCODE;
    //To get focus when mouse is clicked
    procedure MouseDown(Button: TMouseButton; Shift: TShiftState; X, Y: Integer); override;
    //To handle caret showing
    procedure WMSetFocus(var Message: TWMSetFocus); message WM_SETFOCUS;
    //To handle caret disposal
    procedure WMKillFocus(var Message: TWMSetFocus); message WM_KILLFOCUS;
    // Does not allow window resize
    procedure OnCanResizeEventHandler(Sender: TObject; var NewWidth, NewHeight: Integer; var Resize: Boolean);
    // Compute correct width and height
    function CalculateWidth: integer;
    function CalculateHeight: integer;
  public
    //Adds text at cursor, wrapping longer lines and updating cursor position
    procedure AddText(const Text: string);
    //AddText+newline,tab and backspace (only on the same line i.e. it doesn't change
    //  rowposition, only columnposition) special handling.
    procedure AddControlText(const Text: string; const Color, Bkg: TColor);
    //Clears the screen
    procedure Clear;
    //Clears the whole buffer (including screen)
    procedure ClearAll;
    //Clears the whole buffer (including screen) without repainting
    procedure ClearAllBuffer;
    //Clears the screen without repainting
    procedure ClearScreenBuffer;
    //Moves the caret to X,Y on the terminal
    procedure GotoXY(X, Y: Integer);
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    //Inserts void lines and moves screen pointer down. Updates display
    procedure InsertAndScrollF(nrLines: Integer);
    //Moves back screen pointer; returns the number of lines actually scrolled (depends on the nr. of scrollback rows)
    //Updates display
    function Scroll(nrLines: Integer): Integer;
    //Updates on display only what changed
    procedure UpdateDisplay;
    //TVaDispatcher replicas. Used to send data over the serial.
    function WriteBuf(var Buf; Count: Integer): Integer;
    function WriteChar(Ch: Char): Boolean;
    function WriteText(const s: string): Boolean;

    // Modifies the character, color or background of the item at x,y. Does NOT update display
    // 0,0 denotes Left,Bottom Corner
    property BufferBackground[X, Y: Integer]: TColor read GetCharacterBackground write SetBufferBackground;
    property BufferChar[X, Y: Integer]: Char read GetCharacter write SetBufferChar;
    property BufferColor[X, Y: Integer]: TColor read GetCharacterColor write SetBufferColor;
    //Sets/Gets caret position - similar to GotoXY.
    // WARNING: These are logical screen coordinates: 0,0 = BottomLeft
    property CaretX: Integer read FCaretX write SetCaretX;
    property CaretY: Integer read FCaretY write SetCaretY;
    //Modifies the character, color or background of the item at x,y. Updates display
    //0,0 denotes Left,Bottom Corner
    property Character[X, Y: Integer]: Char read GetCharacter write SetCharacter;
    property CharacterBackground[X, Y: Integer]: TColor read GetCharacterBackground write SetCharacterBackground;
    property CharacterColor[X, Y: Integer]: TColor read GetCharacterColor write SetCharacterColor;

  published
    //Set to true if you want to append the capture file
    property CaptureAppend: Boolean read FCaptureAppend write FCaptureAppend;
    //Caret blink time in miliseconds.
    // WARNING: Setting blinktime is system wide i.e. is permanent (just as setting from control panel)
    property BlinkTime: Integer read FBlinkTime write SetBlinkTime;
    //Capture serial incoming data ?
    property Capture: Boolean read FCapture write SetCapture;
    //The file the capture is going to
    property CaptureFile: string read FCaptureFile write SetCaptureFile;
    //Default background color
    property Color: TColor read FColor write SetColor;
    //Number of columns
    property Columns: Integer read FColumns write SetColumns default 80;
    //TVaComm that will provide us communication capabilites
    property Comm: TVaComm read FComm write SetComm;
    //Cursor type: block, underline, none
    property CursorType: TVaTerminalCursorType read FCursorType write SetCursorType;
    //Emulation object. If none is selected it defaults to TTY.
    property Emulation: IVaTerminalEmulation read FEmulation write SetEmulation;
    //Echoes *characters* typed locally i.e. without FXX,Home,End, etc. Calls AddControlText for display
    property LocalEcho: Boolean read FLocalEcho write SetLocalEcho;
    property LocalEchoColor: TColor read FLocalEchoColor write FLocalEchoColor;
    //Inherited properties
    property OnClick;
    property OnDblClick;
    property OnEnter;
    property OnExit;
    property OnKeyDown;
    property OnKeyPress;
    property OnKeyUp;
    property OnMouseDown;
    property OnMouseMove;
    property OnMouseUp;
    //Number of rows
    property Rows: Integer read FRows write SetRows default 25;
    //Scroll back scroll bar visible
    property ScrollBack: Boolean read FScrollBack write SetScrollBack default False;
    //Number of rows to hold in buffer for scrollback
    property ScrollBackRows: Integer read FScrollBackRows write SetScrollBackRows default 250;
    //Inherited properties
    property ShowHint;
    //Font used for drawing characters.
    //WARNING: It must be fixed width, otherwise you get an exception and resets font to default
    property Font: TFont read FFont write SetFont;
    //Capture all keys e.g. Tab,Arrows.
    property WantAllKeys: Boolean read FWantAllKeys write SetWantAllKeys;
  end;


implementation

const
  TABCOUNT = 8;

{ TVaTerminal }

//------------------------------------------------------------------------------

function TVaTerminal.GetCharacter(X, Y: Integer): Char;
begin
  Result := FBuffer.Character[X, Y];
end;

//------------------------------------------------------------------------------

function TVaTerminal.GetCharacterBackground(X, Y: Integer): TColor;
begin
  Result := FBuffer.Background[X, Y];
end;

//------------------------------------------------------------------------------

function TVaTerminal.GetCharacterColor(X, Y: Integer): TColor;
begin
  Result := FBuffer.Color[X, Y];
end;

//------------------------------------------------------------------------------

procedure TVaTerminal.GotoXY(X, Y: Integer);
begin
  if X >= FColumns then
    X := FColumns - 1;

  if X < 0 then X := 0;
  if Y >= FRows then
    Y := FRows - 1;

  if Y < 0 then
    Y := 0;

  FCaretX := X;
  FCaretY := Y;
  PositionCaret;
end;

//------------------------------------------------------------------------------

procedure TVaTerminal.SetBufferBackground(X, Y: Integer;
  const Value: TColor);
begin
  FBuffer.Background[X, Y] := Value;
end;

//------------------------------------------------------------------------------

procedure TVaTerminal.SetBufferChar(X, Y: Integer; const Value: Char);
begin
  FBuffer.Character[X, Y] := Value;
end;

//------------------------------------------------------------------------------

procedure TVaTerminal.SetBufferColor(X, Y: Integer; const Value: TColor);
begin
  FBuffer.Color[X, Y] := Value;
end;

//------------------------------------------------------------------------------

procedure TVaTerminal.InsertAndScrollF(nrLines: Integer);
var
  i: Integer;
begin
  if nrLines < 1 then
    Exit;
  //To draw pending buffer writes before scrolling (by moving the bitmap of what
  // was drawn last time
  UpdateDisplay;
  for i := 1 to nrLines do
    FBuffer.InsertLineAndScroll;
  if Abs(nrLines) < (FRows - 2) then
    DrawScroll(-nrLines)
  else
    Invalidate;
end;

//------------------------------------------------------------------------------

function TVaTerminal.Scroll(nrLines: Integer): Integer;
var
  nl, oldSpos: Integer;
begin
  oldSpos := FScrollPos;
  nl := FScrollBar.Position + nrLines;
  if nl < 0 then
    nl := 0;
  ScrollBarScroll(FScrollBar, scPageUp, nl);
  Result := oldSpos - FScrollPos;
  if Result <> 0 then
  begin
    FFromScroll := True;
    Invalidate;
    FScrollBar.Position := nl;
  end;
end;

//------------------------------------------------------------------------------

procedure TVaTerminal.SetBlinkTime(const Value: Integer);
begin
  FBlinkTime := Value;
  if Focused then
    SetCaretBlinkTime(Value);
end;

//------------------------------------------------------------------------------

procedure TVaTerminal.SetCapture(const Value: Boolean);
begin
  if csDesigning in ComponentState then
  begin
    FCapture := Value;
    Exit;
  end;

  if Value = FCapture then
    Exit;

  //It means we now enter capture state
  if (Value) and (FCaptureFile <> '') then
  begin
    AssignFile(FFileCapt, FCaptureFile);
    //Allow the exception to propagate if the file cannot be opened
    if FCaptureAppend then
    begin
      Reset(FFileCapt, 1);
      if not Eof(FFileCapt) then
        Seek(FFileCapt, FileSize(FFileCapt));
    end
    else
      Rewrite(FFileCapt, 1);
  end
  else
  begin
{$I-}
    CloseFile(FFileCapt);
{$I+}
    IOResult; // clear the error flag
  end;
  FCapture := Value;
end;

//------------------------------------------------------------------------------

procedure TVaTerminal.SetCaptureFile(const Value: string);
begin
  if csDesigning in ComponentState then
  begin
    FCaptureFile := Value;
    Exit;
  end;

  if Value = FCaptureFile then
    Exit;

  if FCapture then
  begin
    //Close previous file if already capturing
{$I-}
    CloseFile(FFileCapt);
{$I+}

    AssignFile(FFileCapt, Value);
    //Allow the exception to propagate if the file cannot be opened
    if FCaptureAppend then
    begin
      Reset(FFileCapt, 1);
      if not Eof(FFileCapt) then
        Seek(FFileCapt, FileSize(FFileCapt));
    end
    else
      Rewrite(FFileCapt, 1);
  end;
  FCaptureFile := Value;
end;

//------------------------------------------------------------------------------

procedure TVaTerminal.SetCharacter(X, Y: Integer; const Value: Char);
var
  yCurent: Integer;
begin
  FBuffer.Character[X, Y] := Value;
  yCurent := FTextTop + Height - (Y + 1 - FScrollPos) * FLineHeight;
  if (yCurent >= Height - FBorderWidth) or (yCurent < 0) then
    Exit;

  if FCaretCreated and FCaretShown then
  begin
    HideCaret(Handle);
    FCaretShown := False;
  end;

  Canvas.Brush.Style := bsSolid;
  Canvas.Brush.Color := FBuffer.Background[X, Y];
  Canvas.Font := FFont;
  Canvas.Font.Color := FBuffer.Color[X, Y];
  Canvas.TextFlags := Canvas.TextFlags or ETO_OPAQUE;
  Canvas.TextOut(FTextLeft + X * FCharWidth, yCurent, Value);

  with FDrawBuffer do
  begin
    Canvas.Brush.Style := bsSolid;
    Canvas.Brush.Color := FBuffer.Background[X, Y];
    Canvas.Font := FFont;
    Canvas.Font.Color := FBuffer.Color[X, Y];
    Canvas.TextFlags := Canvas.TextFlags or ETO_OPAQUE;
    Canvas.TextOut(FTextLeft + X * FCharWidth, yCurent, Value);
  end;

  FOldBuffer.Character[X, Y] := Value;
  if FCaretCreated and (FCursorType <> crsNone) then
  begin
    ShowCaret(Handle);
    FCaretShown := True;
  end;

end;

//------------------------------------------------------------------------------

procedure TVaTerminal.SetCharacterBackground(X, Y: Integer;
  const Value: TColor);
var
  yCurent: Integer;
begin
  FBuffer.Background[X, Y] := Value;
  yCurent := FTextTop + Height - (Y + 1 - FScrollPos) * FLineHeight;
  if (yCurent >= Height - FBorderWidth) or (yCurent < 0) then
    Exit;

  if FCaretCreated and FCaretShown then
  begin
    HideCaret(Handle);
    FCaretShown := False;
  end;

  Canvas.Brush.Style := bsSolid;
  Canvas.Brush.Color := FBuffer.Background[X, Y];
  Canvas.Font := FFont;
  Canvas.Font.Color := FBuffer.Color[X, Y];
  Canvas.TextFlags := Canvas.TextFlags or ETO_OPAQUE;
  Canvas.TextOut(FTextLeft + X * FCharWidth, yCurent, FBuffer.Character[X, Y]);

  with FDrawBuffer do
  begin
    Canvas.Brush.Style := bsSolid;
    Canvas.Brush.Color := FBuffer.Background[X, Y];
    Canvas.Font := FFont;
    Canvas.Font.Color := FBuffer.Color[X, Y];
    Canvas.TextFlags := Canvas.TextFlags or ETO_OPAQUE;
    Canvas.TextOut(FTextLeft + X * FCharWidth, yCurent, FBuffer.Character[X, Y]);
  end;

  FOldBuffer.Background[X, Y] := Value;

  if FCaretCreated and (FCursorType <> crsNone) then
  begin
    ShowCaret(Handle);
    FCaretShown := True;
  end;

end;

//------------------------------------------------------------------------------

procedure TVaTerminal.SetCharacterColor(X, Y: Integer;
  const Value: TColor);
var
  yCurent: Integer;
begin
  FBuffer.Color[X, Y] := Value;
  yCurent := FTextTop + Height - (Y + 1 - FScrollPos) * FLineHeight;

  if (yCurent >= Height - FBorderWidth) or (yCurent < 0) then
    Exit;

  if FCaretCreated and FCaretShown then
  begin
    HideCaret(Handle);
    FCaretShown := False;
  end;

  Canvas.Brush.Style := bsSolid;
  Canvas.Brush.Color := FBuffer.Background[X, Y];
  Canvas.Font := FFont;
  Canvas.Font.Color := FBuffer.Color[X, Y];
  Canvas.TextFlags := Canvas.TextFlags or ETO_OPAQUE;
  Canvas.TextOut(FTextLeft + X * FCharWidth, yCurent, FBuffer.Character[X, Y]);

  with FDrawBuffer do
  begin
    Canvas.Brush.Style := bsSolid;
    Canvas.Brush.Color := FBuffer.Background[X, Y];
    Canvas.Font := FFont;
    Canvas.Font.Color := FBuffer.Color[X, Y];
    Canvas.TextFlags := Canvas.TextFlags or ETO_OPAQUE;
    Canvas.TextOut(FTextLeft + X * FCharWidth, yCurent, FBuffer.Character[X, Y]);
  end;

  FOldBuffer.Color[X, Y] := Value;

  if FCaretCreated and (FCursorType <> crsNone) then
  begin
    ShowCaret(Handle);
    FCaretShown := True;
  end;

end;

//------------------------------------------------------------------------------

procedure TVaTerminal.SetColumns(Value: Integer);
begin
  if Value < 3 then
    Value := 3;
  if (FColumns = Value) then
    Exit;
  FColumns := Value;
  FBuffer.AdjustWidth(Value);
  ReadjustWidth;
end;

//------------------------------------------------------------------------------

procedure TVaTerminal.SetComm(const Value: TVaComm);
begin
  if FComm <> nil then
    FComm.RemoveClient(FDispatcher);
  FComm := Value;
  if FComm <> nil then
    FComm.InsertClient(FDispatcher);
end;

//------------------------------------------------------------------------------

procedure TVaTerminal.SetCursorType(const Value: TVaTerminalCursorType);
begin
  if FCaretShown then
  begin
    HideCaret(Handle);
    FCaretShown := FALSE;
  end;

  if FCaretCreated then
  begin
    DestroyCaret;
    FCaretCreated := FALSE;
  end;

  FCursorType := Value;

  if not Focused then exit;

  case FCursorType of
    crsUnderline:
      begin
        FCaretCreated := CreateCaret(Handle, 0, FCharWidth, 2);
        FCaretHeight := 2;
      end;
    crsBlock:
      begin
        FCaretCreated := CreateCaret(Handle, 0, FCharWidth, FLineHeight);
        FCaretHeight := FLineHeight;
      end;
  end;
  PositionCaret;
end;

//------------------------------------------------------------------------------

procedure TVaTerminal.SetLocalEcho(const Value: Boolean);
begin
  FLocalEcho := Value;
end;

//------------------------------------------------------------------------------

procedure TVaTerminal.SetRows(Value: Integer);
begin
  if Value < 3 then
    Value := 3;
  if (FRows = Value) then
    Exit;
  FBuffer.AdjustHeight(Value + FScrollBackRows);
  //FScrollBar.Position:=0;
  try
    if Value < FRows then
    begin
      FScrollBar.LargeChange := Value;
      FScrollBar.PageSize := Value;
    end;
    if FScrollBackRows > 0 then
      FScrollBar.Max := Value + FScrollBackRows - 1;

    if Value > FRows then
    begin
      FScrollBar.LargeChange := Value;
      FScrollBar.PageSize := Value;
    end;
    FScrollPos := 0;
    if FScrollBackRows > 0 then
    begin
      FScrollBar.Position := FScrollBackRows;
    end;
  finally
    FRows := Value;
    Invalidate;
    ReadjustHeight;
  end;
end;

//------------------------------------------------------------------------------

procedure TVaTerminal.SetScrollBack(const Value: Boolean);
begin
  if Value = FScrollBack then
    Exit;
  FScrollBack := Value;
  if Value and (FScrollBackRows < 1) then
    ScrollBackRows := 1;
  ReadjustWidth;
  if not (csDesigning in ComponentState) then
    FScrollBar.Visible := Value;
  invalidate;
end;

//------------------------------------------------------------------------------

procedure TVaTerminal.SetScrollBackRows(Value: Integer);
begin
  if Value = FScrollBackRows then
    Exit;
  if Value < 1 then
    Value := 1;

  FBuffer.AdjustHeight(Value + FRows);

  //FScrollBar.SetParams(Value+FRows-1-FScrollBar.PageSize,0,Value+FRows-2);
  FScrollBar.Max := Value + FRows - 1;
  FScrollBar.Position := Value;
  FScrollBar.LargeChange := FRows;
  FScrollBar.PageSize := FRows;

  FScrollPos := 0;
  FScrollBackRows := Value;
  Invalidate;
  //FScrollBar.Repaint;
end;

//------------------------------------------------------------------------------

procedure TVaTerminal.SetWantAllKeys(const Value: Boolean);
begin
  FWantAllKeys := Value;
end;

//------------------------------------------------------------------------------

procedure TVaTerminal.Paint;
var
  ccol, cbcol: TColor;
  i, j, xCurent, yCurent: Integer;
  cline: TVaTerminalLine;
  SCurent: string;
begin
  cline := nil;
  if csDesigning in ComponentState then
  begin
    inherited;
    Canvas.Brush.Style := bsSolid;
    Canvas.Brush.Color := Color;
    Canvas.FillRect(Rect(0, 0, Width, Height));
    Exit;
  end;

  if FFromScroll then
  begin
    FFromScroll := False;
    Exit;
  end;

  FDrawFinOK := False;
  try
    FDrawBuffer.Height := FTextHeight + 2 * FBorderWidth;

    if FScrollBack then
      FDrawBuffer.Width := FTextWidth + 2 * FBorderWidth
    else
      FDrawBuffer.Width := FTextWidth + FBorderWidth;

    with FDrawBuffer.Canvas do
    begin
      Brush.Style := bsSolid;
      Brush.Color := Color;

      ccol := FBuffer.Color[0, 0];
      cbcol := FBuffer.Background[0, 0];
      Font := FFont;
      Font.Color := ccol;
      Brush.Color := cbcol;
      Brush.Style := bsSolid;
      TextFlags := TextFlags or ETO_OPAQUE;
      FillRect(self.ClientRect);
      for i := FScrollPos to FRows - 1 + FScrollPos do
      begin
        cline := FBuffer.Line[i];
        yCurent := FTextTop + FTextHeight - (i + 1 - FScrollPos) * FLineHeight;
        SCurent := '';
        xCurent := FTextLeft;
        for j := 0 to FColumns - 1 do
        begin
          if ccol <> cline[j].FCol then
          begin
            if SCurent <> '' then
              windows.TextOut(Handle, xCurent, yCurent, PChar(SCurent), Length(SCurent));

            ccol := cline[j].FCol;
            Font.Color := ccol;
            SCurent := '';
            xCurent := FTextLeft + j * FCharWidth;
          end;

          if cbcol <> cline[j].BCol then
          begin
            if SCurent <> '' then
              windows.TextOut(Handle, xCurent, yCurent, PChar(SCurent), Length(SCurent));

            cbcol := cline[j].BCol;
            Brush.Color := cbcol;
            Brush.Style := bsSolid;
            SCurent := '';
            xCurent := FTextLeft + j * FCharWidth;
          end;
          SCurent := SCurent + cline[j].Ch;
        end;
        if SCurent <> '' then
          Textout(xCurent, yCurent, SCurent);
      end;
    end;
    FDrawFinOK := True;
  finally
    if FCaretCreated and FCaretShown then
    begin
      HideCaret(Handle);
      FCaretShown := False;
    end;

    Canvas.Draw(0, 0, FDrawBuffer);

    if FCaretCreated and (FCursorType <> crsNone) then
    begin
      PositionCaret;
      ShowCaret(Handle);
      FCaretShown := True;
    end;

    //Save what has been drawn so Updatedisplay only draws what changed since now
    FOldBuffer.Assign(FBuffer);
    FBuffer.Modified := False;
  end;
  FPainted := True;
end;

//------------------------------------------------------------------------------

procedure TVaTerminal.Clear;
begin
  FBuffer.ClearScreen;
  Invalidate;
end;

//------------------------------------------------------------------------------

procedure TVaTerminal.ClearAll;
begin
  FBuffer.ClearBuffer;
  Invalidate;
end;

//------------------------------------------------------------------------------

procedure TVaTerminal.ClearAllBuffer;
begin
  FBuffer.ClearBuffer;
end;

//------------------------------------------------------------------------------

procedure TVaTerminal.ClearScreenBuffer;
begin
  FBuffer.ClearScreen;
end;

//------------------------------------------------------------------------------

constructor TVaTerminal.Create(AOwner: TComponent);
var i, j: Integer;
  cline: TVaTerminalLine;
begin
  inherited Create(AOwner);
  ControlStyle := ControlStyle - [csAcceptsControls];
  FScrollBackRows := 250;
  FDrawFinOK := True;
  FFromScroll := False;
  FDrawBuffer := TBitmap.Create;
  FBorderWidth := 0;
  FOwner := AOwner;
  FScrollBack := False;
  FBuffer := TVaTerminalBuffer.Create(self, 80, 26);
  FOldBuffer := TVaTerminalBuffer.Create(self, 80, 26);
  FScrollBar := TScrollBar.Create(Self);
  FScrollPos := 0;
  TabStop := True;
  FLocalEchoColor := clWhite;

  with FScrollBar do
  begin
    Parent := Self;
    Visible := False;
    Kind := sbVertical;
    Align := alRight;
    Min := 0;
    Position := 0;
    Max := 25;
    PageSize := 24;
    LargeChange := 24;
    OnScroll := ScrollBarScroll;
    DoubleBuffered := True;
    TabStop := False;
    ControlStyle := ControlStyle + [csNoDesignVisible];
  end;

  FFont := TFont.Create;
  FFont.Name := 'Terminal';
  FFont.Size := 10;
  FFont.Style := [];
  FFont.Pitch := fpFixed;
  FFont.Color := clWhite;
  FFont.Charset := ANSI_CHARSET;
  FOldFontColor := FFont.Color;
 // FFont.Charset:=OEM_CHARSET;
  SetupFont;
  FInFontChanged := False;
  FFont.OnChange := FontChanged;

  Rows := 25;
  Columns := 80;
  Color := clBlack;
  FDispatcher := TVaDispatcher.Create(Self);
  if csDesigning in ComponentState then
    FDispatcher.Active := False
  else
    FDispatcher.Active := True;

  cline := nil;
  for i := 0 to FRows - 1 + FScrollBackRows do
  begin
    cline := FBuffer.Line[i];
    for j := 0 to FColumns - 1 do
    begin
      cline[j].Ch := #32;
      cline[j].FCol := Font.Color;
      cline[j].BCol := Color;
    end;
  end;
  FTextTop := 0;
  FTextLeft := 0;
  FTextWidth := Width;
  FTextHeight := Height;
  FCaretX := 0;
  FCaretY := 0;

  FCapture := False;
  FCaptureFile := '';
  FPainted := True;
  FBlinkTime := 200;
  FEmulation := nil;
  OnCanResize := OnCanResizeEventHandler;
end;

//------------------------------------------------------------------------------

destructor TVaTerminal.Destroy;
begin
  if Focused then
    RestoreCaretBlinkTime;
  FEmulation := nil;
  FBuffer.Free;
  FFont.Free;
  FDispatcher.Free;
  FDrawBuffer.Free;
  inherited;
end;

//------------------------------------------------------------------------------

procedure TVaTerminal.UpdateDisplay;
var
  ccol, cbcol: TColor;
  i, j, xCurent, yCurent: Integer;
  cline, cOldLine: TVaTerminalLine;
  SCurent: string;
begin
  cline := nil;
  cOldLine := nil;

  if not FBuffer.Modified then
    Exit;

  if FCaretCreated and FCaretShown then
  begin
    HideCaret(Handle);
    FCaretShown := False;
  end;

  FDrawFinOK := False;

  with FDrawBuffer.Canvas do
  begin
    ccol := FBuffer.Color[0, 0];
    cbcol := FBuffer.Background[0, 0];
    Font := FFont;
    Font.Color := ccol;
    Brush.Color := cbcol;
    Brush.Style := bsSolid;
    TextFlags := TextFlags or ETO_OPAQUE;
    for i := FScrollPos to FRows - 1 + FScrollPos do
    begin
      cline := FBuffer.Line[i];
      cOldLine := FOldBuffer.Line[i];
      yCurent := FTextTop + FTextHeight - (i + 1 - FScrollPos) * FLineHeight;
      SCurent := '';
      xCurent := FTextLeft;
      for j := 0 to FColumns - 1 do
      begin
        if (cline[j].Ch = cOldLine[j].Ch) and
          (cline[j].FCol = cOldLine[j].FCol) and
          (cline[j].BCol = cOldLine[j].BCol) then
        begin
          if SCurent <> '' then
          begin
            Textout(xCurent, yCurent, SCurent);
            Scurent := '';
          end;
          Continue;
        end;
        if ccol <> cline[j].FCol then
        begin
          if SCurent <> '' then
            Textout(xCurent, yCurent, SCurent);
          ccol := cline[j].FCol;
          Font.Color := ccol;
          SCurent := '';
        end;
        if cbcol <> cline[j].BCol then
        begin
          if SCurent <> '' then
            Textout(xCurent, yCurent, SCurent);
          cbcol := cline[j].BCol;
          Brush.Color := cbcol;
          Brush.Style := bsSolid;
          SCurent := '';
        end;

        if SCurent = '' then
          xCurent := FTextLeft + j * FCharWidth;
        SCurent := SCurent + cline[j].Ch;
      end;
      if SCurent <> '' then
        Textout(xCurent, yCurent, SCurent);
    end;
  end;

  Canvas.Draw(0, 0, FDrawBuffer);
  FBuffer.Modified := False;
  FDrawFinOK := True;
  if FCaretCreated and (FCursorType <> crsNone) then
  begin
    ShowCaret(Handle);
    FCaretShown := True;
  end;

  FOldBuffer.Assign(FBuffer);
end;

//------------------------------------------------------------------------------

procedure TVaTerminal.SetFont(const Value: TFont);
var
  OldFont: TFont;
begin
  OldFont := TFont.Create;
  OldFont.Assign(FFont);
  FFont.Assign(Value);
  FFont.Pitch := fpFixed;
  try
    SetupFont;
  except
    on Exception do FFont.Assign(OldFont);
  end;
  OldFont.Free;
  SetColorForAll(Value.Color);
  Invalidate;
end;

//------------------------------------------------------------------------------

function GetFixFont(var realfont: TFont): boolean;
var
  DC: HDC;
  Metrics: TTextMetric;
  hObject: THandle;
  i: integer;
  s: string;
begin
  i := 0;
  result := false;
  s := realfont.Name;
  repeat
    if screen.Fonts.Count <= i then
    begin
      realfont.Name := s;
      Exit;
    end;
    DC := GetDC(0);
    fillchar(Metrics, sizeof(Metrics), 0);
    realfont.Name := screen.Fonts.Strings[i];
    realfont.Charset := ANSI_CHARSET;
    realfont.Style := [];
    realfont.Pitch := fpFixed;
    hObject := SelectObject(DC, realfont.Handle);
    GetTextMetrics(dc, Metrics);
    SelectObject(DC, hOBject);
    ReleaseDC(0, DC);
    inc(i);
  until Metrics.tmAveCharWidth = Metrics.tmMaxCharWidth;
  Result := True;
end;

procedure TVaTerminal.SetupFont;
var
  DC: HDC;
  Metrics: TTextMetric;
  hObject: THandle;
  localstyle: Tfontstyles;
begin
  DC := GetDC(0);
  fillchar(Metrics, sizeof(Metrics), 0);
  localstyle := FFont.style;
  FFont.Charset := ANSI_CHARSET;
  FFont.Style := [];
  hObject := SelectObject(DC, FFont.Handle);
  GetTextMetrics(dc, Metrics);
  SelectObject(DC, hOBject);
  ReleaseDC(0, DC);
  FFont.Style := localstyle;

  if Metrics.tmAveCharWidth <> Metrics.tmMaxCharWidth then
  begin
    if not GetFixFont(FFont) then
    begin
      if csDesigning in ComponentState then
        MessageDlg('Can''t find a fixed font', mtError, [mbYes], 0)
      else
        raise Exception.Create('Can''t find a fixed font');
    end;
  end;

  FCharWidth := Metrics.tmMaxCharWidth;
  FLineHeight := Metrics.tmHeight;
  FInternalLeading := Metrics.tmInternalLeading;
end;

//------------------------------------------------------------------------------

procedure TVaTerminal.DataChanged(Data: PVaData; Count: Integer);
begin
  if FCapture then
    BlockWrite(FFileCapt, Data^, Count);

  if Assigned(FEmulation) then
    FEmulation.DataArrived(Data, Count);
end;

//------------------------------------------------------------------------------

procedure TVaTerminal.LineChanged(Events: dword);
begin
  //Do nothing. We are not interested in line events.
end;

//------------------------------------------------------------------------------

procedure TVaTerminal.Notification(AComponent: TComponent;
  Operation: TOperation);
begin
  inherited Notification(AComponent, Operation);

  if (Operation = opRemove) then
  begin
    if AComponent = Comm then
      Comm := nil;
  end;

  if (Operation = opRemove) then
  begin
    if AComponent = FEmulation then
      FEmulation := nil;
  end;

end;

//------------------------------------------------------------------------------

procedure TVaTerminal.WMEraseBkgnd(var M: TMessage);
begin
  M.Result := 1;
end;

//------------------------------------------------------------------------------

procedure TVaTerminal.SetColor(const Value: TColor);
begin
  FColor := Value;
  SetBackgroundForAll(Value);
  Invalidate;
end;

//------------------------------------------------------------------------------

procedure TVaTerminal.SetBackgroundForAll(const Value: TColor);
var
  i, j: Integer;
  cline: TVaTerminalLine;
begin
  cline := nil;
  for i := 0 to FRows + FScrollBackRows - 1 do
  begin
    cline := FBuffer.Line[i];
    for j := 0 to FColumns - 1 do
    begin
      cline[j].BCol := Value;
    end;
  end;

end;

//------------------------------------------------------------------------------

procedure TVaTerminal.setColorForAll(const Value: TColor);
var
  i, j: Integer;
  cline: TVaTerminalLine;
begin
  cline := nil;
  for i := 0 to FRows + FScrollBackRows - 1 do
  begin
    cline := FBuffer.Line[i];
    for j := 0 to FColumns - 1 do
    begin
      cline[j].FCol := Value;
    end;
  end;

end;

//------------------------------------------------------------------------------

procedure TVaTerminal.FontChanged(Sender: TObject);
var
  fnt: TFont;
  i, j: Integer;
  cline: TVaTerminalLine;
begin
  cline := nil;
  if FInFontChanged then
    Exit;
  if not (Sender is TFont) then
    Exit;

  FInFontChanged := True;
  try
    fnt := Sender as TFont;
    //If color changed then change color for all characters which had the default font color
    // Of course, if that was assigned specifically by setcolor, it is also changed.
    if fnt.Color <> FOldFontColor then
    begin
      for i := 0 to FRows - 1 do
      begin
        cline := FBuffer.Line[i];
        for j := 0 to FColumns - 1 do
        begin
          if cline[j].FCol = FOldFontColor then
            cline[j].FCol := fnt.Color;
        end;
      end;
      FOldFontColor := fnt.Color;
    end;
    fnt.Pitch := fpFixed;
    try
      SetupFont;
    except
      on Exception do
      begin
        FFont.Name := 'Terminal';
        FFont.Size := 10;
        FFont.Style := [];
        FFont.Pitch := fpFixed;
        FFont.Color := clWhite;
        FFont.Charset := ANSI_CHARSET;
        FOldFontColor := FFont.Color;
        SetupFont;
        Invalidate;
      end;
    end;
  finally
    FInFontChanged := False;
    ReadjustWidth;
    ReadjustHeight;
  end;
end;

//------------------------------------------------------------------------------

function TVaTerminal.CalculateHeight: integer;
begin
  FTextHeight := FLineHeight * FRows;
  Result := FLineHeight * FRows + 2 * FBorderWidth + FInternalLeading;
end;

procedure TVaTerminal.ReadjustHeight;
begin
  Height := CalculateHeight;
end;

//------------------------------------------------------------------------------

function TVaTerminal.CalculateWidth: integer;
begin
  FTextWidth := FColumns * FCharWidth;
  if FScrollBack then
    Result := FColumns * FCharWidth + FBorderWidth + FScrollBar.Width + 2
  else
    Result := FColumns * FCharWidth + 2 * FBorderWidth;
end;

//------------------------------------------------------------------------------

procedure TVaTerminal.ReadjustWidth;
begin
  Width := CalculateWidth;
end;

//------------------------------------------------------------------------------

procedure TVaTerminal.ScrollBarScroll(Sender: TObject;
  ScrollCode: TScrollCode; var ScrollPos: Integer);
var
  dif: Integer;
  Bad: Boolean;
begin
  if FFromScroll or not FPainted then
    Bad := True
  else
    Bad := False;

  if ScrollPos > (FScrollBar.Max + 1 - FScrollBar.PageSize) then
    ScrollPos := FScrollBar.Max + 1 - FScrollBar.PageSize;

  if ScrollPos = FScrollBar.Position then
    Exit;

  dif := FScrollBar.Max + 1 - FScrollBar.PageSize - ScrollPos - FScrollPos;

  FScrollPos := FScrollBar.Max + 1 - FScrollBar.PageSize - ScrollPos;

  if FScrollPos < 0 then
  begin
    ScrollPos := ScrollPos + FScrollPos;
    FScrollPos := 0;
    Beep;
  end;

  if FScrollPos > FScrollBackRows then
  begin
    ScrollPos := ScrollPos + FScrollPos - FScrollBackRows;
    FScrollPos := FScrollBackRows;
    Beep;
  end;

  if (Abs(dif) < (FRows - 2)) and FDrawFinOK and not Bad then
  begin
    FFromScroll := True;
    //To draw pending buffer writes before scrolling (by moving the bitmap of what
    // was drawn last time
    UpdateDisplay;
    DrawScroll(dif);
  end
  else
  begin
    FFromScroll := False;
    FPainted := False;
  end;
  Invalidate;
end;

//------------------------------------------------------------------------------

procedure TVaTerminal.SetEmulation(Value: IVaTerminalEmulation);
begin
  FEmulation := Value;

  if Assigned(FEmulation) then
    FEmulation.SetTerminal(Self);
end;

//------------------------------------------------------------------------------

procedure TVaTerminal.KeyDown(var Key: Word; Shift: TShiftState);
begin
  inherited;
  if Key = 0 then Exit;
  if not assigned(Emulation) then exit;
  Emulation.KeyDown(Key, Shift);
end;

//------------------------------------------------------------------------------

procedure TVaTerminal.KeyPress(var Key: Char);
begin
  inherited;
  if not assigned(Emulation) then exit;
  if (Key = #0) then
    Exit;
  if FLocalEcho then
    AddControlText(Key, FLocalEchoColor, Color);
  Emulation.KeyPress(Key);
end;

//------------------------------------------------------------------------------

procedure TVaTerminal.KeyUp(var Key: Word; Shift: TShiftState);
begin
  inherited;
  if not assigned(Emulation) then exit;

  if Key <> 0 then
    Emulation.KeyUp(Key, Shift);
end;

//------------------------------------------------------------------------------

procedure TVaTerminal.WMGetDlgCode(var Message: TWMGetDlgCode);
begin
  inherited;
  if FWantAllKeys then
    Message.Result := Message.Result or DLGC_WANTALLKEYS or DLGC_WANTTAB or DLGC_WANTARROWS;
end;

//------------------------------------------------------------------------------

procedure TVaTerminal.MouseDown(Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
begin
  inherited;
  if not Focused then SetFocus;
end;

//------------------------------------------------------------------------------

procedure TVaTerminal.WMSetFocus(var Message: TWMSetFocus);
begin
  inherited;

  { A window should create a caret only when it has the keyboard focus or is
  active. The window should destroy the caret before losing the keyboard focus
  or becoming inactive. This is why we create the caret here and destroy it
  when the focus is removed. }

  case FCursorType of
    crsUnderline:
      begin
        FCaretCreated := CreateCaret(Handle, 0, FCharWidth, 2);
        FCaretHeight := 2;
      end;
    crsBlock:
      begin
        FCaretCreated := CreateCaret(Handle, 0, FCharWidth, FLineHeight);
        FCaretHeight := FLineHeight;
      end;
  end;

  PositionCaret;

  SaveAndSetCaretBlinkTime;

  if FCaretCreated then
  begin
    ShowCaret(Handle);
    FCaretShown := TRUE;
  end;
end;

//------------------------------------------------------------------------------

procedure TVaTerminal.WMKillFocus(var Message: TWMSetFocus);
begin
  inherited;
  RestoreCaretBlinkTime;

  if FCaretShown then
  begin
    HideCaret(Handle);
    FCaretShown := FALSE;
  end;

  if FCaretCreated then
  begin
    DestroyCaret;
    FCaretCreated := FALSE;
  end;

end;

//------------------------------------------------------------------------------

function TVaTerminal.WriteChar(Ch: Char): Boolean;
begin
  Result := False;
  if not FDispatcher.Active then Exit;
  Result := FDispatcher.WriteChar(Ch);
end;

//------------------------------------------------------------------------------

function TVaTerminal.WriteBuf(var Buf; Count: Integer): Integer;
begin
  Result := 0;
  if not FDispatcher.Active then Exit;
  Result := FDispatcher.WriteBuf(Buf, Count);
end;

//------------------------------------------------------------------------------

function TVaTerminal.WriteText(const s: string): Boolean;
begin
  Result := False;
  if not FDispatcher.Active then
    Exit;
  Result := FDispatcher.WriteText(s);
end;

//------------------------------------------------------------------------------

procedure TVaTerminal.PositionCaret;
begin
  SetCaretPos(FTextLeft + FCaretX * FCharWidth, FTextTop +
    (FScrollPos + FRows - FCaretY) * FLineHeight - FCaretHeight);
end;

//------------------------------------------------------------------------------

procedure TVaTerminal.AddText(const Text: string);
var
  i: Integer;
begin
  if FCaretCreated and FCaretShown then
  begin
    HideCaret(Handle);
    FCaretShown := False;
  end;

  for i := 1 to Length(Text) do
  begin
    BufferChar[FCaretX, FCaretY] := Text[i];
    inc(FCaretX);
    if FCaretX >= FColumns then
    begin
      if FCaretY = 0 then
        InsertAndScrollF(1)
      else
        dec(FCaretY);

      FCaretX := 0;
    end;
  end;

  PositionCaret;
  UpdateDisplay;

  if FCaretCreated and (FCursorType <> crsNone) then
  begin
    ShowCaret(Handle);
    FCaretShown := True;
  end;
end;

//------------------------------------------------------------------------------

procedure TVaTerminal.AddControlText(const Text: string; const Color, Bkg: TColor);
var
  i, j: Integer;
begin
  if FCaretCreated and FCaretShown then
  begin
    HideCaret(Handle);
    FCaretShown := False;
  end;

  for i := 1 to Length(Text) do
  begin
    case ord(Text[i]) of
    //LF
      10: if ((i > 1) and (Ord(Text[i - 1]) <> 13)) or (i = 1) then
        begin
          if FCaretY = 0 then
            InsertAndScrollF(1)
          else
            dec(FCaretY);
          FCaretX := 0;
        end;
    //CR
      13: if ((i > 1) and (Ord(Text[i - 1]) <> 10)) or (i = 1) then
        begin
          if FCaretY = 0 then
            InsertAndScrollF(1)
          else
            dec(FCaretY);
          FCaretX := 0;
        end;
    //Backspace
      8: if FCaretX > 0 then
        begin
          dec(FCaretX);
          BufferChar[FCaretX, FCaretY] := ' ';
        end;
    //Tab
      9: for j := 1 to TABCOUNT do
        begin
          BufferChar[FCaretX, FCaretY] := ' ';
          inc(FCaretX);
          if FCaretX >= FColumns then
          begin
            if FCaretY = 0 then
              InsertAndScrollF(1)
            else
              dec(FCaretY);
            FCaretX := 0;
          end;
        end;
    else
      begin
        BufferChar[FCaretX, FCaretY] := Text[i];
        BufferColor[FCaretX, FCaretY] := Color;
        BufferBackGround[FCaretX, FCaretY] := Bkg;
        inc(FCaretX);
        if FCaretX >= FColumns then
        begin
          if FCaretY = 0 then
            InsertAndScrollF(1)
          else
            dec(FCaretY);
          FCaretX := 0;
        end;
      end;
    end;
  end;

  PositionCaret;
  UpdateDisplay;

  if FCaretCreated and (FCursorType <> crsNone) then
  begin
    ShowCaret(Handle);
    FCaretShown := True;
  end;
end;

//------------------------------------------------------------------------------

procedure TVaTerminal.SetCaretX(const Value: Integer);
begin
  GotoXY(Value, FCaretY);
end;

//------------------------------------------------------------------------------

procedure TVaTerminal.SetCaretY(const Value: Integer);
begin
  GotoXY(FCaretX, Value);
end;

//------------------------------------------------------------------------------

procedure TVaTerminal.DrawScroll(const nrLines: Integer);
var
  ccol, cbcol: TColor;
  sbuf: TBitmap;
  i, j, xCurent, yCurent, begDraw, finDraw: Integer;
  cline: TVaTerminalLine;
  SCurent: string;
begin
  cline := nil;
  if (nrLines >= (FRows - 2)) or not FDrawFinOK then
  begin
    FFromScroll := False;
    Invalidate;
    Exit;
  end;

  FDrawFinOK := False;
  sbuf := TBitmap.Create;
  try
    sbuf.Assign(FDrawBuffer);
    FDrawBuffer.Canvas.Draw(0, nrLines * FLineHeight, sbuf);
    with FDrawBuffer.Canvas do
    begin
      ccol := FBuffer.Color[0, 0];
      cbcol := FBuffer.Background[0, 0];
      Font := FFont;
      Font.Color := ccol;
      Brush.Color := cbcol;
      Brush.Style := bsSolid;
      TextFlags := TextFlags or ETO_OPAQUE;

      if nrLines > 0 then
      begin
        begDraw := FRows + FScrollPos - nrLines;
        finDraw := FRows - 1 + FScrollPos;
      end
      else
      begin
        begDraw := FScrollPos;
        finDraw := FScrollPos - nrLines - 1;
      end;
      for i := begDraw to finDraw do
      begin
        cline := FBuffer.Line[i];
        yCurent := FTextTop + Height - (i + 1 - FScrollPos) * FLineHeight;
        SCurent := '';
        xCurent := FTextLeft;
        for j := 0 to FColumns - 1 do
        begin
          if ccol <> cline[j].FCol then
          begin
            if SCurent <> '' then
            begin
              Textout(xCurent, yCurent, SCurent);
            end;
            ccol := cline[j].FCol;
            Font.Color := ccol;
            SCurent := '';
            xCurent := FTextLeft + j * FCharWidth;
          end;
          if cbcol <> cline[j].BCol then
          begin
            if SCurent <> '' then
            begin
              Textout(xCurent, yCurent, SCurent);
            end;
            cbcol := cline[j].BCol;
            Brush.Color := cbcol;
            Brush.Style := bsSolid;
            SCurent := '';
            xCurent := FTextLeft + j * FCharWidth;
          end;
          SCurent := SCurent + cline[j].Ch;
        end;
        if SCurent <> '' then
        begin
          Textout(xCurent, yCurent, SCurent);
        end;
      end;
    end;
  finally
    if FCaretCreated and FCaretShown then
    begin
      HideCaret(Handle);
      FCaretShown := False;
    end;
    Canvas.Draw(0, 0, FDrawBuffer);
    sbuf.Free;
    if FCaretCreated and (FCursorType <> crsNone) then
    begin
      PositionCaret;
      ShowCaret(Handle);
      FCaretShown := True;
    end;
    FDrawFinOK := True;

    //Save what has been drawn so Updatedisplay only draws what changed since now
    FOldBuffer.Assign(FBuffer);
    FBuffer.Modified := False;
  end;
end;


//==============================================================================

{ TVaTerminalBuffer }

//------------------------------------------------------------------------------

procedure TVaTerminalBuffer.AdjustHeight(Value: Integer);
var
  BegScr, i, j, dist: Integer;
  tl: TVaTerminalLine;
  tempbuf: TVaTerminalLineArray;
begin
  tl := nil;
  FModified := True;
  if Value = FHeight then
    Exit;

  BegScr := (FWinPos - FOwner.Rows + 1) mod FHeight;

  if Value > FHeight then
  begin
    SetLength(FBuf, Value, FWidth);
    dist := Value - FHeight;
//When adding rows, set them to the default colors
    for i := FHeight to Value - 1 do
      for j := 0 to FWidth - 1 do
      begin
        FBuf[i, j].Ch := #32;
        FBuf[i, j].FCol := FOwner.Font.Color;
        FBuf[i, j].BCol := FOwner.Color;
      end;
      //move to the end the screen information if one part of the screen is at the end and another at the beginning
    if BegScr > (FWinPos mod FHeight) then
    begin
      for i := FHeight - 1 downto (FWinPos + 1) mod FHeight do
      begin
        tl := FBuf[i];
        FBuf[i] := FBuf[i + dist];
        FBuf[i + dist] := tl;
      end;
    end;
    FWinPos := Value + (FWinPos mod FHeight);
  end
  else
  begin
    SetLength(tempbuf, Value);
    j := Value - 1;
    for i := FWinPos downto FWinPos - Value + 1 do
    begin
      tempbuf[j] := FBuf[i mod FHeight];
      dec(j, 1);
    end;
    FBuf := tempbuf;
    FWinPos := Value + Value - 1;
  end;
  FHeight := Value;
end;

//------------------------------------------------------------------------------

procedure TVaTerminalBuffer.AdjustWidth(Value: Integer);
var
  i, j: Integer;
begin
  FModified := True;
  if Value = FWidth then
    Exit;
  SetLength(FBuf, FHeight, Value);
  //When adding columns, set them to the default colors
  if Value > FWidth then
    for i := 0 to FHeight - 1 do
      for j := FWidth to Value - 1 do
      begin
        FBuf[i, j].Ch := #32;
        FBuf[i, j].FCol := FOwner.Font.Color;
        FBuf[i, j].BCol := FOwner.Color;
      end;
  FWidth := Value;
end;

//------------------------------------------------------------------------------

procedure TVaTerminalBuffer.Assign(const Value: TVaterminalBuffer);
var
  i: Integer;
  cl, clo: TVaTerminalLine;
{$IFNDEF VER140}
  j: Integer;
{$ENDIF}
begin
  if (FHeight <> Value.FHeight) or (FWidth <> Value.FWidth) then
    SetLength(FBuf, Value.FHeight, Value.FWidth);
  FHeight := Value.FHeight;
  FWidth := Value.FWidth;
  FOwner := Value.FOwner;
  FWinPos := Value.FWinPos;
  FModified := Value.FModified;

  cl := nil;
  clo := nil;
  for i := FWinPos downto FWinPos - FOwner.FRows + 1 do
  begin
    cl := FBuf[i mod FHeight];
    clo := Value.FBuf[i mod FHeight];
{$IFDEF VER140}
    //This gives internal compiler error in Delphi 5 !!!
    move(clo[0], cl[0], Length(clo) * SizeOf(TVaTerminalChar));
{$ELSE}
    for j := 0 to FWidth - 1 do cl[j] := clo[j];
{$ENDIF}
  end;

end;

procedure TVaTerminalBuffer.ClearBuffer;
var
  i, j: Integer;
  titem: TVaTerminalChar;
begin
  FModified := True;
  titem.Ch := #32;
  titem.FCol := FOwner.Font.Color;
  titem.BCol := FOwner.Color;
  for i := 0 to FHeight - 1 do
    for j := 0 to FWidth - 1 do FBuf[i, j] := titem;
end;

//------------------------------------------------------------------------------

procedure TVaTerminalBuffer.ClearScreen;
var
  i: Integer;
begin
  for i := 0 to FOwner.Rows - 1 do
    InsertLineAndScroll;
end;

//------------------------------------------------------------------------------

constructor TVaTerminalBuffer.Create(AOwner: TVaTerminal; AWidth, AHeight: Integer);
begin
  FOwner := AOwner;
  if AOwner = nil then
    raise EInvalidPointer.Create('Owner of TVaTerminalBuffer cannot be nil !');
  SetLength(FBuf, AHeight, AWidth);
  FWidth := AWidth;
  FHeight := AHeight;
  FWinPos := AHeight;
  FModified := True;
end;

//------------------------------------------------------------------------------

destructor TVaTerminalBuffer.Destroy;
begin
  FBuf := nil;
end;

//------------------------------------------------------------------------------

function TVaTerminalBuffer.GetBackground(X, Y: Integer): TColor;
begin
{$IFDEF AdditionalChecks}
  if (y < 0) or (y >= FHeight) then
    raise ERangeError.Create('Line out of range in TVaTerminalBuffer.GetCh');

  if (x < 0) or (x >= FWidth) then
    raise ERangeError.Create('Column out of range in TVaTerminalBuffer.GetCh');
{$ENDIF}

  Result := FBuf[(FWinPos - Y) mod FHeight, X].BCol;
end;

//------------------------------------------------------------------------------

function TVaTerminalBuffer.GetCh(X, Y: Integer): TVaTerminalChar;
begin
{$IFDEF AdditionalChecks}
  if (y < 0) or (y >= FHeight) then
    raise ERangeError.Create('Line out of range in TVaTerminalBuffer.GetCh');
  if (x < 0) or (x >= FWidth) then
    raise ERangeError.Create('Column out of range in TVaTerminalBuffer.GetCh');
{$ENDIF}
  Result := FBuf[(FWinPos - Y) mod FHeight, X];
end;

//------------------------------------------------------------------------------

function TVaTerminalBuffer.GetCharacter(X, Y: Integer): Char;
begin
{$IFDEF AdditionalChecks}
  if (y < 0) or (y >= FHeight) then
    raise ERangeError.Create('Line out of range in TVaTerminalBuffer.GetCh');
  if (x < 0) or (x >= FWidth) then
    raise ERangeError.Create('Column out of range in TVaTerminalBuffer.GetCh');
{$ENDIF}
  Result := FBuf[(FWinPos - Y) mod FHeight, X].Ch;
end;

//------------------------------------------------------------------------------

function TVaTerminalBuffer.GetColor(X, Y: Integer): TColor;
begin
{$IFDEF AdditionalChecks}
  if (y < 0) or (y >= FHeight) then
    raise ERangeError.Create('Line out of range in TVaTerminalBuffer.GetCh');
  if (x < 0) or (x >= FWidth) then
    raise ERangeError.Create('Column out of range in TVaTerminalBuffer.GetCh');
{$ENDIF}
  Result := FBuf[(FWinPos - Y) mod FHeight, X].FCol;
end;

//------------------------------------------------------------------------------

function TVaTerminalBuffer.GetLine(I: Integer): TVaTerminalLine;
begin
  FModified := True;
{$IFDEF AdditionalChecks}
  if (I >= FHeight) or (I < 0) then
    raise ERangeError.Create('Line out of range in TVaTerminalBuffer.GetLine');
{$ENDIF}
  Result := FBuf[(FWinPos - I) mod FHeight];
end;

//------------------------------------------------------------------------------

procedure TVaTerminalBuffer.InsertLineAndScroll;
var
  nc, j: Integer;
  titem: TVaTerminalChar;
begin
  FModified := True;
  titem.Ch := #32;
  titem.FCol := FOwner.Font.Color;
  titem.BCol := FOwner.Color;
  nc := (FWinPos + 1) mod FHeight;
  for j := 0 to FWidth - 1 do
    FBuf[nc, j] := titem;

  inc(FWinPos);
  if FWinPos >= (3 * FHeight) then
    FWinPos := FHeight + (FWinPos mod FHeight);
end;

//------------------------------------------------------------------------------

procedure TVaTerminalBuffer.SetBackground(X, Y: Integer;
  const Value: TColor);
begin
  FModified := True;
{$IFDEF AdditionalChecks}
  if y >= FHeight then
    raise ERangeError.Create('Line out of range in TVaTerminalBuffer.GetCh');
  if x >= FWidth then
    raise ERangeError.Create('Column out of range in TVaTerminalBuffer.GetCh');
{$ENDIF}
  FBuf[(FWinPos - Y) mod FHeight, X].BCol := Value;
end;

//------------------------------------------------------------------------------

procedure TVaTerminalBuffer.SetCh(X, Y: Integer;
  const Value: TVaTerminalChar);
begin
  FModified := True;
{$IFDEF AdditionalChecks}
  if y >= FHeight then
    raise ERangeError.Create('Line out of range in TVaTerminalBuffer.GetCh');
  if x >= FWidth then
    raise ERangeError.Create('Column out of range in TVaTerminalBuffer.GetCh');
{$ENDIF}
  FBuf[(FWinPos - Y) mod FHeight, X] := Value;
end;

//------------------------------------------------------------------------------

procedure TVaTerminalBuffer.SetCharacter(X, Y: Integer; const Value: Char);
begin
  FModified := True;
{$IFDEF AdditionalChecks}
  if y >= FHeight then
    raise ERangeError.Create('Line out of range in TVaTerminalBuffer.GetCh');
  if x >= FWidth then
    raise ERangeError.Create('Column out of range in TVaTerminalBuffer.GetCh');
{$ENDIF}
  FBuf[(FWinPos - Y) mod FHeight, X].Ch := Value;
end;

//------------------------------------------------------------------------------

procedure TVaTerminalBuffer.SetColor(X, Y: Integer; const Value: TColor);
begin
  FModified := True;
{$IFDEF AdditionalChecks}
  if y >= FHeight then
    raise ERangeError.Create('Line out of range in TVaTerminalBuffer.GetCh');
  if x >= FWidth then
    raise ERangeError.Create('Column out of range in TVaTerminalBuffer.GetCh');
{$ENDIF}
  FBuf[(FWinPos - Y) mod FHeight, X].FCol := Value;
end;

//------------------------------------------------------------------------------

procedure TVaTerminalBuffer.SetLine(I: Integer;
  const Value: TVaTerminalLine);
begin
  FModified := True;
{$IFDEF AdditionalChecks}
  if I >= FHeight then
    raise ERangeError.Create('Line out of range in TVaTerminalBuffer.GetLine');
{$ENDIF}
  FBuf[(FWinPos - I) mod FHeight] := Value;
end;

{ TVaTTYEmulation }

//------------------------------------------------------------------------------

constructor TVaTTYEmulation.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  FLastChar := #32;
end;

//------------------------------------------------------------------------------

procedure TVaTTYEmulation.DataArrived(Data: PVaData; Count: Integer);
var
  s: string;
  i, beg: Integer;
begin
  beg := 0;
  if (FLastChar = #13) and (Data[0] = #10) then
    beg := 1;

  s := '';
  for i := beg to Count - 1 do
    s := s + Data[i];

  FLastChar := Data[Count - 1];
  if FTerminal <> nil then
    FTerminal.AddControlText(s, FTerminal.Font.Color, FTerminal.Color);
end;

//------------------------------------------------------------------------------

destructor TVaTTYEmulation.Destroy;
begin
  inherited;
end;

//------------------------------------------------------------------------------

procedure TVaTTYEmulation.KeyDown(var Key: Word; Shift: TShiftState);
begin
  //Does nothing. In TTY emulation we send only characters
end;

//------------------------------------------------------------------------------

procedure TVaTTYEmulation.KeyPress(var Key: Char);
begin
  if FTerminal = nil then
    Exit;
  if Key = #13 then
    FTerminal.WriteText(#13#10)
  else
    FTerminal.WriteChar(Key);
end;

//------------------------------------------------------------------------------

procedure TVaTTYEmulation.KeyUp(var Key: Word; Shift: TShiftState);
begin
  //Does nothing. In TTY emulation we send only characters
end;

procedure TVaTerminal.RestoreCaretBlinkTime;
begin
  SetCaretBlinkTime(FSavedBlinkTime);
end;

procedure TVaTerminal.SaveAndSetCaretBlinkTime;
begin
  FSavedBlinkTime := GetCaretBlinkTime;
  SetCaretBlinkTime(FBlinkTime);
end;


procedure TVaTerminal.OnCanResizeEventHandler(Sender: TObject;
  var NewWidth, NewHeight: Integer; var Resize: Boolean);
begin
  Resize :=
    (NewWidth = CalculateWidth) and (NewHeight = CalculateHeight)
    or (NewWidth = CalculateWidth) and (NewHeight = Height)
    or (NewWidth = Width) and (NewHeight = CalculateHeight);
  // The OR expressions are necessary because the component is usually resized
  // in two steps: change the width, change the height.
end;

procedure IVaTerminalEmulation.Notification(AComponent: TComponent;
  Operation: TOperation);
begin
  if (AComponent = FTerminal) and (Operation = opRemove) then
    FTerminal := nil;
  inherited;
end;

{ IVaTerminalEmulation }

procedure IVaTerminalEmulation.SetTerminal(const Value: TVaTerminal);
begin
  FTerminal := Value;
  if Value <> nil then Value.FreeNotification(Self);
end;


end.
