{***************************************************************************}
{ TMS Async32                                                               }
{ for Delphi 4.0,5.0,6.0 & C++Builder 4.0,5.0,6.0                           }
{ v1.1                                                                      }
{ Copyright 1996 - 2002 by TMS Software                                     }
{ Email : info@tmssoftware.com                                              }
{ Web : http://www.tmssoftware.com                                          }
{                                                                           }
{ The source code is given as is. The author is not responsible             }
{ for any possible damage done due to the use of this code.                 }
{ The component can be freely used in any application. The complete         }
{ source code remains property of the author and may not be distributed,    }
{ published, given or sold in any form as such. No parts of the source      }
{ code can be included in any other component or application without        }
{ written authorization of the author.                                      }
{***************************************************************************}

unit VaZmodem;

{$I VALIB.INC}

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  VaClasses, VaComm, VaProtocol, VaCRC32;

type
  zHdrType = array[0..3] of Byte;
  zBufType = array[0..1023] of Byte;

  TVaFileProgressEvent = procedure(Sender: TObject; ByteCount: Integer) of object;

  TVaZModemThread = class(TVaProtocolThread)
  private
    FData: zBufType;
    FAttn: zBufType;
    FLastsent: Byte;
    FUseCRC32: Boolean;
    FRxBufSize: Integer;
    FFrameCRCW: Boolean;
    FCanResume: Boolean;
    FGotZSInit: Boolean;
    FRxFileName: string;
    FRxFileSize: longint;
    FRxFileTime: longint;
    FFileName: string;
    FFileSize: Integer;
    FByteCount: Integer;
    function GetFiles: TStrings;
  protected
    procedure FileInfoEvent;
    procedure FileInfo(Name: string; Size: Integer);
    procedure FileProgressEvent;
    procedure FileProgress(ByteCount: Integer);
    procedure Initialize; override;
    function ReadChar: Integer; override;
    function LongintFromHeader(var Hdr: zHdrType): dwLongint;
    procedure LongintToHeader(Value: dwLongint; var Hdr: zHdrType);
    function ReadZChar: Integer;
    procedure PutString(var Buf: zBufType);
    procedure WriteZByte(aByte: Byte);
    procedure WriteHex(aByte: Byte);
    procedure WriteHexHeader(HType: Byte; var Header: zHdrType);
    procedure WriteBinaryHeader32(HType: Byte; var Header: zHdrType);
    procedure WriteBinaryHeader(HType: Byte; var Header: zHdrType);
    function ReadHex: Integer;
    procedure RemoteCan; override;
    function GetZDLE: Integer;
    function ReadHexHeader(var Hdr: zHdrType): Integer;
    function ReadBinaryHeader16(var Hdr: zHdrType): Integer;
    function ReadBinaryHeader32(var Hdr: zHdrType): Integer;
    function ReadHeader(var hdr: zhdrtype): Integer;
    procedure WriteEot;
    procedure RemoteStartup;
    procedure SyncWithReceiver;
    procedure WriteData32(var Buf: zBufType; ByteCount: Integer; FrameEnd: Byte);
    procedure WriteData(var Buf: zBufType; ByteCount: Integer; FrameEnd: Byte);
    procedure NewFilePos(Header: zHdrType);
    procedure WriteFile(FileName: string);
    procedure ReceiveStartup;
    procedure ReceiveConfirm;
    procedure ReceiveCommand;
    function ReadData32(var Buf: zBufType; ByteCount: Integer; var BytesReceived: Integer): Integer;
    function ReadData(var Buf: zBufType; ByteCount: Integer; var BytesReceived: Integer): Integer;
    procedure ExtractFileInfo;
    function ReadFile(Folder: string): string;
    procedure Upload; override;
    procedure Download; override;
    property Files: TStrings read GetFiles;
  end;

  TVaZModem = class(TVaProtocol)
  private
    FFiles: TStrings;
    FTargetFolder: string;
    FOnFileInfo: TVaFileInfoEvent;
    FOnFileProgress: TVaFileProgressEvent;
    procedure SetFiles(Value: TStrings);
  protected
    procedure FileInfo(Name: string; Size: Integer);
    procedure FileProgress(ByteCount: Integer);
    function CreateThread: TVaProtocolThread; override;
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
  published
    property Files: TStrings read FFiles write SetFiles;
    property TargetFolder: string read FTargetFolder write FTargetFolder;
    property OnFileInfo: TVaFileInfoEvent read FOnFileInfo write FOnFileInfo;
    property OnFileProgress: TVaFileProgressEvent read FOnFileProgress write FOnFileProgress;
  end;


function GMTtoPACKT(GMT: Longint): TDateTime;
function Zfromunixdate(S: string): TDateTime;


implementation

uses
  VaUtils;

const
  BSP =  8;
  CAN = 24;

  ZBUFSIZE = 1024;

  ZF0 = 3;
  ZF1 = 2;
  ZF2 = 1;
  ZF3 = 0;

  ZP0 = 0;
  ZP1 = 1;
  ZP2 = 2;
  ZP3 = 3;

  ZPAD   = 42;  { '*' }
  ZDLE   = 24;  { ^X  }
  ZBIN16 = 65;  { 'A' }
  ZHEX   = 66;  { 'B' }
  ZBIN32 = 67;{ 'C' }

  ZCRESUM = 3;

  { bit masks for ZRINIT }
  CANFDX  = 1;   { can handle full-duplex          (yes for PC's)}
  CANOVIO = 2;   { can overlay disk and serial I/O (ditto)       }
  CANBRK  = 4;   { can send a break - True but superfluous       }
  CANCRY  = 8;   { can encrypt/decrypt - not implemented         }
  CANLZW  = 16;  { can LZ compress - not implemented             }
  CANFC32 = 32;  { can use 32 bit crc frame checks - true        }
  ESCALL  = 64;  { escapes all control chars. not implemented    }
  ESC8    = 128; { escapes the 8th bit. not implemented          }

  ZRXDATA = 100;
  ZERROR  = -1;
  GOTOR   = 256;
  GOTCRCE = 360; { 'h' OR 256 }
  GOTCRCG = 361; { 'i' "   "  }
  GOTCRCQ = 362; { 'j' "   "  }
  GOTCRCW = 363; { 'k' "   "  }
  GOTCAN  = 272;  { CAN OR  "  }

  ZCRCE = 104; { 'h' }
  ZCRCG = 105; { 'i' }
  ZCRCQ = 106; { 'j' }
  ZCRCW = 107; { 'k' }
  ZRUB0 = 108; { 'l' }
  ZRUB1 = 109; { 'm' }

  ZRQINIT = 0;
  ZRINIT = 1;
  ZSINIT = 2;
  ZACK = 3;
  ZFILE = 4;
  ZSKIP = 5;
  ZNAK = 6;
  ZABORT = 7;
  ZFIN = 8;
  ZRPOS = 9;
  ZDATA = 10;
  ZEOF = 11;
  ZFERR = 12;
  ZCRC = 13;
  ZCHALLENGE = 14;
  ZCOMPL = 15;
  ZCAN = 16;
  ZFREECNT = 17;
  ZCOMMAND = 18;
  ZSTDERR = 19;

  StrHex = Chr(ZPAD) + Chr(ZHEX);
  StrBin16 = Chr(ZPAD) + Chr(ZBIN16);
  StrBin32 = Chr(ZPAD) + Chr(ZBIN32);


function Time1970(PT: TDateTime): string;
var
  Y70: TDateTime;
  wk: TDateTime;
  st: string;
begin
  Y70 := 25569;  //1970/01/01 00:00:00
  Wk := PT - Y70;
  st := IntToStr(Trunc(Wk) * 86400);
  st := '0' + st;
  Result := st;
end;

function GMTtoPACKT(GMT: Longint): TDateTime;
var
  DT: double;
begin
  inc(GMT,(60*60*9));
  Dt:=(StrToFloat(IntToStr(GMT)) / 86400);
  Result := TDateTime(DT);
end;

function Zfromunixdate(S: string): TDateTime;
var
  secspast: Longint;
  N: Word;
begin
  secspast:=0;
  for n:=1 to length(s) do secspast:=(secspast shl 3)+ord(s[n])-$30;
  Zfromunixdate:=GMTtoPACKT(secspast);
end;

function FileCRC32(Stream: TFileStream): dwLongint;
var
  BytesRead, I: Integer;
  CRC32: dwLongint;
  Buf: array[1..1024] of Byte;
begin
  CRC32 := InitCRC32;
  Stream.Seek(0, soFromBeginning);
  repeat
    BytesRead := Stream.Read(Buf, Sizeof(Buf));
    for I := 1 to BytesRead do
      CRC32 := CalcCRC32(Buf[I], CRC32);
  until (Stream.Position = Stream.Size);
  Result := CRC32;
end;

{ TVaZModemThread }

procedure TVaZModemThread.Initialize;
begin
  inherited;
  FillChar(FAttn, Sizeof(FAttn), 0);
end;

function TVaZModemThread.LongintFromHeader(var Hdr: zHdrType): dwLongint;
begin
  Result := Hdr[ZP3];
  Result := (Result shl 8) or Hdr[ZP2];
  Result := (Result shl 8) OR Hdr[ZP1];
  Result := (Result shl 8) OR Hdr[ZP0];
end;

procedure TVaZModemThread.LongintToHeader(Value: dwLongint; var Hdr: zHdrType);
begin
  Hdr[ZP0] := Byte(Value);
  Hdr[ZP1] := Byte(Value shr 8);
  Hdr[ZP2] := Byte(Value shr 16);
  Hdr[ZP3] := Byte(Value shr 24)
end;

function TVaZModemThread.ReadChar: Integer;
begin
  Result := inherited ReadChar;
  if AbortTransfer then
  begin
    RemoteCan;
    SetExitCode(ErrorLocalCan);
  end;
end;

function TVaZModemThread.ReadZChar: Integer;
(* strips parity and ignores XON/XOFF characters. *)
begin
  repeat
    Result := ReadChar and $FF7F;
  until (not (Result in [17,19]));
end;

function TVaZModemThread.GetZDLE: Integer;
var
  C: Integer;
begin
  C := ReadChar;
  if (C <> ZDLE) then
  begin
    Result := C;
    Exit;
  end;   {got ZDLE or 1st CAN}

  C := ReadChar;
  if (C = CAN) then  {got 2nd CAN}
  begin
    C := ReadChar;
    if (C = CAN) then {got 3rd CAN}
    begin
      C := ReadChar;
      if (C = CAN) then {got 4th CAN}
        C := ReadChar;
    end;
  end;

  { Flags set in high byte }
  case C of
    CAN:
      begin
        Result := ZCAN;
        SetExitCode(ErrorRemoteCan);
      end;
    {got a frame end marker}
    ZCRCE, ZCRCG, ZCRCQ, ZCRCW:
      Result := (C or GOTOR);
    ZRUB0: Result := $007F; {got an ASCII DELete}
    ZRUB1: Result := $00FF  {any parity         }
    else
      begin
        if ((C and $60) = $40) then  {make sure it was a valid escape}
          Result := C xor $40
        else
          Result := ZERROR;
      end;
   end;
end;

procedure TVaZModemThread.PutString(var Buf: zBufType);
var
  I: Integer;
begin
  I := 0;
  while (I < ZBUFSIZE) and (Buf[I] <> 0) do
  begin
    case Buf[I] of
      221:; //break not supported
      222: SysDelay(2000, false);
      else
        WriteChar(Buf[I])
    end;
    Inc(I);
  end;
end;

procedure TVaZModemThread.WriteHex(aByte: Byte);
const
   Hex: array[0..15] of Char = '0123456789abcdef';
begin
  WriteChar(Ord(Hex[aByte shr 4]));
  WriteChar(Ord(Hex[aByte and $0F]));
end;

function TVaZModemThread.ReadHex: Integer;
var
  I: Integer;
  B: array[0..1] of Integer;
begin
  for I := 0 to 1 do
  begin
    B[I] := ReadZChar;
    B[I] := B[I] - $30;
    if (B[I] > 9) then B[I] := B[I] - 39;
    if (B[I] and $FFF0 <> 0) then
    begin
      Result := ZERROR;
      Exit;
    end;
  end;
  Result := (B[0] shl 4) or B[1];
end;

procedure TVaZModemThread.WriteZByte(aByte: Byte);
begin
  if ((aByte and $7F) in [16,17,19,24]) or
    (((aByte and $7F) = 13) and ((FLastsent and $7F) = 64)) then
  begin
    WriteChar(ZDLE);
    FLastsent := (aByte xor 64)
  end else FLastsent := aByte;
  WriteChar(FLastsent);
end;

procedure TVaZModemThread.RemoteCan;
(* Send a zmodem CANcel sequence - 8 CANs and 8 backspaces *)
var
  I: Integer;
begin
  Owner.Dispatcher.PurgeReadWrite;
  for I := 1 to 8 do
  begin
    WriteChar(CAN);
    SysDelay(100, false);
  end;
  for I := 1 to 10 do WriteChar(BSP);
end;

procedure TVaZModemThread.WriteHexHeader(HType: Byte; var Header: zHdrType);
var
  I: Integer;
  CRC16: TCRC16;
begin
  WriteChar(ZPAD);                  { '*' }
  WriteChar(ZPAD);                  { '*' }
  WriteChar(ZDLE);                  { ^X  }
  WriteChar(ZHEX);                  { 'B' }
  WriteHex(HType);
  CRC16.Value := CalcCRC16(HType, 0);
  for I := 0 to 3 do
  begin
    WriteHex(Header[I]);
    CRC16.Value := CalcCRC16(Header[I], CRC16.Value);
  end;
  for I := 1 downto 0 do
    WriteHex(CRC16.Bytes[I]);
  WriteChar(13);
  WriteChar(10);
end;

procedure TVaZModemThread.WriteBinaryHeader32(HType: Byte; var Header: zHdrType);
var
  I: Integer;
  CRC32: TCRC32;
begin
  WriteChar(ZPAD);
  WriteChar(ZDLE);
  WriteChar(ZBIN32);
  WriteZByte(HType);
  CRC32.Value := CalcCRC32(HType, InitCRC32);
  for I := 0 to 3 do
  begin
    WriteZByte(Header[I]);
    CRC32.Value := CalcCRC32(Header[I], CRC32.Value);
  end;
  CRC32.Value := not CRC32.Value;
  for I := 0 to 3 do
    WriteZByte(CRC32.Bytes[I]);
end;

procedure TVaZModemThread.WriteBinaryHeader(HType: Byte; var Header: zHdrType);
var
  I: Integer;
  CRC16: TCRC16;
begin
  if FUseCRC32 then
  begin
    WriteBinaryHeader32(HType, Header);
    Exit;
  end;
  WriteChar(ZPAD);
  WriteChar(ZDLE);
  WriteChar(ZBIN16);
  WriteZByte(HType);
  CRC16.Value := CalcCRC16(HType, 0);
  for I := 0 to 3 do
  begin
    WriteZByte(Header[I]);
    CRC16.Value := CalcCRC16(Header[I], CRC16.Value);
  end;
  for I := 1 downto 0 do
    WriteZByte(CRC16.Bytes[I]);
end;

function TVaZModemThread.ReadHexHeader(var Hdr: zHdrType): Integer;
var
   CRC: Word;
   CRC16: TCRC16;
   C, I: Integer;
begin
  CRC := InitCRC16;
  Result := ReadHex;
  if Result = ZERROR then
    Exit;
  CRC := CalcCRC16(Lo(Result), CRC);

  for I := 0 to 3 do
  begin
    C := ReadHex;
    if C = ZERROR then
    begin
      Result := ZERROR;
      Exit;
    end;
    Hdr[I] := Lo(C);
    CRC := CalcCRC16(Hdr[I], CRC);
  end;

  for I := 1 downto 0 do
  begin
    C := ReadHex;
    if C = ZERROR then
    begin
      Result := ZERROR;
      Exit;
    end;
    CRC16.Bytes[I] := Lo(C);
  end;

  { Check CRC to validate header }
  if CRC <> CRC16.Value then
  begin
    Result := ZERROR;
    Exit;
  end;

  if (PeekChar = 13) then           {throw away CR/LF}
  begin
    ReadChar;
    ReadChar;
  end;
end;

function TVaZModemThread.ReadBinaryHeader16(var Hdr: zHdrType): Integer;
var
  CRC: Word;
  CRC16: TCRC16;
  I, C: Integer;
begin
  CRC := InitCRC16;
  Result := GetZDLE;
  if Result = ZERROR then
    Exit;
  CRC := CalcCRC16(Lo(Result), CRC);

  for I := 0 to 3 do
  begin
    C := GetZDLE;
    if C = ZERROR then
    begin
      Result := ZERROR;
      Exit;
    end;
    Hdr[I] := Lo(C);
    CRC := CalcCRC16(Hdr[I], CRC);
  end;

  for I := 1 downto 0 do
  begin
    C := GetZDLE;
    if C = ZERROR then
    begin
      Result := ZERROR;
      Exit;
    end;
    CRC16.Bytes[I] := Lo(C);
  end;

  { Check CRC to validate header }
  if CRC <> CRC16.Value then
  begin
    Result := ZERROR;
    Exit;
  end;
end;

function TVaZModemThread.ReadBinaryHeader32(var Hdr: zHdrType): Integer;
var
  CRC: dwLongint;
  CRC32: TCRC32;
  I, C: Integer;
begin
  CRC := InitCRC32;
  Result := GetZDLE;
  if Result = ZERROR then
    Exit;
  CRC := CalcCRC32(Lo(Result), CRC);

  for I := 0 to 3 do
  begin
    C := GetZDLE;
    if C = ZERROR then
    begin
      Result := ZERROR;
      Exit;
    end;
    Hdr[I] := Lo(C);
    CRC := CalcCRC32(Hdr[I], CRC);
  end;

  for I := 0 to 3 do
  begin
    C := GetZDLE;
    if C = ZERROR then
    begin
      Result := ZERROR;
      Exit;
    end;
    CRC32.Bytes[I] := Lo(C);
  end;

  CRC := not CRC;
  { Check CRC to validate header }
  if CRC <> CRC32.Value then
  begin
    Result := ZERROR;
    Exit;
  end;
end;

function TVaZModemThread.ReadHeader(var hdr: zhdrtype): Integer;
var
  C: Integer;
  Done: Boolean;
  CanCount: Integer;
  Temp: string;
begin
  Result := 0;
  Done := false;
  CanCount := 0;
  repeat
    C := ReadZChar;
    if C = CAN then
    begin
      Inc(CanCount);
      if CanCount = 5 then
        SetExitCode(ErrorRemoteCan);
    end else
    begin
      CanCount := 0;
      Temp := Temp + Chr(C);
      if Temp = StrHex then
      begin
        Result := ReadHexHeader(Hdr);
        Done := True;
      end else
      if Temp = StrBin16 then
      begin
        FUseCRC32 := false;
        Result := ReadBinaryHeader16(Hdr);
        Done := True;
      end else
      if Temp = StrBin32 then
      begin
        FUseCRC32 := True;
        Result := ReadBinaryHeader32(Hdr);
        Done := True;
      end;
      if Length(Temp) = 2 then Delete(Temp, 1, 1);
    end;

    if Result = ZERROR then //header was garbled
    begin
      Temp := '';
      Owner.Dispatcher.PurgeWrite;
      ErrorEvent(ErrorUnknownHeader, false); //normal error processing
      Done := false;
    end;

  until Done;
end;

procedure TVaZModemThread.WriteData32(var Buf: zBufType; ByteCount: Integer; FrameEnd: Byte);
var
  I: Integer;
  CRC32: TCRC32;
begin
   CRC32.Value := InitCRC32;
   for I := 0 to (ByteCount - 1) do
   begin
     WriteZByte(Buf[I]);
     CRC32.Value := CalcCRC32(Buf[I], CRC32.Value);
   end;
   CRC32.Value := CalcCRC32(FrameEnd, CRC32.Value);

   WriteChar(ZDLE);
   WriteChar(FrameEnd);

   CRC32.Value := not CRC32.Value;
   for I := 0 to 3 do
     WriteZByte(CRC32.Bytes[I]);
end;

procedure TVaZModemThread.WriteData(var Buf: zBufType; ByteCount: Integer; FrameEnd: Byte);
var
  I: Integer;
  CRC16: TCRC16;
begin
  if (FUseCRC32) then
  begin
    WriteData32(Buf, ByteCount, FrameEnd);
    Exit;
  end;

  CRC16.Value := InitCRC16;
  for I := 0 to (ByteCount - 1) do
  begin
    WriteZByte(Buf[I]);
    CRC16.Value := CalcCRC16(Buf[I], CRC16.Value);
  end;
  CRC16.Value := CalcCRC16(FrameEnd, CRC16.Value);
  WriteChar(ZDLE);
  WriteChar(FrameEnd);
  for I := 1 downto 0 do
    WriteZByte(CRC16.Bytes[I]);
end;

procedure TVaZModemThread.WriteEot;
var
  C: Integer;
  Done: Boolean;
  Header: zHdrType;
begin
  Done := false;
  repeat
    LongintToHeader(dwLongint(0), Header);
    WriteBinaryHeader(ZFIN, Header);
    repeat
      C := ReadHeader(Header);
      case C of
        ZNAK: //repeat header
          ErrorEvent(ErrorRepeatHeader, false);
        ZFIN:
          begin
            WriteChar(Ord('O'));
            WriteChar(Ord('O'));
            SysDelay(500, false);
            SetExitCode(ErrorNone);
          end;
        else
          begin
            LongintToHeader(0, Header);
            WriteBinaryHeader(ZNAK, Header);
            ErrorEvent(ErrorUnknownHeader, False);
          end;
      end;
    until C = ZNAK;
  until Done;
end;

procedure TVaZModemThread.RemoteStartup;
var
  Header: zHdrType;
  Done: Boolean;
  C, Position: Integer;
begin
  WriteChar(Ord('r'));
  WriteChar(Ord('z'));
  WriteChar(13);
  LongintToHeader(0, Header);
  WriteHexHeader(ZRQINIT, Header);
  Done := false;
  repeat
    C := ReadHeader(Header);
    case C of
      ZCHALLENGE:
        begin
          Position := LongintFromHeader(Header);
          LongintToHeader(Position, Header);
          WriteHexHeader(ZACK, Header);
        end;
      ZCOMMAND:
        begin
          LongintToHeader(LONGINT(0), Header);
          WriteHexHeader(ZRQINIT, Header);
        end;
      ZRINIT:
        begin
          FRxBufSize := (WORD(Header[ZP1]) shl 8) or Header[ZP0];
          FUseCRC32 := ((Header[ZF0] and CANFC32) <> 0);
          Exit;
        end;
      else
        if (C <> ZRQINIT) then //echo?
        begin
          WriteHexHeader(ZNAK, Header);
          ErrorEvent(ErrorUnknownHeader, false);
        end;
      end; {case}
  until Done;
end;

procedure TVaZModemThread.NewFilePos(Header: zHdrType);
var
  Position: Integer;
begin
  Owner.Dispatcher.PurgeReadWrite;
  Position := LongintFromHeader(Header);
  try
    Stream.Seek(Position, soFromBeginning);
    LongintToHeader(Stream.Position, Header);
    WriteBinaryHeader(ZDATA, Header);
    FFrameCRCW := True;
  except
    WriteHexHeader(ZFERR, Header);
    SetExitCode(ErrorStreamSeek);
  end;
end;

procedure TVaZModemThread.SyncWithReceiver;
var
  C: Integer;
  Done: Boolean;
  Header: zHdrType;
begin
  Done := false;
  repeat
    C := ReadHeader(Header);  //CAN or ZPAD
    case C of
      ZACK: Exit;
      ZRPOS:
        begin
          NewFilePos(Header);
          Done := True;
        end;
      ZCAN: SetExitCode(ErrorRemoteCan);
      ZABORT: //respond with end of batch
        begin
          WriteEot;
          SetExitCode(ErrorRemoteCan);
        end;
      ZSKIP:
        SetExitCode(ErrorNone); //skip this file
      else
        begin  //unknown header...
          LongintToHeader(0, Header);
          WriteBinaryHeader(ZNAK, Header);
        end;
    end;
  until Done;
end;

procedure TVaZModemThread.WriteFile(FileName: string);
var
  C, FrameEnd, BufOutSize, BytesRead: Integer;
  Header: zHdrType;
  FInfo, FN, FS, FD: string;

  procedure ResetBufOut;
  begin
    FillChar(FData, Sizeof(FData), 0);
  end;

begin
  FN := ExtractFileName(FileName);
  FD := Time1970(FileDateToDateTime(FileAge(FileName)));

  OpenStream(FileName);
  try
    FileInfo(FileName, Stream.Size);

    Str(Stream.Size, FS);
    FInfo := AnsiLowerCase(FN+#0+FS+#32+FD);

    if (FRxBufSize = 0) or (FRxBufSize > ZBUFSIZE) then
      BufOutSize := ZBUFSIZE
    else BufOutSize := FRxBufSize;

    repeat
      FillChar(Header, Sizeof(Header), 0);
      Header[ZF0] := ZCRESUM; {can recover}
      WriteBinaryHeader(ZFILE, Header);

      Move(Pointer(FInfo)^, FData, Length(FInfo));
      WriteData(FData, ZBUFSIZE, ZCRCW);
      repeat
        C := ReadHeader(Header);
        case C of
          ZNAK: //repeat header
            ErrorEvent(ErrorRepeatHeader, false);
          ZSKIP:
            SetExitCode(ErrorNone); //continue with next file
          ZCRC:
            begin
              LongintToHeader(FileCRC32(Stream), Header);
              WriteHexHeader(ZCRC, Header);
            end;
          ZRPOS:
            begin
              NewFilePos(Header);
              FFrameCRCW := false;
            end;
        end;
      until (C = ZRPOS) or (C = ZNAK); //ZNAK = repeat header
    until C = ZRPOS; //ZFILE loop

    repeat

      repeat //ZCRCE loop
        ResetBufOut;
        BytesRead := 0;
        try
          BytesRead := Stream.Read(FData, BufOutSize);
        except
          RemoteCan;
          SetExitCode(ErrorStreamRead);
        end;

        if FFrameCRCW then
        begin
          FrameEnd := ZCRCW;
          FFrameCRCW := false;
        end else
        if EndOfStream then
          FrameEnd := ZCRCE
        else FrameEnd := ZCRCG;

        WriteData(FData, BytesRead, FrameEnd);

        FileProgress(Stream.Position);

        if FrameEnd = ZCRCW then
        begin
          SyncWithReceiver;
          FFrameCRCW := false;
        end;

        while CharReceived do
        begin
          if PeekChar in [CAN, ZPAD] then
          begin
            ResetBufOut;
            Owner.Dispatcher.PurgeWrite;
            WriteData(FData, 0, ZCRCE);
            SyncWithReceiver;
          end else ReadChar;  //remove character...
        end;

        if AbortTransfer then
        begin
          RemoteCan;
          SetExitCode(ErrorLocalCan);
        end;

      until FrameEnd = ZCRCE; //eof

      repeat
        LongintToHeader(Stream.Position, Header);
        WriteBinaryHeader(ZEOF, Header);
        C := ReadHeader(Header);
        case C of
          ZNAK:
            ErrorEvent(ErrorRepeatHeader, false); //repeat header
          ZRPOS:
            NewFilePos(Header); {Oops, we need to write some more}
          ZRINIT:; //file received ok
          ZFERR:   //receiver in trouble
            begin
              RemoteCan;
              SetExitCode(ErrorRemoteFileErr);
            end;
        end;
      until (C = ZRINIT) or (C = ZRPOS);

    until C = ZRINIT; //main ZDATA loop

  finally
    CloseStream;
  end;
end;

function TVaZModemThread.GetFiles: TStrings;
begin
  Result := TVaZModem(Owner).Files;
end;

procedure TVaZModemThread.Upload;
var
  Index: Integer;
begin
  FLastSent := 0;
  RemoteStartup;
  for Index := 0 to Files.Count - 1 do
  begin
    if FileExists(Files[Index]) then
      WriteFile(Files[Index]);
    if Index = Files.Count - 1 then
      WriteEot;
  end;
end;

{ ***** RECEIVER PROCS ***** }

function TVaZModemThread.ReadData32(var Buf: zBufType; ByteCount: Integer;
  var BytesReceived: Integer): Integer;
var
  C, I: Integer;
  CRC: dwLongint;
  CRC32: TCRC32;
  Done: Boolean;
  FrameEnd: Integer;
begin
  BytesReceived := 0;
  CRC := InitCRC32;
  Result := ZERROR;
  Done := FALSE;
  repeat
    C := GetZDLE;
    if C = ZERROR then
    begin
      Result := ZERROR;
      Exit;
    end;

    if (Hi(C) <> 0) then
    begin
      if (C = GOTCRCE) or (C = GOTCRCG) or
         (C = GOTCRCQ) or (C = GOTCRCW) then
      begin
        FrameEnd := C;
        CRC := CalcCRC32(Lo(C), CRC);

        for I := 0 to 3 do
        begin
          C := GetZDLE;
          if C = ZERROR then
          begin
            Result := ZERROR;
            Exit;
          end;
          CRC32.Bytes[I] := Lo(C);
        end;
        CRC := not CRC;
        if CRC <> CRC32.Value then
          Result := ZERROR
        else
          Result := FrameEnd;
        Exit;
      end;
    end;

    Dec(ByteCount);
    if ByteCount < 0 then
    begin
      RemoteCan;
      SetExitCode(ErrorPacketToLarge);
    end;
    Buf[BytesReceived] := Lo(C);
    Inc(BytesReceived);
    CRC := CalcCRC32(Lo(C), CRC);
  until Done;
end;

function TVaZModemThread.ReadData(var Buf: zBufType; ByteCount: Integer;
  var BytesReceived: Integer): Integer;
var
  C, I: Integer;
  CRC: Word;
  CRC16: TCRC16;
  Done: Boolean;
  FrameEnd: Integer;
begin
  if (FUseCRC32) then
  begin
    Result := ReadData32(Buf, ByteCount, BytesReceived);
    Exit;
  end;
  Result := ZERROR;
  BytesReceived := 0;
  CRC := InitCRC16;
  Done := FALSE;
  repeat
    C := GetZDLE;
    if C = ZERROR then
    begin
      Result := ZERROR;
      Exit;
    end;

    if (Hi(C) <> 0) then
    begin
      if (C = GOTCRCE) or (C = GOTCRCG) or
         (C = GOTCRCQ) or (C = GOTCRCW) then
      begin
        FrameEnd := C;
        CRC := CalcCRC16(Lo(C), CRC);

        for I := 1 downto 0 do
        begin
          C := GetZDLE;
          if C = ZERROR then
          begin
            Result := ZERROR;
            Exit;
          end;
          CRC16.Bytes[I] := Lo(C);
        end;

        if CRC <> CRC16.Value then
          Result := ZERROR
        else Result := FrameEnd;
        Exit;
      end;
    end;

    Dec(ByteCount);
    if ByteCount < 0 then
    begin
      RemoteCan;
      SetExitCode(ErrorPacketToLarge);
    end;
    Buf[BytesReceived] := Lo(C);
    Inc(BytesReceived);
    CRC := CalcCRC16(Lo(C), CRC);
  until Done;
end;

procedure TVaZModemThread.ReceiveConfirm;
(* Acknowledge the other ends request to terminate *)
var
  C, Cnt: Integer;
  Header: zHdrType;
  Done: Boolean;
  TimeEvent: TVaTimeEvent;
begin
  Cnt := 0;
  Owner.Dispatcher.PurgeRead;
  Done := false;
  repeat
    LongintToHeader(0, Header);
    WriteHexHeader(ZFIN, Header);

    if not CharReceived then
    begin
      InitTimer(TimeEvent, 10000); //10 seconds wait
      repeat
        if TimerExpired(TimeEvent) then
          Exit; //exit anyway...don't wait for the 'OO'
      until (CharReceived) or (AbortTransfer);
    end;

    while CharReceived do
    begin
      C := ReadZChar;
      if C = 79 then
      begin
        Inc(Cnt);
        if Cnt = 2 then
          SetExitCode(ErrorNone);
      end else Cnt := 0;
    end;
  until Done;
end;

//Processing of ZModem ZCommand
procedure TVaZModemThread.ReceiveCommand;
var
  C: Integer;
  Done: Boolean;
  Header: zHdrType;
  BytesReceived: Integer;
begin
  C := ReadData(FData, ZBUFSIZE, BytesReceived);
  if (C = GOTCRCW) then
  begin
    Done := false;
    repeat //exit by MaxError or ReceiveConfirm
      LongintToHeader(0, Header);
      WriteHexHeader(ZCOMPL, Header);
      repeat
        C := ReadHeader(Header);
        case C of
          ZNAK: //repeat header
            ErrorEvent(ErrorRepeatHeader, false);
          ZFIN:
            begin
              ReceiveConfirm;
              Exit;
            end;
        end;
      until C = ZNAK;
    until Done;
  end;
end;

procedure TVaZModemThread.ReceiveStartup;
var
  I, C: Integer;
  Header: zHdrType;
  BytesReceived: Integer;
begin
  for I := 0 to 10 do
  begin
    LongintToHeader(0, Header);
    Header[ZF0] := CANFDX OR CANOVIO OR CANFC32; {Full dplx, overlay I/O and CRC32}
    WriteHexHeader(ZRINIT, Header);
    repeat
      C := ReadHeader(Header);

      case C of
        ZRQINIT:   // v1.1 : ZRQINIT handling added 
          begin
            WriteHexHeader(ZRINIT, Header);
          end;
        ZNAK: //repeat header
          ErrorEvent(ErrorRepeatHeader, false);
        ZFILE:
          begin
            FCanResume := Header[ZF0] = ZCRESUM;
            C := ReadData(FData, ZBUFSIZE, BytesReceived);
            if (C = GOTCRCW) then Exit;
            WriteHexHeader(ZNAK, Header);
            ErrorEvent(ErrorInvalidPacket, false);
          end;
        ZSINIT:  //receive attention string
          begin
            C := ReadData(FAttn, ZBUFSIZE, BytesReceived);
            if (C = GOTCRCW) then
            begin
              FGotZSInit := True;
              WriteHexHeader(ZACK, Header);
            end else
            begin
              WriteHexHeader(ZNAK, Header);
              ErrorEvent(ErrorInvalidPacket, false);
            end;
          end;
        ZFREECNT:
          begin
            LongintToHeader(dwLongint(DiskFree(0)), Header); //could cause a problem!
            WriteHexHeader(ZACK, Header);
          end;
        ZCOMMAND:
          begin
            ReceiveCommand;
            WriteHexHeader(ZNAK, Header);
            ErrorEvent(ErrorZCommand, false);
          end;
        ZCAN:
          SetExitCode(ErrorRemoteCan);
        ZFIN:
          begin
            ReceiveConfirm;
            WriteHexHeader(ZNAK, Header);
            ErrorEvent(ErrorOutOfSync, false);
          end;
        else
          ErrorEvent(ErrorUnknownHeader, false);
      end;
    until (C = ZNAK);
  end;
  SetExitCode(ErrorRemoteStartup);
end;

procedure TVaZModemThread.ExtractFileInfo;
var
  I: Integer;
  Temp: string;
begin
  FRxFileName := '';
  FRxFileSize := 0;
  FRxFileTime := 0;

  I := 0;
  while FData[I] <> 0 do
  begin
    FRxFileName := FRxFileName + Chr(FData[I]);
    Inc(I);
  end;
  Inc(I);

  Temp := '';
  while FData[I] > 32 do
  begin
    Temp := Temp + Chr(FData[I]);
    Inc(I);
  end;

  try
    FRxFileSize := StrToInt(Temp);
  except
    FRxFileSize := 0;
  end;
  Inc(I);

  while (FData[I] >= Ord('0')) and
        (FData[I] <= Ord('9')) do
  begin
    FRxFileTime := FRxFileTime*8 + FData[I] - ord('0');
    Inc(I);
  end;
end;

function TVaZModemThread.ReadFile(Folder: string): string;
var
  FileName: string;
  Done, Completed: Boolean;
  C, BytesReceived: Integer;
  Header: zHdrType;

  procedure SetPosition;
  begin
    LongintToHeader(Stream.Position, Header);
    WriteBinaryHeader(ZRPOS, Header);
  end;

  procedure SaveData(ByteCount: Integer);
  begin
    try
      Stream.Write(FData, ByteCount);
      FileProgress(Stream.Position);
    except
      RemoteCan;
      SetExitCode(ErrorStreamWrite);
    end;
  end;

begin
  Result := '';
  ReceiveStartup; //ZFILE or exit
  ExtractFileInfo;

  FileName := Folder + ExtractFileName(FRxFileName);
  if FileExists(FileName) then
  begin
    if FCanResume then OpenStream(FileName)
    else
    begin
      FileName := CreateUniqueFileName(FileName);
      CreateStream(FileName);
    end;
  end else CreateStream(FileName);

  try
    Stream.Seek(Stream.Size, soFromBeginning);
  except
    LongintToHeader(0, Header);
    WriteHexHeader(ZSKIP, Header);
    ErrorEvent(ErrorStreamSeek, false);
    CloseStream;
    Exit;
  end;

  FileInfo(FileName, FRxFileSize);

  try
    SetPosition;
    Done := False;
    repeat
      C := ReadHeader(Header);
      case C of
        ZNAK: //repeat header
          begin
            ErrorEvent(ErrorRepeatHeader, false);
            SetPosition;
          end;
        ZDATA:
          begin
            Completed := false;
            repeat
              C := ReadData(FData, ZBUFSIZE, BytesReceived);
              case C of
                ZERROR:
                  begin
                    ErrorEvent(ErrorInvalidPacket, false);
                    PutString(FAttn);
                    SetPosition;
                    Completed := True;
                  end;
                GOTCRCG: //read some more
                  SaveData(BytesReceived);
                GOTCRCE:
                  begin
                    SaveData(BytesReceived);
                    Completed := True; //wait for eof
                  end;
                GOTCRCW, GOTCRCQ: //need acknowledge
                  begin
                    SaveData(BytesReceived);
                    LongintToHeader(0, Header);
                    WriteHexHeader(ZACK, Header);
                    Completed := True; //wait for next zdata header...
                  end;
              end;
            until Completed;
          end; //zdata
        ZEOF: Done := True;
      end;
    until Done;
    Result := FileName;
  finally
    CloseStream;
  end;
end;

procedure TVaZModemThread.Download;
var
  BatchDone: Boolean;
  Folder, FileName: string;
begin
  Files.Clear;
  with TVaZModem(Owner) do
    Folder := AddPathSlash(TargetFolder);
  FLastSent := 0;
  FGotZSInit := False;
  BatchDone := false;
  repeat
    FileName := ReadFile(Folder);
    if FileName <> '' then //zskip
      Files.Add(FileName);
  until BatchDone;
end;

procedure TVaZModemThread.FileInfoEvent;
begin
  TVaZModem(Owner).FileInfo(FFileName, FFileSize);
end;

procedure TVaZModemThread.FileInfo(Name: string; Size: Integer);
begin
  FFileName := Name;
  FFileSize := Size;
  Synchronize(FileInfoEvent);
end;

procedure TVaZModemThread.FileProgressEvent;
begin
  TVaZModem(Owner).FileProgress(FByteCount);
end;

procedure TVaZModemThread.FileProgress(ByteCount: Integer);
begin
  FByteCount := ByteCount;
  Synchronize(FileProgressEvent);
end;

{ TVaZModem }

constructor TVaZModem.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  FFiles := TStringList.Create;
end;

destructor TVaZModem.Destroy;
begin
  FFiles.Free;
  inherited Destroy;
end;

procedure TVaZModem.FileInfo(Name: string; Size: Integer);
begin
  try
    if Assigned(FOnFileInfo) then
      FOnFileInfo(Self, Name, Size);
  except
    Application.HandleException(Self);
  end;
end;

procedure TVaZModem.FileProgress(ByteCount: Integer);
begin
  try
    if Assigned(FOnFileProgress) then
      FOnFileProgress(Self, ByteCount);
  except
    Application.HandleException(Self);
  end;
end;

procedure TVaZModem.SetFiles(Value: TStrings);
begin
  FFiles.Assign(Value);
end;

function TVaZModem.CreateThread: TVaProtocolThread;
begin
  Result := TVaZModemThread.Create(Self);
end;


end.
